#!/bin/bash
echo "================= cook configurator (part of MACSIMUS) ====================="
echo "|  This script must be called from the macsimus/cook directory             |"
echo "|  The cook version will be configured and compiled in a SUBDIRectory      |"
echo "|  Synopsis:                                                               |"
echo "|    configure.sh [SUBDIR [EXENAME]]                                       |"
echo "|  Environment:                                                            |"
echo "|    file macsimus/makemake.sh contains options passed to makemake (df.=cc)|"
echo "|  You will be asked for compile-time options:                             |"
echo "|    Answer by a letter or a number; just <Enter> means \"no\" or \"default\"  |"
echo "============================================================================"

if [ -e ../makemake.sh ] ; then
  source ../makemake.sh
else
  echo "../makemake.sh not found - MACSIMUS has not been properly configured!"
  echo "Enter compiler name:"
  read
  MAKEMAKE=$REPLY
fi

if [ "$MAKEMAKE" = "" ]; then
  MAKEMAKE=cc
  echo "NOTE: $MAKEMAKE will be passed to makemake"
fi

EXENAME=""
if [ $# -gt 0 ] ; then
  SUBDIR=$1
  echo NOTE: $SUBDIR/ specified as subdirectory name
  if [ $# -gt 1 ] ; then
    EXENAME="$2"
  echo NOTE: $EXENAME specified as the executable name
  fi
else
  echo -n "Enter subdirectory name (will be created if necessary): "
  read
  SUBDIR=$REPLY
  if [ "$SUBDIR" = "" ] ; then
    echo "no directory specified - stop"
    exit
  fi
fi

if [ ! -e $SUBDIR ] ; then
  mkdir -p $SUBDIR
  echo $PWD/$SUBDIR created
else
  if [ -d $SUBDIR ] ; then
    echo "NOTE: directory $PWD/$SUBDIR/ already exists"
  else
    echo "ERROR: $SUBDIR exists and is not a directory"
    exit
  fi
fi

cd $SUBDIR

NAME="cook"
MODULE=""
PROJECT="COOK"

if [ -e simopt.h ] ; then
  mv simopt.h simopt.h~
  echo "NOTE: existing (previous) simopt.h was renamed to simopt.h~"
  echo
fi

echo "/* This file was generated by configure.sh */" > simopt.h
echo "/* See MACSIMUS/generic/simopt.h for more details */" >> simopt.h
echo "#define COOK" >> simopt.h

echo "===== Select BOUNDARY CONDITIONS:"
echo "> p = Periodic in 3D, spherical cut-off of Lennard-Jones interactions"
echo "      Allows later modifications (wall, slab) (DEFAULT)"
echo "  f = Free (no boundary), true electrostatic 1/r (no cutoff)"
echo "      Use for e.g. one molecule in vacuum"
echo "  n = Periodic in 3D, all pair interactions use nearest-image distance"
echo "      (special, cutoff corections not implemented; NOT TESTED RECENTLY)"
read
BC=$REPLY
if [ "$BC" = "" ] ; then
  BC='p'
fi
case $BC in
n )
  NAME=$NAME"ni"
  PROJECT=$PROJECT"-ni"
  MODULE=$MODULE" ewald"
  COULOMB=0 # to ask correctly for widom
  echo "#define NIBC" >> simopt.h
  ;;
f )
  NAME=$NAME"free"
  PROJECT=$PROJECT"-free"
  COULOMB=0 # to ask correctly for widom
  echo "#define FREEBC" >> simopt.h
  ;;
p )
  echo "===== Select algorithm for ELECTROSTATIC calculations:"
  echo " -1 = Ewald summation, most accurate and most time consuming"
  echo "      All r-space terms (erfc) are calculated by hyperbolic splines"
  echo ">-2 = The same as -1 with more accurate erfc for 1-2 and 1-3 interactions"
  echo "      For models with partial charges close together (DEFAULT)"
  echo "  0 = Cut-and-shift approximation (MACSIMUS style) calculated directly"
  echo "  2 = Cut-and-shift approximation (MACSIMUS style) by quadratic splines"
  echo "      0 and 2 are equivalent, 0 is usually a bit faster"
  echo "  3 = Cut-and-shift approximation by Fennell, Gezelter, JCP 124, 234104 (2006)"
  echo "      Not suitable for models with large partial charges very close together"
  read
  MODULE=$MODULE" elst ewald"
  COULOMB=$REPLY
  if [ "$COULOMB" = "" ] ; then
    COULOMB="-2"
  fi
  echo "#define COULOMB" ${COULOMB} >> simopt.h
  case $COULOMB in
  -1 | -2 )
     PROJECT=$PROJECT"-Ewald"
     NAME=$NAME"ew"
     ;;
   0 | 2 )
     PROJECT=$PROJECT"-cutelst"$COULOMB
     NAME=$NAME"ce"
     ;;
   3 ) NAME=$NAME"fg" ;;
   * ) echo "ERROR: unsupported algorithm"; exit ;;
  esac
  ;;
esac

echo "===== Select POLARIZABILITY support:"
echo "      (polarizability requires a special version of blend and force field)"
echo "> n = No polarizability (DEFAULT)"
echo "  0 = Only scalar linear polarizability supported."
echo "      More features can be obtained by summing the following flags:"
echo "  1 = Repulsive antipolarizability (shell core), all terms included"
echo "  2 = Saturated polarizability"
echo "  4 = Use this if there are also uncharged atoms (faster code)"
echo "  8 = Axial polarizability"
# 16 = old development version: see old+misc/constrdaa.c and old+misc/gear2polg.c
echo " 32 = Fluctuating charge model (under development)"
echo " 64 = incl. intramolecular induced dipole--induced dipole interactions (ADIM)"
read
POLAR=$REPLY
if [ "$POLAR" = "" ] ; then
  POLAR="n"
fi
if [ "$POLAR" != "n" ] ; then
  echo "#define POLAR $POLAR" >> simopt.h
  NAME=$NAME"p"$POLAR
  PROJECT=$PROJECT"-polar"$POLAR
fi

echo "===== Select DETAILS of boundary conditions:"
echo "> p = Plain version (DEFAULT)"
echo "  s = Slab (includes: z-forces, density profiles, surface tension)"
echo "  w = Wall (at z=0) or a slit pore"
echo "  g = Gold (conducting) wall, charge inversion method"
echo "      Incompatible with Ewald+linked-cell list; NOT TESTED RECENTLY"
echo "NOTES: Versions w and g imply z-density profile measurements"
echo "       WIDOM (if selected) is measured in z-slabs, too"
read
SV=$REPLY
if [ "$SV" = "" ] ; then
  SV="p"
fi
case $SV in
s )
  PROJECT=$PROJECT"-slab"
  NAME=$NAME"s"
  echo "#define SLAB 0" >> simopt.h
  ;;
w )
  PROJECT=$PROJECT"-wall"
  NAME=$NAME"w"
  echo "#define WALL" >> simopt.h
  echo "#define SLAB 0" >> simopt.h
  ;;
g )
  NAME=$NAME"g"
  PROJECT=$PROJECT"-gold"
  echo "#define GOLD" >> simopt.h
  echo "#define SLAB 0" >> simopt.h
  ;;
* )
esac

echo "===== The NON-BONDED forces are:"
echo "      (subdirectory of sim/ with sitesite.c and sitesite.h is specified)"
echo "      (non-lj requires special version of blend and force field)"
echo "> lj       = Lennard-Jones (DEFAULT), various combining rules available"
echo "  exp6     = exp-6 with Lennard-Jones-like combining rules, POLAR=0 supported"
echo "  buck     = exp-6 with Buckingham combining rules"
echo "             compatible both with POLAR&1 (shell-core) and simple POLAR"
echo "  busing   = exp-6 with Busing combining rules + r^-6 (use with POLAR&1)"
echo "  metal    = tight-binding potential"
echo "  wcalj    = Lennard-Jones cut at the minimum and shifted"
read
NONBOND=$REPLY
if [ "$NONBOND" = "" ] ; then
  NONBOND=lj
elif [ "$NONBOND" = "buck" -o "$NONBOND" = "busing" ] ; then
  echo "===== Add term D/r^12 (#define REP12, applies to exp-6-like potentials) (y/N)?"
  read
  if [ "$REPLY" == "y" ]; then
    echo "#define REP12" >> simopt.h
  fi
fi

if [ $BC = "f" ]; then
  MODULE=$MODULE" water"
else
  echo "===== Are all your molecules smaller than half the box size (y/N)?"
  echo "      (It is safer to say No if you are not sure)"
  read
  if [ "$REPLY" != "y" ]; then
    echo "#define WORM" >> simopt.h
  fi

  echo "===== Select algorithm for PAIR FORCES calculations (affects speed):"
  echo "      (The cutoff refers to the elst cutoff; always cutoff>=LJcutoff)"
  echo "> d = Direct evaluation of pair forces      (accepts PARALLEL=2)"
  echo "      The cutoff is comparable to half the box size (DEFAULT)"
  echo "  c = Direct evaluation of pair forces      (accepts PARALLEL=2)"
  echo "      The cutoff is significantly smaller than half the box size"
  echo "  l = Linked-cell list method               (accepts PARALLEL=1 and 2)"
  echo "      Efficient for large systems and/or the cutoff sphere significantly"
  echo "      smaller than the box.  WARNING: not compatible with all other options"
  echo "  l0 = As above, with local copies of forces within cells"
  echo "  l1 = As above, with PARALLEL=1: 1st site in pair indirected, 2nd copied"
  echo "      (Which linked-cell version is more efficient depends mainly on the"
  echo "      processor and cache architecture, data size, and parallelism)"
  read
  ALG=$REPLY
  if [ "$ALG" = "" ] ; then
    ALG="d"
  fi
  case $ALG in
  c )
    NAME=$NAME"c"
    MODULE=$MODULE" water"
    echo "#define CUT" >> simopt.h
    ;;
  l )
    MODULE=$MODULE" linklist"
    NAME=$NAME"lc"
    PROJECT=$PROJECT"-lc"
    echo "#define LINKCELL 3" >> simopt.h
    ;;
  l0 )
    MODULE=$MODULE" linklist"
    NAME=$NAME"lc"
    PROJECT=$PROJECT"-lc0"
    echo "#define LINKCELL 0" >> simopt.h
    ;;
  l1 )
    MODULE=$MODULE" linklist"
    NAME=$NAME"lc"
    PROJECT=$PROJECT"-lc1"
    echo "#define LINKCELL 1" >> simopt.h
    ;;
  * )
    MODULE=$MODULE" water"
    ;;
  esac

  echo "===== Select PARALLELIZATION (linux shared memory + threads):"
  echo "      (Not all combinations with the pair-forces method are valid)"
  echo "> 0 = standard serial version (DEFAULT)"
  echo "  1 = Linked-cell list method parallelized in x-slabs, Ewald parallelized"
  echo "  2 = pair forces running in parallel with Ewald (2 threads)"
  read
  PAR=$REPLY
  if [ "$PAR" == "0" ] ; then
# PARALLEL=0 is the same as undef, so if 0 given, no P suffix...
     PAR=""
  fi
  if [ "$PAR" != "" ] ; then
    echo "#define PARALLEL $PAR" >> simopt.h
    PROJECT=$PROJECT"-par"$PAR
    NAME=$NAME"P"$PAR
  fi

  echo "===== Request PRESSURE TENSOR calculations; sum the following flags:"
  echo "> 0 = no pressure tensor calculated (DEFAULT)"
  echo "  1 = calculate the virial (static) part incl. constraint forces"
  echo "  2 = calculate the kinetic part"
  echo "  4 = calculate also off-diagonal terms Pxy,Pyz,Pxz"
  echo "      (if not set, only diagonal terms Pxx,Pyy,Pzz are calculated)"
  echo "  8 = calculate also center-of-mass based (molecular) tensor"
  echo "      - applies for the kinetic part only (use with 2)!"
  echo "      (if not set, only site-based part is calculated)"
  echo " 16 = SPECIAL: kurtosis and 6-th normalized central moment of v_CM"
  echo "HINT: For direct calculation of surface tension, use 3 for models with"
  echo "      constrained bonds or 1 for vibrating bonds or monoatomic models"
  read
  PRESSURETENSOR=$REPLY
  if [ "$PRESSURETENSOR" != "" ] ; then
    echo "#define PRESSURETENSOR $PRESSURETENSOR" >> simopt.h
    PROJECT=$PROJECT"-PT"$PRESSURETENSOR
  fi
fi

if [ "$NONBOND" = "metal" ] ; then
  echo "#define METAL" >> simopt.h
  MODULE=${MODULE/water/}
  PROJECT=$PROJECT"-METAL"
fi

echo "#define VERLET 3" >> simopt.h
echo "#define SHAKE 1" >> simopt.h

if [ $COULOMB -ge "0" ]; then
  echo "===== Add support for the Widom particle insertion method (y/N)"
  read
  if [ "$REPLY" = "y" ]; then
    echo "#define WIDOM" >> simopt.h
  fi
fi

echo "===== Add support for \"ANCHORING\" a molecule and mean force calculations (y/N)"
read
if [ "$REPLY" = "y" ]; then
  echo "#define ANCHOR" >> simopt.h
fi

echo "===== Add support for RADIUS OF GYRATION, end-to-end distance, and direct"
echo "calculation of the VELOCITY-VELOCITY time autocorrelation functions (y/N)"
read
if [ "$REPLY" = "y" ]; then
  echo "#define RGYR" >> simopt.h
fi

echo "===== Add support for dihedral (proper torsion) angle distribution measurements:"
echo "> n = None (DEFAULT)"
echo "  2 = Only the central bond distinguishes dihedrals"
echo "  4 = All 4 atoms distinguish dihedrals"
echo " -1 = All dihedrals in all molecules different:"
echo "      Support for selection (SIMNAME.ddh) enabled"
echo "  1 = One summary histogram for all dihedrals"
read
DIH=$REPLY
if [ "$DIH" = "" ] ; then
  DIH="n"
fi
if [ $DIH != "n" ]; then
  echo "#define DIHHIST" $DIH >> simopt.h
fi

echo "===== Add support for CROSS-SECTION measurements (y/N)"
read
if [ "$REPLY" = "y" ]; then
  echo "#define XSECTION" >> simopt.h
fi

if [ $BC != "f" ]; then
  echo "===== Add support for VISCOSITY measurements by NEMD (periodic z-force) (y/N)"
  read
  if [ "$REPLY" = "y" ]; then
    echo "#define SHEAR" >> simopt.h
  fi

  echo "===== Add support for CLUSTER analysis (y/N)"
  read
  if [ "$REPLY" = "y" ]; then
    echo "#define CLUSTERS" >> simopt.h
  fi
fi

echo "===== Add support for HARMONIC EXPANSION coefficients and full correlation function:"
echo "> n = None (DEFAULT)"
echo "  1 = (methanol?)"
echo "  2 = water"
echo "  3 = linear molecule"
read
HARM=$REPLY
if [ "$HARM" = "" ] ; then
  HARM="n"
fi
if [ $HARM != "n" ]; then
  echo "#define HARMONICS" $HARM >> simopt.h
fi

echo "#define PROJECT" \"$PROJECT\" >> simopt.h

if [ -e metamake ] ; then
  mv metamake metamake~
  echo "NOTE: existing (previous) metamake was renamed to metamake~"
  echo
fi

echo "!! This file was generated by configure.sh" > metamake
echo "!! See MACSIMUS/generic/metamake more details" >> metamake
echo >> metamake
echo "O = -DSCR" >> metamake

if [ -e ../../home.mmk ]; then
  echo "!include \"../../home.mmk\"" >> metamake
  echo "!include \"../../compile.mmk\"" >> metamake
elif [ -e ../../../home.mmk ]; then
  echo "!include \"../../../home.mmk\"" >> metamake
  echo "!include \"../../../compile.mmk\"" >> metamake
elif [ -e ../../../../home.mmk ]; then
  echo "!include \"../../../../home.mmk\"" >> metamake
  echo "!include \"../../../../compile.mmk\"" >> metamake
else
  echo "ERROR: home.mmk not found!"
  echo "(none of ../../home.mmk ../../../home.mmk ../../../../home.mmk exists)"
  echo "The directory in which cook is configured must be the 1st, 2nd"
  echo "or 3rd-level subdirectry of MACSIMUS/cook/."
  exit
fi

echo "!! !dir is relative to !home (see \"home.mmk\")" >> metamake
echo "!dir = gen sim sim/$NONBOND cook cook/$SUBDIR" >> metamake

if [ "$EXENAME" != "" ] ; then
  NAME=$EXENAME
else
  echo "The executable name was determined from the configuration"
fi

echo -n "!"$NAME ": " >> metamake
echo "alloc conjgrad ground rndgen statics varfile bitfile cputime \\" >> metamake
echo "  asksig simcg simgear interpot setss intrapot sitesite norm xforces linregr \\" >> metamake
echo "  constrd rhs simglob pakcp simils siminit main simpot simmeas sfdx \\" >> metamake
echo "  forces simdef \\" >> metamake
echo " "$MODULE >> metamake
echo >> metamake

echo "The executable name is" $PWD/$NAME
echo "You may change it in file" $PWD/metamake
echo
echo "<Enter> to convert metamake to the makefile (linux version), ^C to cancel"

read

if [ -e ../../bin/makemake ]; then
  MM="../../bin/makemake"
else
  MM=makemake
fi

if [ "$PAR" != "" ] ; then
  MMOPT=pthread
fi

echo $MM unix $MAKEMAKE $MMOPT
$MM unix $MAKEMAKE $MMOPT

echo
echo -n "SITE-SITE: "
fgrep "D3 =" makefile

echo "<Enter> to erase all previous *.o modules and compile, ^C to cancel"
read
rm *.o

echo "========================================================================"
make $NAME
echo "recompile by:"
echo "  cd $SUBDIR; make $NAME"
