/* J.Kolafa 1991, update 1995 - see alloc.h for info                alloc.c */
/* changes 11/2009:
   FNLEN increased (from 28 to 40)
   output pointer hex only 
   5/2012: small bug fixed - no trailing 0 in id
*/   

#include "ground.h"
#include "alloc.h"
#include "mystring.c"

#undef free /* original system free must be valid here !!! */

#ifdef DOS
long AllocSizeLim=0xFFFFL;   /* 64KB-1 */
#else
int4 AllocSizeLim=0x40000000; /* changed from 32 to 128 MiB 6/2004, to 1 GiB 6/2012 
                                 NB: to be made 64 bit*/
#endif

#if CHECKHEAP > 2 || CHECKHEAP < -2
#error CHECKHEAP out of range
#endif

#ifdef SHM
/* using shared memory instead the normal heap */ 
#warning "using shared memory instead the normal heap"
#include "shmalloc.c"
#define malloc shmalloc
#define free shmfree
#endif

/***** alloc/free *****/


#if CHECKHEAP == 0 /* ==================================================== 0 */

/***** no check *****/

void *myfree(void *f)
{
#if FREENULL
  if (f)
#endif
    free(f);
  return NULL;
} /* myfree */

#elif CHECKHEAP == 1 || CHECKHEAP == -1 /* ========================== +1,-1 */

/***** heap overflow and bad calls to free are checked *****/

void *mymalloc(int4 size,int zero)
{
  void *a;
  
  if (size>AllocSizeLim) {
    Error("AllocSizeLim exceeded");
    return NULL; }
  if ((a=(void*)malloc(size)) == NULL) {
    Error("malloc failed");
    return NULL; }
  if (zero) memset(a,0,size);
  
  return a;
} /* mymalloc */

void *myfree(void *f)
{
  if (f) { free(f); f = NULL; }
#if !FREENULL
  else Error("free(NULL)");
#endif
  return f;
} /* myfree */


#elif CHECKHEAP == 2 || CHECKHEAP == -2 /* ========================== +2,-2 */

/* IDLEN-1 chars of id's recorded */
#define IDLEN 12
#define IDFMT "%12s"

#ifdef DOS
#define FNLEN 12
#else
#define FNLEN 40
#endif

int4 AllocRange, AllocTrace;

static unsigned char AllocRangeData[16]={
  85,122,49,97,165,202,35,27,109,181,59,65,206,108,162,170};

struct AllocRange_s {
  unsigned char check[16];
  void *p;
  struct AllocRange_s *next;
  int4 size;
  int line;
  char fn[FNLEN];
  char id[1]; /* variable length */
} *Range0;

static int AllocCount=0;

static struct AllocRange_s *staticR;

static void RError(const char *format,...)
{
  va_list args;
  char enough[256];

  sprintf(enough,
    "RANGE CHECK for `%s':\n*** allocated with %ld B in %s:%d\n*** ",
    staticR->id,(long)staticR->size,staticR->fn,staticR->line);

  va_start(args,format);
  vsprintf(enough+strlen(enough),format,args);
  myError(enough);
  va_end(args);
}

static void checkone(struct AllocRange_s *R)
{
  if (!AllocRange) 
    return;
  else {
    unsigned char *head,*tail;
    int4 size;
    unsigned int i;

    staticR=R;

    if (AllocRange<0 || AllocRange&15) {
      RError("AllocRange smashed");
      AllocRange=16; }

    head=(unsigned char*)R->p-(AllocRange+16);
    size=*(int4*)(head+8);
    if (size<1 || size>AllocSizeLim) {
      RError("size overwritten");
      return; }
    if (*(int4*)(head+12)!=AllocRange) {
      RError("write %ld B before",AllocRange+8L);
      *(int4*)(head+12)=AllocRange; }
    head+=16;
    loop (i,0,AllocRange) if (head[i]!=AllocRangeData[i&15]) {
      RError("write %ld B before",(long)AllocRange-i);
      head[i]=AllocRangeData[i&15]; }
    tail=head+size+AllocRange;
    loop (i,0,AllocRange) if (tail[i]!=AllocRangeData[i&15]) {
      RError("write %ld B after",(long)i+1L);
      tail[i]=AllocRangeData[i&15]; } }
}


void mycheckranges(int verbose,char *fn,int line)
{
  struct AllocRange_s *R;
  int n=0;

  myErrorInfo.Level=1; myErrorInfo.Line=line; myErrorInfo.File=fn;

  if (verbose)
    prt("ALLOC and RANGE check called from %s:%d %s",
        fn,line,Range0?"":"- empty!");

  for (R=Range0; R; R=R->next) {
    staticR=R;
    if (verbose)
      prt("%3d " IDFMT " %08p %lu B by %s:%d",
         ++n,R->id,R->p,R->size,R->fn,R->line);
    if (memcmp(R->check,AllocRangeData,16))
      RError("bad control data");
    checkone(R); }
}


void *mymalloc(int4 size,int zero,char *c,char *fn,int line)
/* note: zero&2 marks call from ralloc */
{
  void *a;

  myErrorInfo.Level=1; myErrorInfo.Line=line; myErrorInfo.File=fn;

  if (size<1 || size>AllocSizeLim)
    myError("alloc(%s,%ld): size out of range",c,(long)size);

  if (AllocRange) {
    unsigned char *head,*tail;
    unsigned int i;
    struct AllocRange_s *R;

    /*
    if (AllocRange<0 || AllocRange&15) {
      myError("alloc(%s,%ld): %ld is bad AllocRange",
              c,(long)size,(long)AllocRange);
      AllocRange=16; }
    */
    /* 12/2004: change AllocRange smoothly to a multiple of 16: */
    AllocRange=(AllocRange+15)/16*16;

    mycheckranges(0,fn,line);

    head=(unsigned char*)malloc(size+2*AllocRange+16);
    if (!head)
      myError("alloc(%s,%ld): no heap",c,(long)size);
    *(int4*)(head+8)=size;
    *(int4*)(head+12)=AllocRange;
    head+=16;
    tail=head+size+AllocRange;
    loop (i,0,AllocRange) head[i]=tail[i]=AllocRangeData[i&15];
    a=(void*)(head+AllocRange);
    R=malloc(sizeof(struct AllocRange_s)+strlen(c));
    if (!R)
      myError("alloc(%s,%ld): no heap (check)",c,(long)size);
    else {
      R->next=Range0;
      R->p=a;
      R->size=size;
      copy(R->check,AllocRangeData,16);
      strcpy(R->id,c);
      /* copy only last FNLEN-1 chars of fn if longer than FNLEN */
      strcpy(R->fn,fn+max(strlen(fn)-(FNLEN-1),0));
      R->line=line;
      Range0=R; }
    }
  else
    if ((a=(void*)malloc(size)) == NULL)
       myError("alloc(%s,%ld): no heap",c,(long)size);

  if (zero&1) memset(a,0,size);

  if (AllocTrace)
    prt("%3d:" IDFMT " %08p %calloc %luB%s in %s:%d",
         ++AllocCount,
         c,
         a,
         zero&2?'r':' ',
         (long)size,
         zero?":=0":"",
         fn,line);
  return a;
} /* mymalloc */

char *mystrdup(const char *s,char *c,char *fn,int line)
{
  char *a=mymalloc(strlen(s)+1,0,c,fn,line);
  strcpy(a,s);
  return a;
}

void *myfree(void *f,char *c,char *fn,int line)
{
  myErrorInfo.Level=1; myErrorInfo.Line=line; myErrorInfo.File=fn;

  if (f == NULL) {
#if FREENULL
    prt("free(%s==NULL) in %s:%d",c,fn,line);
#else
    myError("free(%s==NULL)",c);
#endif
    return NULL; }

  if (AllocRange) {
    struct AllocRange_s *R,**C=NULL;

#if 0
    /* check of the memory beeing freed first */
    staticR=malloc(sizeof(struct AllocRange_s)+strlen(c));
    if (!staticR) myError("no heap for control data");
    strcpy(staticR->id,c);
    staticR->p=*f;
    /* statement that has allocated it is not known */
    strcpy(staticR->fn,"?");
    staticR->line=-1;
    checkone(staticR);
    free(staticR);
#endif
  
    mycheckranges(0,fn,line);

    for (R=Range0; R; R=R->next)
      if (f==R->p) {
        if (C) *C=R->next;
        else Range0=R->next;
        if (AllocTrace)
	  prt("%3i:" IDFMT " %08p as %s in %s:%d freed",
	      AllocCount--,c,f,
	      R->id,R->fn,R->line); 
        free(R);
        break; }
      else
        C=&R->next;
    free((char*)f-AllocRange-16); }
  else
    free(f);

  if (AllocTrace & !AllocRange)
    prt("%3i:" IDFMT " %08p freed",
         AllocCount--,c,f);

  return NULL;
}

#endif /* +2,-2 */


/***** ralloc/release *****/


#if CHECKHEAP < 0

#ifndef RHEAPTABLEN
#define RHEAPTABLEN 32
#endif

typedef struct {
  void *p;
#if CHECKHEAP == -2
  char id[IDLEN];
#endif
} rheap_item;

typedef struct rheap_s {
  rheap_item i[RHEAPTABLEN];
  int pos;
  struct rheap_s *next;
} rheap_t;

rheap_t *rheap;

#if CHECKHEAP == -1

void *myralloc(int4 size,int zero)
{
  rheap_t *h;
  void *a;
  
  if (rheap==NULL || rheap->pos>=RHEAPTABLEN) {
    h=rheap;
    alloc(rheap,sizeof(rheap_t));
    rheap->next=h;
    rheap->pos=0; }

  if ((a=(void*)malloc(size)) == NULL) Error("ralloc");
  if (zero) memset(a,0,size);

  rheap->i[rheap->pos++].p=a;
  return a;
} /* myralloc */

void *myrelease(void *p)
{
  void *x;
  
  rheap_t *h;
  while (rheap!=NULL) {
    x=rheap->i[--rheap->pos].p;
    rheap->i[rheap->pos].p=myfree(rheap->i[rheap->pos].p);
    if (rheap->pos <= 0) {
      h=rheap; rheap=rheap->next;
      h=TYPEOF(h) myfree(h); }
    if (x==p) return NULL; }
  
  Error("release");
  return NULL;
}

#else /* CHECKHEAP = -2 */

void *myralloc(int4 size,int zero,char *c,char *fn,int line)
{
  rheap_t *h;
  void *a;

  myErrorInfo.Level=1; myErrorInfo.Line=line; myErrorInfo.File=fn;

  if (rheap==NULL || rheap->pos>=RHEAPTABLEN) {
    h=rheap;
    alloc(rheap,sizeof(rheap_t));
    rheap->next=h;
    rheap->pos=0; }

  a=mymalloc(size,zero|2,c,fn,line);

  if (AllocTrace) { 
    copy(rheap->i[rheap->pos].id,c,IDLEN);
    rheap->i[rheap->pos].id[IDLEN-1]=0; }

  rheap->i[rheap->pos++].p=a;

  return a;
} /* myralloc */


void *myrelease(void *p,char *c,char *fn,int line)
{
  void *x;
  rheap_t *h;

  myErrorInfo.Level=1; myErrorInfo.Line=line; myErrorInfo.File=fn;

  while (rheap!=NULL) {
    x=rheap->i[--rheap->pos].p;

    if (AllocTrace)
      rheap->i[rheap->pos].p=myfree(rheap->i[rheap->pos].p,rheap->i[rheap->pos].id,fn,line);
    else
      rheap->i[rheap->pos].p=myfree(rheap->i[rheap->pos].p,c,fn,line);

    if (rheap->pos <= 0) {
      h=rheap; rheap=rheap->next;
      h=myfree(h,"rheap",fn,line); }
    if (x==p) return NULL; }

  myError("release %s",c);
  return NULL;
}

#endif

#endif /* <0 */
