/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%        one step t --> t+h of the Gear Predictor-Corrector method           %
%                 for first-order differential equations                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% da/dt=b(t,a) is solved, where a,b are vectors of neq real elements and
% the right hand side b(t,a) is computed by procedure rhs(b,a)
%
% Globals required:
%   real t=time, h=time step
%   int GearOrder = order in {2,3,4,5,6}
% typedef required:
%   ToIntPtr=pointer to anything containing a contiguous vector of neq reals
% Macro or function giving pointer to the first real of the vector must be 
% #defined:  D(a)
% type real must be float or double
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(normally include file for peogear.c, declarations in peogear.h)
*/

#define MaxGearOrder 6

real GearCoef[MaxGearOrder-1][MaxGearOrder] = {
/*2*/ {0.5,      1.0},
/*3*/ {5.0/12,   1.0, 0.5},
/*4*/ {3.0/8,    1.0, 0.75,    1.0/6},
/*5*/ {251.0/720,1.0, 11.0/12, 1.0/3,   1.0/24},
/*6*/ {95.0/288, 1.0, 25.0/24, 35.0/72, 5.0/48, 1.0/120} };

real C0=-9e9; /* to replace the coefficient C_0 */

void Gear(int neq,ToIntPtr *a)
{
int i,j,k,bin;
real s,d,*gc,*r,*hdr,*pred,*rhspred;

if (GearOrder<2 || GearOrder>MaxGearOrder) Error("bad GearOrder");
gc=GearCoef[GearOrder-2];
if (C0>-8e9) gc[0]=C0;

r=D(a[0]);
hdr=D(a[1]);
pred=D(a[GearOrder]);
rhspred=D(a[GearOrder+1]);

switch (GearOrder) {

case 4: /* optimized code (for non-vector computers) */
{
real *a0,*a1,*a2,*a3,*p,*pm,*rp;
real gc0=gc[0],gc2=gc[2],gc3=gc[3];
/* predictor */
for (a0=r,a1=hdr,a2=D(a[2]),a3=D(a[3]),p=pred,pm=pred+neq;
     p<pm; a0++,a1++,a2++,a3++,p++)
  *p = *a0+*a1+*a2+*a3;

/* r.h.s. */
t += h;
rhs(a[GearOrder+1],a[GearOrder]);

/* corrector */
for (a0=r,a1=hdr,a2=D(a[2]),a3=D(a[3]),p=pred,pm=pred+neq,rp=rhspred;
     p<pm; a0++,a1++,a2++,a3++,p++,rp++) {
  s = h* *rp;
  d = s - *a1;
  *a1 = s;
  d -= 2* *a2 + (s=3* *a3);
  *a0 = *p + d*gc0;
  *a2 += s + d*gc2;
  *a3 += d*gc3; }
}
break;

default: /* general case - slow code !!! */
/* predictor */
for (i=0; i<neq; i++) {
  s = 0;
  for (j=0; j<GearOrder; j++) s += (D(a[j]))[i];
  pred[i] = s; }

/* r.h.s. */
t += h;
rhs(a[GearOrder+1],a[GearOrder]);

/* corrector */
for (i=0; i<neq; i++) {
  s = h*rhspred[i];
  d = s-hdr[i];
  hdr[i] = s;
  for (j=2; j<GearOrder; j++) d -= j*(D(a[j]))[i];
  r[i] = pred[i] + d*gc[0];
  for (j=2; j<GearOrder; j++) {
    s = d*gc[j]; bin = 1;
    for (k=j; k<GearOrder; k++) {
      s += (D(a[k]))[i]*bin;
      bin = bin*(k+1)/(k+1-j); }
    (D(a[j]))[i] = s; }
  }

} /* switch */
}
