/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%        one step t --> t+h of the Gear Predictor-Corrector method           %
%                 for second order differential equations                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  2    2
% d a/dt  = b(t,a,da/dt) (#ifdef VELOCITY)
%   or
%  2    2
% d a/dt  = b(t,a) (#ifndef VELOCITY)
%   is solved, where a,b are vectors of neq real elements and the right
%   hand side.  b(t,a[,da/dt]) is computed by procedure rhs(b,a[,da/dt])
%
% Globals required:
%   real t=time, h=time step
%   int GearOrder = order in {3,4,5,6}
% typedef required:
%   ToIntPtr=pointer to anything containing a contiguous vector of neq reals
% Macro or function giving pointer to the first real of the vector:
%   D(a)
% if VELOCITY is #defined, rhs is assumed to depend on da/dt
% the contents of  a  [and  da ] should not be changed, small corrections
%   of the order of the method (e.g. to correct constraints) are allowed
%
% typedef real must be float or double
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(normally include file for peogear.c, declarations in peogear.h) 
*/

#include "simgear.h" /* added 11/2001 */

#ifdef PAR
#include "cputime.h"
#endif

#define MaxGearOrder 6

real GearCoef[MaxGearOrder-2][MaxGearOrder] = {
/*3*/ {0.0,      1.0,      1.0},
/*4*/ {1.0/6,    5.0/6,    1.0, 1.0/3},
#ifdef VELOCITY
/*5*/ {19.0/90,  0.75,     1.0, 0.5,     1.0/12},
/*6*/ {3.0/16,   251.0/360,1.0, 11.0/18, 1.0/6,  1.0/60},
#else
/*5*/ {19.0/120, 0.75,     1.0, 0.5,     1.0/12},
/*6*/ {3.0/20,   251.0/360,1.0, 11.0/18, 1.0/6,  1.0/60},
#endif
  };

real C0=-9e9; /* to replace the coefficient C_0 */

void Gear(int neq,ToIntPtr *a)
{
int i,j,k,bin;
real hh2=h*h/2,s,d,ss,*gc,*r,*hdr,*pred,*rhspred;
#ifdef VELOCITY
real *preddr;
#endif

if (GearOrder<3 || GearOrder>MaxGearOrder)
  extError("bad GearOrder",__FILE__,__LINE__);
gc=GearCoef[GearOrder-3];
if (C0>-8e9) gc[0]=C0;

r=D(a[0]);
hdr=D(a[2]);
pred=D(a[GearOrder]);
#ifdef VELOCITY
preddr=D(a[GearOrder+2]);
#endif
rhspred=D(a[GearOrder+1]);

switch (GearOrder) {

case 4: /* optimized code */

/* predictor */
loop (i,0,neq) {
  pred[i] = (D(a[3]))[i] + (D(a[2]))[i] + (D(a[1]))[i] + (D(a[0]))[i];
#ifdef VELOCITY
  preddr[i] = (3*(D(a[3]))[i] + 2*(D(a[2]))[i] + (D(a[1]))[i])/h;
#endif
  }

/* r.h.s. */
t += h;

#ifdef PAR
                                                             CPUtime("gear");
#endif

#ifdef VELOCITY
rhs(a[5],a[4],a[6]);
#else
rhs(a[5],a[4]);
#endif

/* corrector */
loop (i,0,neq) {
  ss = hh2*rhspred[i];
  d = ss-hdr[i]-3*(D(a[3]))[i];
  r[i] = pred[i] + d*gc[0];
#ifdef VELOCITY
  (D(a[1]))[i] = h*preddr[i] + d*gc[1];
#else
  (D(a[1]))[i] = d*gc[1] + 3*(D(a[3]))[i] + 2*(D(a[2]))[i] + (D(a[1]))[i];
#endif
  (D(a[3]))[i] = d*gc[3] + (D(a[3]))[i];
  hdr[i] = ss; }

break;

default: /* general case - slow code !!! */

/* predictor */
loop (i,0,neq) {
  s = 0;
  for (j=0; j<GearOrder; j++) s += (D(a[j]))[i];
  pred[i] = s;
#ifdef VELOCITY
  s = 0;
  loop (j,1,GearOrder) s += j*(D(a[j]))[i];
  preddr[i] = s/h;
#endif
  }

/* r.h.s. */
t += h;

#ifdef VELOCITY
rhs(a[GearOrder+1],a[GearOrder],a[GearOrder+2]);
#else
rhs(a[GearOrder+1],a[GearOrder]);
#endif

/* corrector */
loop (i,0,neq) {
  ss = hh2*rhspred[i];
  d = ss-hdr[i];
  loop (j,3,GearOrder) d -= j*(j-1)/2*(D(a[j]))[i];
  r[i] = pred[i] + d*gc[0];
#ifdef VELOCITY
  (D(a[1]))[i] = h*preddr[i] + d*gc[1];
#else
  s = d*gc[1];
  loop (j,1,GearOrder) s += j*(D(a[j]))[i];
  (D(a[1]))[i]=s;
#endif
  loop (j,3,GearOrder) {
    s = d*gc[j]; bin = 1;
    loop (k,j,GearOrder) {
      s += (D(a[k]))[i]*bin;
      bin = bin*(k+1)/(k+1-j); }
    (D(a[j]))[i] = s; }
  hdr[i] = ss; }

} /* switch */
#ifdef PAR
                                                             CPUtime("gear");
#endif

}
