/*
  2/10: PAR removed, EOF on in just message (no error)
  4/06: error output to out suppressed if in=NULL
  7/05: stdarg incompatibility fixed in myError
  10/01: minimum buffer for initscroll is 1024B
  6/00: bug fixed in commands $L $C
  11/99: initscroll(negative) causes ERROR not to ask but stop

  This module depends on the following #defines (which should be
    specified as -D options to ALL modules using ground.h):
  -DSCR   : scrolling of previous output available (default=no scrolling);
            also $iFILE on input requires -DSCR
  -DCALC=0  only numbers available in get data statements
  -DCALC=1  number expressions and elementary functions available
  -DCALC=2  also += -= *= /= %= ^= statements, # as result of previous
            calculations, and inspecting values of identifiers
  -DCALC=3  as above + identifiers in expressions allowed
  -DCALC=4  as above + extra functions (gamma, erfc, hyperbolic)
  (undefined CALC is the same as CALC=4)
  -DDOS  : DOS Turbo/Borland C (16 bit) versions

  further #define variables with changable defaults are:
    LINES COLUMNS
    STRLEN GETBUFLEN
*/

#include "ground.h"
#ifdef DOS
#include <conio.h>
#endif

#define REAL double
#define FN(X) X
#include "func.c"
#undef REAL
#undef FN

#if PRECISION==2
/* see ground.h */
#define REAL long double
#define CAT(X,Y) X##Y
#define FN(X) CAT(X,l)
#include "func.c"
#undef FN
#undef CAT
#undef REAL
#endif

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                                  in and out
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

formatted i/o functions (redirectable by scroll)
~~~~~~~~~~~~~~~~~~~~~~~
prt(format,...) = fprintf(out,format,...) ended by line feed
prt_(format,...) = fprintf(out,format,...);
prts_(char *s) = fputs(s,out);  (without LF)
prts(char *s) = fputs(s,out); + LF
prtc(c) = fputc(c,out)
***/

FILE *in,*out; /*** must be initialized when program is started
                - static initializing by stdin or stdout is not portable !!! */
static FILE *masterin; /* to solve $iFILE scroll statement */

char *_S_buf=NULL; /* buffer for scroll */

#ifdef SCR

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% scroll %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(c) J.Kolafa 10/1991

This unit defines prt, prt_, and prtc so that (since initscroll(int buflen)
has been called) all data printed to out are stored in a buffer.
Any call to scroll() stops the program execution and allows looking to the
previous output up to the buffer capacity.

See the help below for available commands!

NOTES/CAVEATS/BUGS:
  assumes screen 24 lines long; this can be changed by command L or $LINES
  assumes screen 80 chars wide; this can be changed by command C or $COLUMNS
  lines longer than 80 chars are truncated and no LF is sent to display
  all control characters < Esc are interpreted as LF (tabs included!)
  does not work if data contain null characters
  may crash in extreme conditions (particularly if data of single fprintf
    call are longer than half the buffer size)
  if there is not enough memory, initscroll returns -1 and scroll simply
    does not scroll
  <stdarg.h> and ANSI standard required
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

static struct {
  int L;      /* lines/screen (incl. prompt line) */
  int C;      /* no of columns */
  int ibuf;   /* _S_buf pointer */
  int buflen; /* length of _S_buf */
  int clear;  /* DOS: NONE=clrscr() */
  int stoponerr; /* ERROR(()), WARNING(()) don't, always stop */
#ifdef CLOSEGRAPH
  int closegraph; /* whether to call closegraph() after quit */
#endif
} S={25,80,0,0,
#ifdef DOS
               0,
#else
               1,
#endif
                 0};

#define NONE 0
#define ANSI 1
#define LINEFEEDS 2
#define SPACES 3

static void _S_getscreen(int *size,char *c)
{
  if (c) *size=atoi(++c);
}

void setscroll(int lines,int columns,int clear,int closegraph)
{
  if (lines) S.L=lines;
  if (columns) S.C=columns;
  S.clear=clear;
#ifdef CLOSEGRAPH
  S.closegraph=closegraph;
#endif
}

int initscroll(int buflen) /************************************* initscroll */
{
  char *LINES=getenv("LINES");
  char *COLUMNS=getenv("COLUMNS");

  if (in==NULL) in=stdin;
  if (out==NULL) out=stdout;

  if (buflen<=1023) {
    S.stoponerr=buflen;
    return 0; }

  if (LINES) S.L=atoi(LINES);
  if (COLUMNS) S.C=atoi(COLUMNS);

  buflen -= buflen & 1; /* must be even !!! */
  if (_S_buf!=NULL) free(_S_buf);
  if ( (_S_buf=(char*)malloc(buflen))==NULL) return -1;
  memset(_S_buf,0,buflen);
  S.buflen=buflen;

  _S_buf[0]='\n';
  S.ibuf=1; /* to point after \n */
  _Id.echo=(in==stdin && out==stdout)+1;

  return 0;
}

static void _S_swapbuf(void) /*********************************** _S_swapbuf */
{
  int i,bh=S.buflen/2;
  char *b,c;

  if (S.ibuf>bh) {
    b=_S_buf+bh;
    loop (i,0,bh) { c=_S_buf[i]; _S_buf[i]=b[i]; b[i]=c; }
    S.ibuf-=bh; }
}

int prtc(char c) /***************************************************** prtc */
{
  int i;

  i=putc(c,out);
  if (out==stdout) if (_S_buf) {
    _S_buf[S.ibuf++]=c; _S_buf[S.ibuf]=0;
    _S_swapbuf(); }

  return i;
}

int prts_(const char *c) /********************************************* prts */
{
  int i;

  i=fputs(c,out);
  if (out==stdout) if (_S_buf) {
    while (*c) _S_buf[S.ibuf++]=*c++;
    _S_buf[S.ibuf]=0;
    _S_swapbuf(); }

  return i;
}

static int vprt_(const char *format,va_list args) /******************* vprt_ */
{
  int i;
  char *s;

  if (out==stdout && _S_buf) {
    i=vsprintf(s=_S_buf+S.ibuf,format,args);
    if (fputs(s,out)==EOF) i=EOF;
    else { i=0; while (*s) s++,i++; /* to return # of chars */ }
    while (_S_buf[S.ibuf]) S.ibuf++;
    _S_swapbuf(); }
  else
    i=vfprintf(out,format,args);

  return i;
}


void scroll(void) /*************************************************** scroll */
{
#ifndef STRLEN
#ifdef DOS
#define STRLEN 128
#else
#define STRLEN 256
#endif
#endif

  char x[STRLEN], *xnl;

#define B(I) _S_buf[I+S.buflen*(I<0)]

  if (_S_buf) {
    static char s[STRLEN];
    static int l;
    FILE *f=NULL;
    int pos=S.ibuf-1,i,nl=0,key=0;
   
    if (S.buflen) if (_S_buf[pos]>=27) prtc('\n');
    for (;;) {
    again:
      /*.....    fputs(key ? key>0 ? "bot$\n" : "top$\n" : "$\n",stdout);*/
      fputs(key ? key>0 ? "bot$" : "top$" : "$",stdout);
      fgets(x,STRLEN,stdin);
      if ( (xnl=strchr(x,'\n')) ) *xnl=0; /* remove possible LF */

      switch (*x) {
        case 'i': masterin=in;
          if ( !(in=fopen(x+1,"rt")) ) {
            if (_Id.echo>=0) 
              ERROR(("$i%s: no include file",x+1))
              in=masterin; }
        case '.':
        case 'q': return;
        case  0 : nl=1; break;
        case 'u': nl= -(S.L-2); break;
        case 'd': case ' ': nl=S.L-2; break;
        case '/': xnl=x+1; nl=strlen(xnl);
                  if (nl) { strcpy(s,xnl); l=nl; }
                  i=pos;
                  /* screen-line up to start search from 2nd line */
                  nl= -(S.L-2);
                  while (nl<0) {
                    while (i--,B(i)>=27);
                    if (!B(i)) { i++; break; }
                    nl++; }
                  for ( ; B(i); i++)
                    if (!memcmp(s,&B(i),l)) { pos=i; nl=S.L-1; break; }
                  break;
        case 'w': if (x[1]<=' ') { nl=0; break; }
                  f=fopen(x+1,"wt");
        case 't': case 'g': nl= -32000; break;
        case 'L': _S_getscreen(&S.L,x); Min(S.L,200) Max(S.L,4) break;
        case 'C': _S_getscreen(&S.C,x); Min(S.C,200) Max(S.C,10) break;
        case '!': _S_getscreen(&S.clear,x); break;
        case 'e':
#ifndef DOS
                  fputc('\n',out);
#endif
                  if (!strcmp(x,"exit"))
#ifdef CLOSEGRAPH
                  if (S.closegraph) closegraph();
#endif
                  exit(1);
        case 'b': case 'G': nl=32000; break;
        case '?': case 'h': fprintf(stderr,"\n\
%i lines x %i cols  clrscr=%d\n\n\
$t $g    top\n\
$b $G    bottom\n\
$<Enter> line down\n\
$-       line up\n\
$u       page up\n\
$d $<Spc>page down\n\
$#       by # lines\n\
$L#      set lines\n\
$C#      set cols\n\
$!#      set clrscr mode\n\
$wFILE   write\n\
$iFILE   include\n\
$/STRING find\n\
$/       last find\n\
$q $.    quit scroll\n\
$exit    exit prog\n\n\
",S.L,S.C,S.clear);
                  goto again;
        case '-': if (x[1]==0) { nl=-1; break; }
        default: nl=atoi(x); }

      key=0;
      /* by nl lines forward */
      while (nl>0) {
        while (pos++,B(pos)>=27);
        if (!B(pos)) { pos--; nl=0; key++; break; }
        nl--; }

      /* subtract screenlen to move cursor 1 screen up */
      nl-=(S.L-1);

      /* by nl lines backward */
      while (nl<0) {
        while (pos--,B(pos)>=27);
        if (!B(pos)) { pos++; key--; break; }
        nl++; }

      switch (S.clear) {
        case NONE:
#ifdef DOS
          clrscr();
#endif
          break;
        case ANSI:
          printf("\n\33[2J\n");
          break;
        case SPACES:
          loop (i,0,S.C*S.L) putc(' ',stdout);
          putc('\n',stdout);
          break;
        case LINEFEEDS:
          loop (i,0,S.L) putc('\n',stdout); }

      if (f) { /* to file */
        i=pos;
        while (B(i)) {
          putc(B(i),f);
          i++; }
        fclose(f); f=NULL; }

      nl=S.L-1;

      while (nl>0) {
        i=0;
        while (pos++,B(pos)>=27) if (i++<S.C) putc(B(pos),stdout);
        if (i<S.C) putc('\n',stdout);
        if (!B(pos)) { pos--; break; }
        nl--; }
      } /* for (;;) */
    }
  else {
    /* batch mode: only $i solved here */
    fgets(x,STRLEN,in);
    if ( (xnl=strchr(x,'\n')) ) *xnl=0; /* remove possible LF */
    if (*x=='i') {
      masterin=in;
      if ( !(in=fopen(x+1,"rt")) ) {
        if (_Id.echo>=0) ERROR(("$i%s: no include file",x+1))
        in=masterin; } }
    }
#undef B
#undef STRLEN
} /*scroll*/

#else
/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% no scroll %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/
int initscroll(int buflen) /************************************* initscroll */
{
  if (in==NULL) in=stdin;
  if (out==NULL) out=stdout;
  if (buflen) _S_buf=(char*)1; /* to mark only */
  _Id.echo=(in==stdin && out==stdout)+1;

  return 0;
}

int prtc(char c) /***************************************************** prtc */
{
  return putc(c,out);
}

int prts_(const char *c) /******************************************** prts_ */
{
  return fputs(c,out);
}

static int vprt_(const char *format,va_list args) /******************* vprt_ */
{
  return vfprintf(out,format,args);
}

void scroll(void) /************************************************** scroll */
{
#ifdef DOS
  char x[32];
#else
  char x[256];
#endif
  if (_S_buf) fgets(x,8,in);
}

#endif /* SCR */

/*** common for SCR and !SCR ***/

int prts(const char *c) /********************************************** prts */
{
  int i=prts_(c);

  if (i>=0) return prtc('\n');
  else return i;
}

static int vprt(const char *format,va_list args) /********************* vprt */
{
  int i,j;

  i=vprt_(format,args);
  j=prtc('\n');

  return j<0 ? j : i;
}

int prt_(const char *format,...) /************************************* prt_ */
{
  int ret;
  va_list args;

  va_start(args,format);
  ret=vprt_(format,args);
  va_end(args);

  return ret;
}

int prt(const char *format,...) /************************************** prt */
{
  int ret;
  va_list args;

  va_start(args,format);
  ret=vprt(format,args);
  va_end(args);

  return ret;
}



/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                                    Error
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

struct myErrorInfo_s myErrorInfo;

void myError(const char *format, ...) /*************************** myError */
{
  va_list args;
  const char *errfmt="\n*** %s in %s:%d\n*** ";
  int ierr=20;

  if (myErrorInfo.Level<0 || myErrorInfo.Level>2) {
    fprintf(stderr,"\n?! TOTAL BREAKDOWN: unknown errorlevel");
    myErrorInfo.Level=2; }


if (out!=stdout 
#ifdef SCR
                || S.stoponerr
#endif
                              ) {
  fprintf(stderr,errfmt,
          "WARNING\0ERROR\0  DISASTER"+myErrorInfo.Level*8,
          myErrorInfo.File,myErrorInfo.Line);
  va_start(args,format);
  vfprintf(stderr,format,args);
  va_end(args);
  fputc('\n',stderr);
  }

 prt_(errfmt,
      "WARNING\0ERROR\0  DISASTER"+myErrorInfo.Level*8,
      myErrorInfo.File,myErrorInfo.Line);
  va_start(args,format);
  vprt(format,args);
  va_end(args);
  _n

  if (
#ifdef SCR
      !S.stoponerr && 
#endif
                      in==stdin && out==stdout) {
    char s[32];

  prompt:
#ifdef SCR
    fprintf(stderr,"*** select:  (a)bort  (e)xit%s%s\n",
      myErrorInfo.Level==0?"  (c)ontinue":myErrorInfo.Level==1?"  (i)gnore":"",
      S.buflen>1?"  (s)croll":"");
#else
    fprintf(stderr,"*** select:  (a)bort  (e)xit  %s\n",
      myErrorInfo.Level==0?"  (c)ontinue":myErrorInfo.Level==1?"  (i)gnore":"");
#endif
    if (fgets(s,32,in)==NULL) { prt("*** EOF ***"); s[0]='e'; }
    if (!ierr--) s[0]='e';
    switch (s[0]) {
      case 'a': fclose(out);
#ifdef CLOSEGRAPH
                closegraph();
#endif
                abort();
      case 'e': fclose(out);
#ifdef CLOSEGRAPH
                closegraph();
#endif
                exit(myErrorInfo.Level);
      case 'c': if (myErrorInfo.Level) goto prompt;
                prt("*** continuing ***"); return;
      case 'i': if (myErrorInfo.Level==2) goto prompt;
                prt("*** ignoring error ***"); return;
#ifdef SCR
      case 's': case '$': scroll();
#endif
      default : goto prompt; }
    }
  else if (myErrorInfo.Level){
    fclose(out);
    exit(myErrorInfo.Level);  }
}

/* old version, kept for compatibility reasons (and PAR): */

void extError(const char *msg, const char *f, int l) /*********** extError */
{
  static char e[]="*** %s:%i ERROR %s ***\n";

  fprintf(stderr,e,f,l,msg);

  if (in) { /* added 4/2006 */
    prtc('\n');
    prt_(e,f,l,msg); }
  if (in==stdin && out==stdout) {
    char s[32];
  prompt:
#ifdef SCR
    fprintf(stderr,"select:  (a)bort  (e)xit  (r)esume  (s)croll\n");
#else
    fprintf(stderr,"select:  (a)bort  (e)xit  (r)esume\n");
#endif
    if (fgets(s,32,in)==NULL) {
      prt("*** EOF ***");
      s[0]='e'; }
    switch (*s) {
      case 'a': fclose(out);
#ifdef CLOSEGRAPH
                closegraph();
#endif
                abort();
      case 'e': fclose(out);
#ifdef CLOSEGRAPH
                closegraph();
#endif
                exit(msg[0]+(msg[0]==0));
      case 'r': prt("*** execution resumed ***"); 
                return;
#ifdef SCR
      case 's': case '$': scroll();
#endif
      default : goto prompt; } }
  else {
    fclose(out);
#ifdef CLOSEGRAPH
    closegraph();
#endif
    exit(msg[0]+(msg[0]==0)); }
}


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           C-version of the PL/I GET DATA and PUT DATA statements            %
%                            (c) J.Kolafa 1991                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
* input stream should contain a list of assignments of the form
    identifier=expression
  or
    identifier[decimal_const]suffix=expression
  ended by a semicolon (suffix may be empty)
* identifiers must begin with 'A'..'Z', 'a'..'z' ( not '_' )
* input is read from file `in'
* output is written to file `out'
*   `out' and `in' must be defined by the user or by initscroll()
* use in the source must start by `getdata' and end by `enddata'
* macro get(S) reads scalar S ( statement in input stream is S=expression )
* getvec(V,S,maxindex) reads one item from a vector of structures
  ( statement in input stream is V[index]S=expression )
  ( use getvec(V,,maxindex) for a simple vector, V[index]=expression )
* checkdata checks for misspelled identifiers, to be used just before enddata
* #define CALC 0  only numbers allowed in getdata
               1: expressions allowed in getdata (calc.c needed)
               2: also old value available in expr (as #)
               3: also variables in expression available (only scalars)
               4: + extra functions
* statement `?' single on line prints a brief help
* `?=' toggles echo mode (auto print of expressions results)
* `??' prints a list of all identifiers
* #ifdef SCR then using $ in the input stream enables scrolling
  $ must be followed by a scroll command (single letter or number)
* any text following '!' until end of line is discarded (treated as a comment)

* example of use (with CALC==3)

  #include "ground.h"

  double dd=7.7;
  double z[3]={1.1,2.2,3.3};
  double abc[2];

  int main()
  {
    int i=11, J3=33;
    double xx;
    int xxset=0;
    struct { int i; double d; } s[10];
    struct keys_s initkeys[]={
      {"cont",0},  {"continue",0},
      {"append",1},
      {"start",2},
      {"random",3},
      {NULL,0}};

    initscroll(0); // selects in=stdin, out=stdout, echo on, no scrolling

    getdata
      get(i) get(J3) get(abc[0]) get(abc[1])
      getvec(z,,3) getvec(s,.i,10) getvec(s,.d,10)
      getmark(dd,ddset)
    checkdata enddata
    if (!ddset) ERROR(("dd has not appeared in input data"))

    put(dd)
    put2(i,sqrt((double)J3))
    put3(z[0]*2,z[1]*3,z[2]*4)
    put2(abc[0],abc[1])
    put2(s[3].i,s[3].d)
  
    return 0;
  }

* example of input data

  dd=acos(-1)^2 J3=123.4^2
  i=2
  i = #*3*4*5*6
  z[1] = pi/3
  s[3].i=123 s[3].d=(1+1/(2+3))^.5
  abc[1]=1;

* will generate the following output

i=2
  i = 2
i = #*3*4*5*6
  i = 720
z[1] = pi/3
  z[1] = 1.04719755
s[3].i=123 s[3].d=(1+1/(2+3))^.5
  s[3].i = 123
  s[3].d = 1.09544512
abc[1]=1;
  abc[1] = 1

         dd=7.7
          i=720           sqrt((double)J3)=5.74456
     z[0]*2=2.2                z[1]*3=3.14159            z[2]*4=13.2
     abc[0]=0                  abc[1]=1
     s[3].i=123                s[3].d=1.09545

* NOTES/CAVEATS/BUGS:
- Indexed variables cannot be used in expressions; however, the following 
  workaround may be used:
    L[1]
    x=# ! now x=L[1]  
- One assignment statement in the input stream must be within one line of
  maximum length 127 (DOS) or 255 (may be changed by #define GETBUFLEN)
- In the case of more dimensions the varying index must be in the first []
- Error messages appear after the whole line is read and (if _Id.echo==1)
  copied to out - the position of the error is not marked
- An expression is evaluated in double, the result is undefined if it cannot
  be cast into the type requested
- If the variable is int, result printed as float
- pi, PI, and standard function names cannot be used as identifiers
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#if CALC
#include "calc.c"
#endif

const char *_putformat="%11s=%-13.6g ";

#ifndef GETBUFLEN
#ifdef DOS
#define GETBUFLEN 128
#else
#define GETBUFLEN 256
#endif
#endif

/* this structure must be global - used in getdata macros */
#if CALC>=3
struct _Id_s _Id = { " = %-16.9g", "", 1.0, 0};
#else
struct _Id_s _Id = { " = %-16.9g" };
#endif

static void prtecho(void)
{
  prt("<echo is %s>",_Id.echo?_Id.echo==1?"copy input":"on":"off");
}

int _GetId(void) /*************************************************** _GetId */
/*
  gets one statement from `in'
  reads a whole line (until `;') if necessary
  scans lot of statements
  returns 1 if get() statements are to be scanned, otherwise 0
*/
{
  int c,sg,excl=0;
  char *x,*end;
  static char zero=0;

#if CALC>=2
  double lhs=_Id.nr;
#endif

  _Id.str=NULL;

#if CALC>2
  /* 1st pass just to scan all scalar variables to create a list with values */
  if (_Id.list) return 1;
#endif

#if CALC>=2
  /* += -= /= *= ^= %= pending */
  _Id.used=0;
  if (_Id.asg>1) goto evalexpr;
#endif

  /* reads one line (until `;') if necessary */
  if (*_Id.b==0) {
    x=_Id.buf;
    do {
      c=getc(in);
      if (c<0) {
        /* EOF found: swap back to master `in' if this is EOF on file
           included by $iFILE (this is handled by scroll() */
        fclose(in);
        in=masterin; masterin=NULL;
        if (in==NULL) {
          fprintf(stderr,"EOF on in - exit\n");
          if (out) prt("EOF on in - exit");
          exit(0); }
        c='\n'; }
#ifdef SCR
      while (c=='$') {
        scroll();
        c=getc(in); }
#endif
      if (_Id.echo==1) if (c>0) prtc(c);
      if (c<' ') { if (c=='\n') goto eoln; c=' '; }
      if (c=='!') excl=1; /* comment: skip to eoln */
      if (!excl) *x++=c;
      if (x-_Id.buf>=GETBUFLEN-1) {
        *(_Id.b=_Id.buf)=0;
        Error("data:line too long");
        return 0; }
      } while (c!=';' || excl); /* `;' after `!' ignored - not end getdata set */
    eoln:
    *x++=0;
    *x=0;
    _Id.b=_Id.buf; }

/* remove trailing spaces */
  while (*_Id.b==' ') _Id.b++;

  _Id.prt=_Id.asg=0;
  _Id.id=_Id.b;
  _Id.suff= &zero;
  _Id.br=NULL;

#include "question.c"

  /* empty line or `;' */
  if ((_Id.empty = (*_Id.id==';' || *_Id.id==0))) {
    noscan:
    _Id.id= &zero;
    _Id.used=1;    /* suppress check `identifier found' */
    goto ret; }

  if (_Id.echo==2) _Id.prt=1; /* ID name to be printed */


  /*** identifier scan, incl. # and indexed id's ***/
  
  if (isalpha(*_Id.b))
    while (*_Id.b && (isalnum(*_Id.b) || strchr("_.[",*_Id.b))) {
      if (!_Id.br && *_Id.b=='[') {
        /* scanning [INDEX] (the 1st []) */
        _Id.br=_Id.b++; sg=1; _Id.indx=0;
        if (*_Id.b=='-') { sg= -1; _Id.b++; }
        while (isdigit(*_Id.b)) { _Id.indx=_Id.indx*10+sg*(*_Id.b++-'0'); }
        _Id.suff=_Id.b;
        if (*_Id.b!=']') { Error("missing ]"); goto next; }
        _Id.suff=_Id.b+1; }
      next: _Id.b++; }
  
  while (*_Id.b==' ') *_Id.b++=0;

#if CALC>=2
  /* postponing operators += -= *= ... */
  if (strchr("+-*/%^|&$",*_Id.b))
    if (*(_Id.b+1)=='=') {
      _Id.asg=*_Id.b; *_Id.b=0; _Id.b+=2;
      goto ret; }
#endif

  if (_Id.b==_Id.id || *_Id.b!='=') {
    /* anything else than `=' after ID : ID value or expression printed */
    _Id.prt=1;  /* to print ID name */
    _Id.used=0; /* to check if ID is OK */
    _Id.asg=1;  /* to assign to _Id.nr (= # in expr), CALC>=2 only */
#if CALC>=1
    if (_Id.b==_Id.id
     || (*_Id.b && strchr("+-*/^%:()",*_Id.b))
     || !strcmp("pi",_Id.id) || !strcmp("PI",_Id.id)  ) {
      /* expression to evaluate and print */
      /* if (_Id.br) Error("indexed variable in expression"); */
      _Id.used=2; /* suppress check + mark not id=expr */
      _Id.b=_Id.id; 
      goto evalexpr; }
#endif
#if CALC<2
    if (_Id.echo>=0) Error("syntax");
    /* to print ID value only */
#endif
    goto ret; }

  /* there was = (or += +- ...) after ID */
  *_Id.b++=0; /* mark string end (replace `=' etc. by zero) */

  /*** expression evaluation ***/
 evalexpr:

  while (*_Id.b==' ' || *_Id.b=='\t') _Id.b++;
  
  if (*_Id.b=='\"') {
    /* NEW: support of ID="string" -- _Id.str exported instead of _Id.nr */
    _Id.str=end=++_Id.b;
    _Id.nr=_Id.str[0]; /* ascii of 1st char */
    _Id.used=0; /* force error */
    while (*end) {
      if (end-_Id.b>64) Error("missing right \" or string too long");
      if (*end=='\"') { *end++=0; break; }
      end++; } }
  else {
    _Id.str=NULL;
#if CALC
    _Id.nr=Calc(_Id.b,&end);

#if CALC>=2
    /* += -= /= *= solved */
    switch (_Id.asg) {
      case '+': _Id.nr+=lhs; break;
      case '*': _Id.nr*=lhs; break;
      case '^': _Id.nr=pow(lhs,_Id.nr); break;
      case '-': _Id.nr=lhs-_Id.nr; break;
      case '/': _Id.nr=lhs/_Id.nr; break;
      case '%': _Id.nr=lhs-(int)(lhs/_Id.nr)*_Id.nr; break; }
    _Id.asg=0;
#endif

#else
    _Id.nr=strtod(_Id.b,&end);
#endif
    }

  if (end==_Id.b) {
    *(_Id.b=_Id.buf)=0;
    if (_Id.echo>=0) Error("data:bad number or expression (\"?\" for help)");
    return 0; }

  _Id.b=end;

 ret:
  _Id.end= *_Id.b==';';

  return 1;
} /* _GetId */

static struct keys_s *passkey;
static int errorkey;

static int findkey(void)
{
  struct keys_s *key=passkey;

  errorkey=0;
  while (key->key) {
    if (!strcmp(_Id.str,key->key)) return key->val;
    key++; }
  errorkey++;

  return 0;
}


#if CALC<2
int _GetCmp(char *id) /********************************************* _GetCmp */
#else
int _GetCmp(char *id,double val)
#endif
{
#if CALC>=3
  _Idlist *l;

  if (_Id.list) {
    l=(_Idlist *)malloc(sizeof(_Idlist)+strlen(id));
    if (!l) {
      fprintf(stderr,"no heap for _Id.list\n");
#ifdef CLOSEGRAPH
      closegraph();
#endif
      exit(-2); }
    l->next=_Id.head; _Id.head=l;
    strcpy(l->id,id);
    l->val=val;
    return 0; }
#endif
  
  if (_Id.prt>=2) {
    int x=strlen(id)+1;
    if ((_Id.col+=x)>79) { _Id.col=x; prtc('\n'); }
    prt_("%s ",id);
    return 0; }
  
  if (_Id.br!=NULL) *_Id.br='[';
  if (strcmp(id,_Id.id)) return 0;
  
  if (_Id.str) {
    if (passkey) {
      _Id.nr=findkey();
      if (errorkey) ERROR(("unknown string value in expression %s=\"%s\"",id,_Id.str)) }
    else
      ERROR(("variable %s accepts only numbers, not strings",id)) }
  
#if CALC>=2
  if (_Id.asg) _Id.nr=val;
#endif
  _Id.used=1;
  if (_Id.prt) prt_("  %s",id);
  
#if CALC>=3
  /* try to find the identifier in the list to update value */
  for (l=_Id.head; l; l=l->next)
    if (!strcmp(l->id,id)) l->val=_Id.nr;
#endif
  
  return 1;
} /* _GetCmp */


#if CALC<2
int _GetVecCmp(char *id,char *xsuff,int maxindex)
#else
int _GetVecCmp(char *id,char *xsuff,int maxindex,double val)
#endif
{
  /* this patch is needed for compilers which make string " " and not ""
     from an empty macro argument (e.g. Cray) */
  char suff[32];
  strcpy(suff,xsuff);
  if (*suff==' ') *suff=0; /* replace suffix " " by "" */
  
#if CALC>2
  /* nothing instead of list creating - cannot use indexed variables in expr. */
  if (_Id.list) return 0;
#endif
  
  if (_Id.prt>=2) {
    int x=strlen(id)+strlen(suff)+8;
    if (maxindex>999) x+=2;
    if ((_Id.col+=x)>79) { _Id.col=x; prtc('\n'); }
    prt_("%s[0..%d]%s ",id,maxindex-1,suff); return 0; }
  
  if (_Id.br==NULL) return 0;
  *_Id.br=0;
  if (strcmp(id,_Id.id) || strcmp(suff,_Id.suff)) return 0;
#if CALC>=2
  if (_Id.asg) _Id.nr=val;
#endif
  _Id.used=1;
  if (_Id.prt) prt_("  %s[%i]%s",id,_Id.indx,suff);
  if (_Id.indx>=maxindex || _Id.indx<0) {
    extError("data:index",id,_Id.indx);
    return 0; }
    
  return 1;
} /* _GetVecCmp */


void _CheckData(void) /****************************************** _CheckData */
{
  if (!_Id.used)
#if CALC>2
    if (!_Id.list)
#endif
      {
      *(_Id.b=_Id.buf)=0;
      if (_Id.echo>=0) Error("data:bad identifier or syntax (\"?\" for help)");
      }
}

int _GetEnd(void) /************************************************ _GetEnd */
{
#if CALC>2
  if (_Id.list) {
    /* was just after creating the list of identifiers */
    _Id.list=0;
    return 0; }
#endif
  
  if (_Id.prt) {
    if (_Id.prt>=2) prtc('\n');
    else prt(_Id.fmt,_Id.nr); }
  
  if (_Id.end) {
    if (_Id.echo==2) prtc('\n');
  
#if CALC>2
    { /* cleaning the list */
      _Idlist *Q,*QQ;
      QQ=_Id.head;
      while (QQ) {
        Q=QQ; QQ=QQ->next; free(Q); }
      _Id.head=NULL; }
#endif
    }
    
  return _Id.end;
}

#if CALC<2
int _GetKey(char *id,struct keys_s *key) /************************* _GetKey */
#else
int _GetKey(char *id,double val,struct keys_s *key)
#endif
{
  int ret;

  passkey=key;

#if CALC<2
  ret=_GetCmp(id);
#else
  ret=_GetCmp(id,val);
#endif
 
  passkey=NULL;

  return ret;
}

int prtkey_(int val,struct keys_s *key,int how) /******************* prtkey_ */
/*
  prints string ="NAME" according to val and key
  in case of several NAMEs with the same val:
    how=0:  prints the first occurrence in key
    how=1:  prints the last occurrence in key
    how=2:  prints all occurrences in key
  returns the number of characters printed
*/
{
  char *found=NULL;
  int i=0;

  while (key->key) {
    if (val==key->val) {
      if (how==1) found=key->key;
      else i+=prt_("=\"%s\"",key->key);
      if (how==0) break; }
    key++; }

  if (found) i+=prt_("=\"%s\"",found);

  return i;
}

void _PrtKey_(char *id,int val,struct keys_s *key) /************** _PrtKey_ */
/* 
  for putkey(id): id=VALUE="NAME" (1st occurrence) is printed
  filled to 26 chars
*/
{
  int i=prt_("%11s=%d",id,val);
 
  prtkey_(val,key,0);

  while (i<26) prtc(' '),i++;
}

/****** headers of tables ******/

void putline(char c,int l) /*************************************** putline */
{
  int i;
  /*** prints a line of l characters c on out ***/
  /* new: negative l means no LF */
  if (l) loop (i,0,abs(l)) prtc(c);
  if (l>0) _n
}

void header(char *h) /********************************************** header */
/***
  to make headers of tables. Usage:
    header("AA BB ");
    ...
    header("");
  will produce
  ======
  AA BB
  ------
  ...
  ======
  on out
  if the 1st character of the arg of header is special, it is used in front of
  hlines 
  new 8/99: h may contain \n
*/
{
  static int len=0;
  static char c1,c2;
  char *c;
  
  if (h && h[0]) {
    if (strchr("!@#$%^&*+[]:<>()./\\",h[0])) c1=c2=h[0];
    else { c1='='; c2='-'; }
  
    len=0;
    for (c=h; *c; c++) if (*c<' ') len=0; else len++;
    _n
    putline(c1,len);
    prt_("%s\n%c",h,c2);
    putline(c2,len-1); }
  else
    putline(c1,len);
}

void underline(char *t) /**************************************** underline */
/*
 produces

title
^^^^^

in BLEND, prepended by !
*/
{
#ifdef BLEND
  prt("\n! %s",t);
  prt_("! "); putline('^',strlen(t));
#else
  prt("\n%s",t);
  putline('^',strlen(t));
#endif
}

void graph(double x,int linelen) /*********************************** graph */
/*
  Pseudo-graphics with a char/4 resolution.
  For 0<=x<=1, a multi-character is printed at the position scaled by linelen.
*/
{
  int c,i;
  static char a[]="X X>><<X";
  static char l[128]; 
  
  if (linelen>125) linelen=125;

/*.....char *l; alloc(l,linelen+3);*/

  memset(l,' ',linelen+2);
  l[1]=l[linelen]='|'; 
  if (linelen&1) l[linelen/2+1]='|';
  c=(int)((linelen-1)*4*x+4.5); 
  i=(c+256)%4; c=(c-i)/4;
  if (c>=0 && c<=linelen) memcpy(l+c,a+2*i,2);
  l[linelen+1]=0;
  prt("%s",l+1);
/*.....free(l);*/
}

#ifdef MYSQRT
/* optional sqrt function replacement, see ground.h */

unsigned sqrt_tab[64] = {
   5376U,  12024U,  19815U,  27394U,  34750U,  41877U,  48762U,  55398U,
  61770U,  67867U,  73677U,  79186U,  84377U,  89240U,  93748U,  97887U,
 101635U, 104961U, 107845U, 110218U, 112068U, 113409U, 114654U, 115341U,
 115422U, 114907U, 113267U, 110876U, 107701U, 103741U,  98918U,  93152U,
  92222U,  96728U, 100941U, 104850U, 108425U, 111685U, 114529U, 117129U,
 119149U, 120699U, 122335U, 123565U, 124351U, 124648U, 124500U, 123853U,
 122666U, 120373U, 117526U, 114220U, 110158U, 105438U,  99982U,  93742U,
  86655U,  78666U,  69713U,  59729U,  48641U,  36365U,  22820U,   7901U };

#if MYSQRT==1
/* function */
double mySqrt(double _X)
{
  long unsigned sqrt_u;
  double sqrt_t, sqrt_x;
  union { long unsigned u; double d; } sqrt_ud;
  
  sqrt_ud.d = _X;
  sqrt_u = 0x5fe80000UL-(sqrt_ud.u>>33);
  sqrt_ud.u = sqrt_u-(long unsigned)sqrt_tab[(sqrt_u>>14) & 63UL]<<32;
  sqrt_x = sqrt_ud.d;
  sqrt_x *= 3.0-_X*sqrt_x*sqrt_x;
  sqrt_x *= 12.0-_X*sqrt_x*sqrt_x;
  sqrt_x *= 0.0625;
  sqrt_t = _X*sqrt_x;
  
  return sqrt_t+sqrt_x*0.5*(_X-sqrt_t*sqrt_t);
} /* end mySqrt */
              
#else
/* macro (inline), see ground.h */
long unsigned sqrt_u;
double sqrt_t, sqrt_x;
union sqrt_un sqrt_ud;
#endif
#endif

#include "fortran.c"
