/* cc -g -DX11 -DDEBUG -c xdraw.c
  MACSIMUS graphics interface
  (merges old-versions of: draw, drawx11, mouselib, mousekey; buttons)

  Long time ago, MACSIMUS used the Turbo C graphics library and DOS full
  screen graphics...
  Then I wrote a Turbo C graphics emulator under X11.
  Then I added some other layers, like DOS mouse (could be on/off) and
  extended keys (e.g., a single function key or cursor was transformed
  to a DOS 0+KEY code, and then translated to a single key again; mouse
  clicks were detected, but re-analyzed later from push/release data;
  all keys were buffered, but not mouse events..)
  DOS is dead!
  This is an attempt to clean the mess a bit, without too many changes
  in the existing MACSIMUS code.

  Note that MACSIMUS uses the following functions:
    kbhit(): returns 1 in case of any user-event available
    readkey(): a key or code returned:
               mouse position is in dosmouse.x, dosmouse.y
               blocks (waits) if none event queued
  Any event is transformed into a key (now int) and treated uniformly.
  Note that ESC=ctrl-[, TAB=ctrl-I, etc., as in terminal or DOS

  Mouse events are transformed into (used to be in mousekey, now here):
    LEFTCLICK = left mouse pressed, left mouse released, no x,y change
    LEFTDRAG = left mouse down, x,y change
    MIDCLICK,MIDDRAG,LEFTCLICK,LEFTDRAG = similarly
  MotionNotify without a mouse pressed does not generate any user-event
  MotionNotify with a key pressed can skip some events if slow
*/

#define Sqr(X) ((X)*(X))

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h> /* fabs needed */
#include "xdraw.h"

/*
  #define DEBUG: detailed
  #define GUI : Gtk-based support for menu (in a separate window)

  environment variable LAZYX11
  positive value: waits given # of seconds (max 5) before redraw
                  to be used for lazy networks
  negative value: suppresses geometry message to stderr
  -0            : suppresses geometry message to stderr, not lazy
*/

int x_color=15,x_fillcolor=15; /* as set by setcolor, setfillstyle */
static int currentcolor=-1; /* as color of gc's */

#include "drawbit.c"

struct xwindowhints_s xwindowhints = {
  drawbit_width,drawbit_height,/* X icon size */
  drawbit_bits,   /* icon data */
  NULL,           /* standard x-window geometry (passed from -g option) */
  "black",        /* background color (passed from -bg option) */
  "JK X-draw",    /* X window name */
  "X-draw"        /* X icon name */
};

struct dosmouse_s dosmouse; /* obsolete name */

Display *display;
int screen_num,display_width,display_height,no_colors;
Window win;
XSizeHints size_hints;
XEvent event;
XColor xcolor,hwcolor;
Colormap cmap;
GC gc;

#ifdef GUI
#include "config.h"
static Atom MyAtom;
struct gui_s gui;
#endif

#include "xbuttons.c"

long unsigned *xcoltab;

#define CLICK (LEFTCLICK-1)
#define DRAG (LEFTDRAG-1)

static struct pending_s {
  int x,y;    /* pointer position */
  int button; /* mouse button pressed; 1,2,3 or 0=none */
  int state;  /* one of CLICK, DRAG, and 0 */
} pending;

int x_xsize,x_ysize;

#define XERR(X,R) { fprintf(stderr,"draw/X11 error: %s\n",X); return R; }
#define XERE(X) { fprintf(stderr,"draw/X11 ERROR: %s\n",X); exit(-3); }

unsigned kbhit(void) /************************************************ kbhit */
/*
  basic X event loop goes through here:
  - if there is a pending event (not processed yet), 1 is returned
  - if no event is pending, a new one is checked for:
    - if one has come, it is stored (as pending) and 1 is returned
    - orherwise 0 is returned
  see also: readkey()
*/
{
  int aux;
#ifdef GUI
  XClientMessageEvent *evt;
  if (!MyAtom) MyAtom = XInternAtom(display, CTRL_ATOM_NAME, 0);
#endif

  if (!display) return 0;

  dosmouse.del=10; /* micro s */

#ifdef DEBUG
  {
    static int n=0;
    n++;
    fprintf(stderr,"[%d] dosmouse.key=%d\n",n,dosmouse.key);
  }
#endif
  //    usleep(1000);//DEBUG

  if (dosmouse.key) return 1; /* event already read and waiting for readkey() */

  /*      XNextEvent(display,&event); NB: XNextEvent blocks if no event */

 again:
  if (!XCheckMaskEvent(display,0xffffffff,&event))
    if (!XCheckTypedEvent(display,ClientMessage,&event)) {
      dosmouse.del=10000;
      return 0; }

#ifdef DEBUG
  fprintf(stderr,"kbhit: event came...");
#endif

  event.type &= 0x7f; /* mask sent events */

#ifdef DEBUG
  //  fprintf(stderr,"readkey: %c%c = %d\n", " ^"[ch<' '], ch>=' '?ch:ch+'@', ch);
#endif

  switch (event.type) {
#ifdef DEBUG
    /* ???: when a keyboard map is changed,
       keeps generating MappingNotify forever */
    case MappingNotify:
      fprintf(stderr,"MappingNotify\n");
      return dosmouse.key=X_REDRAW_ME;
#endif

    case Expose:
#ifdef DEBUG
      fprintf(stderr,"Expose\n");
#endif
      //      if (event.xexpose.count!=0) break;
      return dosmouse.key=X_REDRAW_ME;

    case ConfigureNotify:
      x_xsize=event.xconfigure.width;
      x_ysize=event.xconfigure.height;
#ifdef DEBUG
      fprintf(stderr,"ConfigureNotify: x-size=%d y-size=%d\n",x_xsize,x_ysize);
#endif
      return dosmouse.key=X_REDRAW_ME;

    case ButtonPress:
#ifdef DEBUG
      fprintf(stderr,"ButtonPress: x=%d y=%d  button=%d\n",
              dosmouse.x,dosmouse.y,event.xbutton.button);
#endif
      if (event.xbutton.button==Button4) return dosmouse.key=WHEELFORWARD;
      if (event.xbutton.button==Button5) return dosmouse.key=WHEELBACKWARD;
      pending.x=event.xbutton.x;
      pending.y=event.xbutton.y;
      if (event.xbutton.button==Button1) {
        struct buttons_s *b=getbutton();
        if (b) return dosmouse.key=b->key;
        pending.button=1; }
      else if (event.xbutton.button==Button3) {
        struct buttons_s *b=getbutton();
        if (b) {
          outbwtextxy(2,2,2,b->help);
          readkey();
          return dosmouse.key=X_REDRAW_ME; }
        pending.button=3; }
      else if (event.xbutton.button==Button2) pending.button=2;
      pending.state=CLICK;
      /* no key returned:
         waiting until pointer motion or button release */
      return dosmouse.key=0;
      break;

    case MotionNotify:
#ifdef DEBUG
      fprintf(stderr,"MotionNotify: x=%d y=%d  pending.button=%d\n",
              dosmouse.x,dosmouse.y,pending.button);
#endif
      if (pending.button) {
        /* motion with button pressed -> drag event */

        /* skiping all MotionNotify's and accepting the final one only */
        while (XCheckMaskEvent(display,ButtonMotionMask,&event));

        dosmouse.xl=pending.x;
        dosmouse.yl=pending.y;

        if (pending.state==CLICK) {
          /* 1st move after click: return the initial position */
          dosmouse.dx=dosmouse.dy=0;
          pending.state=DRAG;
          dosmouse.drag=0;
          dosmouse.x=dosmouse.x0=pending.x; pending.x=event.xbutton.x;
          dosmouse.y=dosmouse.y0=pending.y; pending.y=event.xbutton.y; }
        else {
          dosmouse.drag=1;
          dosmouse.dx=event.xbutton.x-pending.x;
          dosmouse.dy=event.xbutton.y-pending.y;
          /* following drag: returning the true position */
          dosmouse.x=pending.x=event.xbutton.x;
          dosmouse.y=pending.y=event.xbutton.y; }

        pending.state=DRAG;
        return dosmouse.key=LEFTDRAG-1+pending.button; }

      /* motion without button pressed -> no user-event */
      dosmouse.x=event.xbutton.x; /* not needed */
      dosmouse.y=event.xbutton.y; /* not needed */
      dosmouse.key=0;
      goto again;
      //      return dosmouse.key=0;
      break;

    case ButtonRelease:
#ifdef DEBUG
      fprintf(stderr,"ButtonRelease: x=%d y=%d\n",
              event.xbutton.x,event.xbutton.y);
#endif
      dosmouse.drag=2; /* end of drag */

      /* release of a key that has not been pressed...? */
      if (!pending.button) return dosmouse.key=0;

      dosmouse.x=event.xbutton.x;
      dosmouse.y=event.xbutton.y;
      if (pending.state==LEFTCLICK-1) {
        /* release after press -> click */
        aux=pending.button;
        pending.button=0;
        dosmouse.drag=3; /* not needed */
        return dosmouse.key=LEFTCLICK-1+aux; }

      /* end of drag: return the LAST position; the current is ignored (almost always is the same) */
      dosmouse.dx=event.xbutton.x-pending.x;
      dosmouse.dy=event.xbutton.y-pending.y;
      dosmouse.x=pending.x;
      dosmouse.y=pending.y;
      aux=pending.button;
      pending.button=0; /* no button pressed (multiple buttons pressed ignored) */
      return dosmouse.key=LEFTDRAG-1+aux;
      break;

#ifdef GUI
      /* message received from xmenu */
    case ClientMessage:
      evt = (XClientMessageEvent *) &event;

      if (evt->message_type == MyAtom) {
#ifdef DEBUG
        time_t ttt;
        fprintf(stderr,"ClientMessage (from xmenu): type=%ld, %ld at %ld s\n",
                (long int)evt->data.l[0],(long int)evt->data.l[1],(long int)time(&ttt));
#endif
        switch (evt->data.l[0]) {
          case ACTION_CONTROL_KEY:
            return dosmouse.key=evt->data.l[1];
            break;
          case ACTION_APP_KILL:
            exit(0); } }

      break;
#endif

    case KeyPress: {
      int n;
      char buf[128];
      KeySym key;
      XComposeStatus compose;

      n=XLookupString((XKeyEvent*)&event,buf,128,&key,&compose);

#ifdef DEBUG
      fprintf(stderr,"KeyPress: keysym=%d=%x\n",
              (int)key,(int)key);
#endif

      switch (key) {
        case  UP:
        case  DOWN:
        case  LEFT:
        case  RIGHT:
        case  INS:
        case  DEL:
        case  HOME:
        case  END:
        case  PGUP:
        case  PGDN:
        case  F1:
        case  F2:
        case  F3:
        case  F4:
        case  F5:
        case  F6:
        case  F7:
        case  F8:
        case  F9:
        case  F10:
        case  F11:
        case  F12:
          return dosmouse.key=key;

        case XK_Shift_L: case XK_Shift_R:
        case XK_Control_L: case XK_Control_R:
        case XK_Alt_L: case XK_Alt_R: return dosmouse.key=0; /* no event */

        default:
          if (buf[0]) return dosmouse.key=buf[0]; /* normal ASCII character */
          else return dosmouse.key=X_UNKNOWN_KEY; /* any control/function key not in
                                                     the above table */
      }
    }

  }
  return dosmouse.key=0;
}

void Imworking(int worktime) /************************************ Imworking */
{
  /* not needed anymore */
}

void repeatprefix(int key) /*********************************** repeatprefix */
{
  dosmouse.prefix=key;
  dosmouse.counter=-1;
  dosmouse.compose=0;
  dosmouse.repkey='?'; /* for debug only */
}

void repeatkey(int count,int key) /****************************** repeatkey */
{
  dosmouse.counter=count;
  dosmouse.compose=0;
  dosmouse.repkey=key;
}

int readkey(void) /***************************************************** readkey */
/*
  the interface finished:
  - an ASCII letter returns this letter, cast to (int); e.g., 'a' returns 97
  - shift-letter returns simply the uppercase letter; 'B' returns 66
  - ctrl-letter returns the ctrl code; Ctrl-C returns 3
  - other key returns its keysym
  - a mouse button returns CLICK and DRAG
  - mouse motion (without click) is suppressed (no event)
  - only shift, control, alt are suppressed (no event
  see kbhit() for more info!
*/
{
  unsigned aux;

  for (;;) {

    if (kbhit()) {
      /* an event has happened and is prepared in dosmouse.* */

      if (dosmouse.prefix) {
        if (dosmouse.key==dosmouse.prefix) {
          /* count prefix */
          if (dosmouse.compose) {
            /* already composing - cancel */
            dosmouse.compose=0;
            dosmouse.counter=0;
            dosmouse.key=0;
            continue; }
          /* initializing the counter */
          dosmouse.counter=0;
          dosmouse.key=0;
          dosmouse.compose=1;
          continue; }
        else if (dosmouse.compose) {
          if (dosmouse.key>='0' && dosmouse.key<='9') {
            /* composing the counter */
            dosmouse.counter=dosmouse.counter*10+(dosmouse.key-'0');
            dosmouse.key=0;
            continue; }
          else {
            /* key to repeat; noting passed directly */
            dosmouse.repkey=dosmouse.key;
            dosmouse.compose=0;
            dosmouse.key=0;
            continue; } }
        else if (dosmouse.counter>=0) {
          /* interrupting repetition */
          dosmouse.counter=-1;
          if (dosmouse.key==ESC) {
            /* ESC swallowed, other events passed */
            dosmouse.key=0;
            continue; } }
      }

      if (!dosmouse.key) continue; /* should not happen */
      aux=dosmouse.key;
#ifdef DEBUG
      fprintf(stderr,"readkey: %c%c = %d\n", " ^"[aux<' '], aux>=' '?aux:aux+'@', aux);
#endif
      dosmouse.key=0;
      return aux; }

    else { /* no event has happend */

      if (dosmouse.prefix && !dosmouse.compose) {
        if (dosmouse.counter==0)
          /* infinity */
          return dosmouse.repkey;
        else if (dosmouse.counter>0) {
          dosmouse.counter--;
          if (dosmouse.counter==0) dosmouse.counter=-1; /* no repeat anymore */
          return dosmouse.repkey; } } }

#ifdef DEBUG
    fprintf(stderr,"sleeping %d us\n",dosmouse.del);
#endif
    usleep(dosmouse.del);
  }
}

void delay(unsigned msec) /******************************************* delay */
/* not 100% portable ... */
{
  usleep(msec*1000); /* cf. also nanosleep () */
}

void clearviewport(void) /************************************ clearviewport */
{
  XClearWindow(display,win);
}

long unsigned winbackground;
static int xormode;

int startgraph(int mode) /*************************************** startgraph */
{
  int atx=1,aty=1,i;

  /* create X-window and set all necessary things */

  if (display) {
    /* already created - clear screen and raise instead */
    XRaiseWindow(display,win);
    XMapWindow(display,win);
    /*
       It looks like that XRaiseWindow is not sufficient, but why
       XMapWindow works, is a puzzle
    */
    clearviewport();

    {
      static XEvent dummy;
      dummy.xexpose.type=Expose;
      dummy.xexpose.display=display;
      dummy.xexpose.window=win;
      XSendEvent(display,win,False,ExposureMask,&dummy);
    }

    /*
       to eat all window exposure events (?!- at least one must come)
       (in fact, I don't know how to pop a window WITHOUT any event
    */
    i=readkey();
    if (i!=X_REDRAW_ME) fprintf(stderr,"unexpected event/key %c%c = %d\n",
                                " ^"[i<' '], i>=' '?i:i+'@', i);
    while (kbhit()==X_REDRAW_ME) (void)readkey();

    return abs(mode); }

  if (mode>16) no_colors=mode;
  else if (mode>0 && (mode&1)==0) no_colors=256;
  else no_colors=16;
  xcoltab=(long unsigned*)malloc(no_colors*sizeof(xcoltab[0]));
  if (xcoltab==NULL) XERE("no heap")
  memset(xcoltab,0,no_colors*sizeof(xcoltab[0]));

  display=XOpenDisplay(getenv("DISPLAY"));
  if (!display) XERR("XOpenDisplay",-1)
  screen_num=DefaultScreen(display);
  display_width=DisplayWidth(display,screen_num);
  display_height=DisplayHeight(display,screen_num);

#ifndef min
#define min(A,B) ((A)<(B)?(A):(B))
#endif
  x_xsize=min(640,display_width*2/3);
  x_ysize=min(480,display_height*2/3);

  if (xwindowhints.geometry) {
    char *ch;
    /* -g option */
    x_xsize=atoi(xwindowhints.geometry);
    if ( (ch=strchr(xwindowhints.geometry,'x')) ) {
      x_ysize=atoi(ch+1);
      if ( (ch=strpbrk(ch,"+-")) ) {
	atx=atoi(ch);
	if ( (ch=strpbrk(ch+1,"+-")) ) aty=atoi(ch); } } }

  if (x_xsize<8) x_xsize=100;
  if (x_xsize>display_width) x_xsize=display_width;

  if (x_ysize<8) x_ysize=100;
  if (x_ysize>display_height) x_ysize=display_height;

  { char *c=getenv("LAZYX11");
    if (!c || *c!='-')
      fprintf(stderr,"window geometry=%dx%d%+d%+d\n",x_xsize,x_ysize,atx,aty);
    }

  if (atx<0) atx+=display_width-x_xsize;
  if (aty<0) aty+=display_height-x_ysize;

  cmap=DefaultColormap(display,screen_num);

  /*.....winbackground=BlackPixel(display,screen_num);*/
  /* does not work generally for xor */

  if (!xwindowhints.background) xwindowhints.background="black";

  if (XAllocNamedColor(display,cmap,xwindowhints.background,&xcolor,&hwcolor))
    winbackground=xcolor.pixel;

  /*.....fprintf(stderr,"background=%ld\n",xcolor.pixel);*/

  win=XCreateSimpleWindow(display,RootWindow(display,screen_num),
			  atx,aty,
			  x_xsize,x_ysize,
			  0,
			  WhitePixel(display,screen_num),
			  winbackground);

#ifdef DEBUG
  fprintf(stderr,"Window ID: %d\n", (int)win);
#endif
#ifdef GUI
  {
    char *path=getenv("GUIPATH");

    if (path && gui.exe && gui.name) {
      int l=strlen(path)*3+strlen(gui.exe)+strlen(gui.name)*2+128;
      char *s=malloc(l);

      sprintf(s,"%s/%s -wid %ld -display %s -%s %s/xml/%s.xml -path %s &",
              path,gui.exe,
              (long int)win,
              DisplayString(display),
              gui.name,path,gui.name,
              path);
      fprintf(stderr,"%s\n",s);
      system(s);
      free(s); }
  }
#endif

  if (x_xsize>=DisplayHeight(display,screen_num) ||
      x_ysize>=DisplayHeight(display,screen_num)) {
    size_hints.flags=PPosition | PSize | PMinSize | PMaxSize | PWinGravity;
    size_hints.x=0;
    size_hints.y=0;
    size_hints.min_width=x_xsize;
    size_hints.max_width=x_xsize;
    size_hints.width=x_xsize;
    size_hints.min_height=x_ysize;
    size_hints.max_height=x_ysize;
    size_hints.height=x_ysize;
    size_hints.win_gravity=CenterGravity; }
  else {
    size_hints.flags=PPosition | PSize | PMinSize;
    size_hints.min_width=x_xsize;
    size_hints.min_height=x_ysize; }

  {
    static XWMHints wm_hints;
    static XClassHint class_hint;
    static int xnarg; /*?*/
    static char **xarg; /*?*/

    XTextProperty winname,iconame;

    Pixmap icon_pixmap=XCreateBitmapFromData(display,win,
      xwindowhints.icobits,xwindowhints.icowidth,xwindowhints.icoheight);

    XStringListToTextProperty(&xwindowhints.winname,1,&winname);
    XStringListToTextProperty(&xwindowhints.iconame,1,&iconame);

    wm_hints.initial_state=NormalState;
    wm_hints.input=True;
    wm_hints.flags=StateHint | InputHint | IconPixmapHint;

    wm_hints.icon_pixmap=icon_pixmap;

    XSetWMProperties(display,win,&winname,&iconame,xarg,xnarg,&size_hints,&wm_hints,&class_hint);
}

  XSelectInput(display,win,
	       ExposureMask | StructureNotifyMask
	       | KeyPressMask
	       | ButtonPressMask | ButtonReleaseMask | PointerMotionMask );

  {
    XGCValues gcvalues;

#if 1
    static char *turbo[16]={
      "#000000",
      "#000080",
      "#008000",
      "#008080",
      "#800000",
      "#800080",
      "#804000",
      "#c0c0c0",
      "#707070",
      "#0000ff",
      "#00ff00",
      "#00ffff",
      "#ff0000",
      "#ff00ff",
      "#ffff00",
      "#ffffff"};
#else
    static char *turbo[16]={
      "black",
      "darkblue",
      "darkgreen",
      "darkcyan",
      "darkred",
      "darkmagenta",
      "chocolate",
      "gray70",
      "gray40",
      "blue",
      "green",
      "cyan",
      "red",
      "magenta",
      "yellow",
      "white"};
#endif
    int i;

    gcvalues.function=GXcopy;
    gc=XCreateGC(display,win,GCFunction,&gcvalues);

    if (no_colors==16) {
      loop (i,1,16)
	if (!XAllocNamedColor(display,cmap,turbo[i],&xcolor,&hwcolor))
	  fprintf(stderr,"draw/X11: %s is unknown color\n",turbo[i]);
	else xcoltab[i]=xcolor.pixel;
      xcoltab[0]=winbackground;
      setcolor(WHITE); }
    }

  XMapWindow(display,win);

  /* wait until window is actually created, i.e., the first Expose */
  do {
    XNextEvent(display,&event);
#ifdef DEBUG
    fprintf(stderr,"startgraph: %s\n",
	    event.type==Expose?"Expose":"Other event");
#endif
    } while (event.type!=Expose);

  { char *c=getenv("LAZYX11");
    int i;

    if (c) {
      i=abs(atoi(c));
      if (i>5) i=5;
      if (i) sleep(i); } }

  return abs(mode);
}

void closegraph(void) /****************************************** closegraph */
/*
   graphics shut down ignored
   window remains created but inactive for future
*/
{
  if (!display) return;
  setwritemode(0);
  setfillstyle(1,LIGHTRED);
  setcolor(YELLOW);
  bar(0,0,90,40);
  settextstyle(0,0,1);
  settextjustify(0,2);
  outtextxy(12,8," WINDOW");
  outtextxy(12,22,"INACTIVE");
  setcolor(LIGHTRED);
  rectangle(0,0,getmaxx(),getmaxy());
  rectangle(1,1,getmaxx()-1,getmaxy()-1);
#if 0
  XFreeGC(display,gc);
  XCloseDisplay(display);
  display=NULL;
#endif
}

//void restorecrtmode(void) {}

void setpal(int col,int r,int g,int b,int maxbright) /*************** setpal */
/* maxbright<0 requests freeing colors-??? */
{
  int ret;
  if (col<0 || col>no_colors)
    XERR("color number out of range",)
  else {
    if (maxbright<0) {
      if (xcoltab[col]) XFreeColors(display,cmap,xcoltab+col,1,0);
      maxbright=-maxbright; }
    if (r+g+b==0) xcoltab[col]=winbackground;
    else {
      xcolor.red=r*0x10000/maxbright;
      xcolor.green=g*0x10000/maxbright;
      xcolor.blue=b*0x10000/maxbright;
      ret=XAllocColor(display,cmap,&xcolor);
      xcoltab[col]=xcolor.pixel;
      if (ret==0) XERR("cannot allocate color",) } }
}

void setcolor(int color) /***************************************** setcolor */
{
  static int first=1;
  if (first) {
    XSetForeground(display,gc,xcoltab[currentcolor=WHITE]);
    XSetBackground(display,gc,BLACK);
    first=0; }
  x_color=color;
}

void setgccolor(int color) /************************************* setgccolor */
{
  if (color<0 || color>=no_colors)
    XERR("color",)
      if (currentcolor!=color) {
    if (xormode)
      XSetForeground(display,gc,xcoltab[currentcolor=color]^winbackground);
    else
      XSetForeground(display,gc,xcoltab[currentcolor=color]); }
}

void setfillstyle(int mode,int color) /************************ setfillstyle */
{
  /* only mode=1 (normal fill) supported */
  x_fillcolor=color;
}

void setlinestyle(int mode,int pattern,int thick)  /*********** setlinestyle */
/* only 1 dotted line supported */
/* thick more general than for DOS */
{
  XGCValues gcvalues;
  unsigned long mask=GCLineStyle|GCLineWidth;

  if (mode==0)
    gcvalues.line_style=LineSolid;
  else {
    mask|=GCDashList;
    gcvalues.line_style=LineOnOffDash;
    gcvalues.dashes=2;
    XSetDashes(display,gc,2,"\1\3",2);
  /* "\1\3" is dot-black-black-black */ }

  gcvalues.line_width=thick;
  XChangeGC(display,gc,mask,&gcvalues);
}

void setwritemode(int mode) /********************************** setwritemode */
{
  XGCValues gcvalues;
  if (mode) gcvalues.function=GXxor;
  else gcvalues.function=GXcopy;
  XChangeGC(display,gc,GCFunction,&gcvalues);
  xormode=mode;
  currentcolor=-1;
}

static int ggx,ggy;

void line(int x,int y,int xx,int yy)  /******************************** line */
{
  setgccolor(x_color);
  XDrawLine(display,win,gc,x,y,ggx=xx,ggy=yy);
}

void draw(REAL x1,REAL y1) /******************************************* draw */
{
  int gx=SX(x1), gy=SY(y1);
  setgccolor(x_color);
  if (pen) XDrawLine(display,win,gc,ggx,ggy,gx,gy);
  ggx=gx;
  ggy=gy;
  pen=1;
}

void bar(int x0,int y0,int x1,int y1) /********************************* bar */
{
  /* warning: good only if x1>x0, y1>y0 !!! */
  setgccolor(x_fillcolor);
  XFillRectangle(display,win,gc,x0,y0,x1-x0+1,y1-y0+1);
}

void rectangle(int x0,int y0,int x1,int y1) /********************* rectangle */
{
  setgccolor(x_color);
  XDrawRectangle(display,win,gc,min(x0,x1),min(y0,y1),abs(x1-x0),abs(y1-y0));
}

void putpixel(int x,int y,int color) /***************************** putpixel */
{
  setgccolor(color);
  XDrawPoint(display,win,gc,x,y);
}

void fillellipse(int x,int y,int rx,int ry) /******************* fillellipse */
{
  setgccolor(x_fillcolor);
  XFillArc(display,win,gc,x-rx,y-rx,rx*2,ry*2,0,360*64);
  setgccolor(x_color);
  XDrawArc(display,win,gc,x-rx,y-rx,rx*2,ry*2,0,360*64);
}

void circle(int x,int y,int r) /************************************* circle */
{
  setgccolor(x_color);
  XDrawArc(display,win,gc,x-r,y-r,r*2,r*2,0,360*64);
}

static int textsize=1;

void settextstyle(int a,int b,int size)
{
  textsize=size;
}

static int xjust,yjust=2;

void settextjustify(int x,int y)
{
  xjust=x; yjust=y;
}

/* To install a new font, consult ppm2xfont.c and generate a C-include file.
   Then you need to change:
   - NXFONTS in xdraw.h
   - below:
     #define XFONT xfont<NUMBER>
     #include "<NEWFONT>.c"
     #undef XFONT
   - in selectfont():
     xfonts[<NUMBER>]=xfont<NUMBER>;
*/

#define XFONT xfont0
#include "xfont7.c"
#undef XFONT

#define XFONT xfont1
#include "xfbgi.c"
#undef XFONT

#define XFONT xfont2
#include "xfont10.c"
#undef XFONT

#define XFONT xfont3
#include "xfont15.c"
#undef XFONT

/* NEWFONT: add similar 3 lines as above */

                     /* width,xwidth,height,top,base,from,to,data */
struct xfont_s xfont = {8,8,9,0,7,0,256, xfont1+7}; /* BGI */
static int selectedfont;

void selectfont(int font)
{
  static unsigned *xfonts[NXFONTS];
  xfonts[0]=xfont0;
  xfonts[1]=xfont1;
  xfonts[2]=xfont2;
  xfonts[3]=xfont3;
  /* NEWFONT: add a line here */

  if (font<0 || font>=NXFONTS) font=1;
  selectedfont=font;

  xfont.width=xfonts[font][0];
  xfont.xwidth=xfonts[font][1];
  xfont.height=xfonts[font][2];
  xfont.top=xfonts[font][3];
  xfont.base=xfonts[font][4];
  xfont.from=xfonts[font][5];
  xfont.to=xfonts[font][6];
  xfont.data=xfonts[font]+7;
}

int outtextxy(int x,int y,const char *s)
{
  unsigned X,Y,B,l=0;
  unsigned ch;

  x-=xjust*(strlen(s)*xfont.width/2-1)*textsize;
  switch (yjust) {
    case 0: y-=xfont.base*textsize; break;
    case 1: y-=(xfont.base+xfont.top)/2*textsize; break;
    case 2: y-=xfont.top*textsize; }

  while (*s) {
    ch=*(unsigned char*)s;

    if (ch>=xfont.from && ch<xfont.to) loop (X,0,xfont.xwidth) {
        B=xfont.data[(ch-xfont.from)*xfont.xwidth+X];

      if (textsize==1)
        loop (Y,0,xfont.height) {
          if (B&1) putpixel(x+X,y+Y,getcolor());
          B>>=1; }
      else
        loop (Y,0,xfont.height) {
          if (B&1) {
            int ii,jj;
            loop (ii,X*textsize,X*textsize+textsize)
              loop (jj,Y*textsize,Y*textsize+textsize)
              putpixel(x+ii,y+jj,getcolor()); }
          B>>=1; } }

    l+=xfont.width*textsize;
    x+=xfont.width*textsize;
    s++; }

  return l;
}

int outtext(const char *s)
{
  return outtextxy(ggx,ggy,s);
}

int outbwtextxy(int x,int y,int font,const char *s) /*********** outbwtextxy */
/* output text as black on white using font, restore colors/font */
{
  int oldf=selectedfont,oldc=x_color,oldbg=x_fillcolor,oldts=textsize;
  unsigned char *locs=(unsigned char*)strdup(s),*c;
  int aty=2+xfont.top,l=0,nl=0;
  unsigned char end;

  selectfont(font);
  setcolor(BLACK);
  settextstyle(0,0,1);
  setfillstyle(1,WHITE);
  
  end=' '; c=locs;
  while (end) {
    unsigned char *e=c;

    while (!strchr("\n\r",*e)) e++;
    end=*e; if (l<e-c) l=e-c;
    c=e+1; }

  end=' '; c=locs;
  while (end) {
    unsigned char *e=c;

    while (!strchr("\n\r",*e)) e++;
    end=*e; *e=0;
    line(0, aty+xfont.base+2, l*xfont.width+5, aty+xfont.base+2);
    line(l*xfont.width+5, aty-xfont.top-1, l*xfont.width+5, aty+xfont.base+1);
    bar(0, aty-xfont.top-1, l*xfont.width+4, aty+xfont.base+1);
    outtextxy(2,aty,(char*)c); nl++;
    c=e+1; 
    if (end!='\r') aty+=xfont.height+2; }

  XFlush(display);
  free(locs);
  selectfont(oldf);
  setcolor(oldc);
  setfillstyle(1,oldbg);
  settextstyle(0,0,oldts);

  return nl;
}

/*** scaled graphics in x,y Cartesian coordinates ***/

struct scaling_s scaling;
int pen;

void up(void) /******************************************************** up */
{
  pen=0;
}

static int wx0,wx1=639,wy0,wy1=479; /* VGA 16col as default */

void lwindow(int x0,int x1,int y0,int y1) /************************* window */
/* renamed because of window() in conio.h */
/* normally, x0<x1 and y0<y1: y0 is thus TOP and y1 BOTTOM */
{
  if (x1<=x0) x1=getmaxx();
  if (y1<=y0) y1=getmaxy();
  wx0=x0; wx1=x1;
  wy0=y0; wy1=y1;
}

void scale(REAL x0,REAL x1,REAL y0,REAL y1) /*********************** scale */
{
  scaling.x=(wx1-wx0)/(x1-x0);
  scaling.fromx=x0-wx0/scaling.x;
  scaling.y=(wy1-wy0)/(y1-y0);
  scaling.toy=y1+wy0/scaling.y;
}

int SX(REAL x) /****************************************************** SX */
{
  int i=Int(scaling.x*(x-scaling.fromx)+16383.5)-16383;

  if (i<-32767) i=-32767;
  if (i>32767) i=32767;

  return i;
}

REAL iSX(int ix)
{
  return ix/scaling.x+scaling.fromx;
}

int SY(REAL y) /***************************************************** SY */
{
  int i=Int(scaling.y*(scaling.toy-y)+16383.5)-16383;

  if (i<-32767) i=-32767;
  if (i>32767) i=32767;

  return i;
}

REAL iSY(int iy)
{
  return scaling.toy-iy/scaling.y;
}

void lline(REAL x1,REAL y1,REAL x2,REAL y2) /************************* lline */
{
  up(); draw(x1,y1); draw(x2,y2);
}

void dot(REAL x1,REAL y1) /********************************************* dot */
{
  putpixel(SX(x1),SY(y1),getcolor());
}

/* radius of put pixel area:
   2pix/dot on average; min 0.707 to have 1 pix sure */
REAL pointsize=0.798;
int insidecolor=BLACK; /* for open points */

void point(REAL x1,REAL y1,char type) /******************************* point */
{
  REAL size=pointsize;
  int open=(unsigned)type<128,iopen;

  type=tolower(type&127);
  if (strchr("+x",type)) open=0;

  if (strchr("+dv^",type)) size*=1.414;

  if (size<=0)
    dot(x1,y1);
  else {
    REAL sx=scaling.x*(x1-scaling.fromx);
    REAL sy=scaling.y*(scaling.toy-y1);

    if (fabs(sx)<4096.0 && fabs(sy)<4096.0) loopto (iopen,0,open) {

      switch (type) {

        case 'x': case '*': {
          int
            x0=Int(sx+0.5-size),
            x9=Int(sx+0.5+size),
            y0=Int(sy+0.5-size),
            y9=Int(sy+0.5+size);

          line(x0,y0,x9,y9); line(x0,y9,x9,y0); }

          if (type=='x') break;
          else size*=1.414;

        case '-': case '+': {
          int
            x=Int(sx+0.5),
            y=Int(sy+0.5);

          if (strchr("*+-",type)) line(Int(sx+0.5-size),y,Int(sx+0.5+size),y);
          if (strchr("*+|",type)) line(x,Int(sy+0.5-size),x,Int(sy+0.5+size));
          break; }

        default: {
          int
            x =Int(sx+1-size),
            x9=Int(sx+1+size),
            y,
            y0=Int(sy+1-size),
            y9=Int(sy+1+size);
          REAL ss=Sqr(size),sv=size/2;
          unsigned col=iopen ? insidecolor : getcolor();

          for (; x<x9; x++)
            for (y=y0; y<y9; y++)
              if ((type=='o' && Sqr(x-sx)+Sqr(y-sy)<ss)
               || (type=='d' && fabs(x-sx)+fabs(y-sy)<size)
               || (type=='s' && fabs(x-sx)<size && fabs(y-sy)<size)
               || (type=='^' && y-sy<=sv && 1.7320508*fabs(x-sx)-(y-sy)<size)
               || (type=='v' && sy-y<=sv && 1.7320508*fabs(x-sx)-(sy-y)<size))
                putpixel(x,y,col);
          break; } }

      size=size*0.75-1; } }
}

void bigdot(REAL x1,REAL y1) /************************************** bigdot */
{
  int i=SX(x1),j=SY(y1),c=getcolor();

  putpixel(i,j,c);
  putpixel(i+1,j,c); putpixel(i-1,j,c);
  putpixel(i,j+1,c); putpixel(i,j-1,c);
}

