#include "guiapp.h"
#include <gdk/gdkx.h> // for GDK_WINDOW_XID macro


namespace gctrl {

using sigc::mem_fun;


void GUIApp::connect_action_clicked ( Gtk::Button *b, long data, long action )
{
  if (b) {
    b->signal_clicked().connect( 
        sigc::bind<long,long>(mem_fun(*this, &GUIApp::send_event), action, data ));
  }
}


void GUIApp::connect_action_clicked ( Gtk::ToolButton *b, long data, 
                                      Glib::ustring img_name, long action )
{
  if (b) {
    b->signal_clicked().connect( 
        sigc::bind<long,long>(mem_fun(*this, &GUIApp::send_event), action, data ));
    if ( img_name.size() > 0 ){
      Glib::ustring path = Glib::build_filename(g_base_path, DIR_ICONS );
      Gtk::Image *img = new Gtk::Image( Glib::build_filename( path, img_name) );
      b->set_icon_widget(*img);
      img->show();
    }
  }
}


GUIApp::GUIApp ( long tgt_win_id, char *xml_file, Glib::ustring window_title_ )
  : tgt_win_id_(tgt_win_id), window_title(window_title_)
{
  this->set_title(window_title);
  this->resize(320,240);
  this->add_events(Gdk::ALL_EVENTS_MASK);
  this->set_resizable(false);

//  std::cout << "g_display_name: " << g_display_name << std::endl;

  set_help( HELP_DEFAULT );
  help_dialog = new HelpDialog ( str_help );

  // buttons
  b_help = new Gtk::Button(Gtk::Stock::HELP);
  b_quit = new Gtk::Button(Gtk::Stock::QUIT);
  b_save = new Gtk::Button(Gtk::Stock::SAVE);


  // XXX: IMPORTANT - decision between simple GUI and xml specified 
  // Load the GtkBuilder file and instantiate its widgets:
  if (xml_file) {
    builder = Gtk::Builder::create();
    builder->add_from_file( xml_file );
  
    builder->get_widget("main_vbox", vbox_main);
    vbox_main->reparent(*this);

    if ( !vbox_main ){
      std::cerr << "Error: GUIApp::GUIApp(): File \"" << xml_file 
        << "\" doesn't contain Gtk::VBox named \"main_vbox\"." << std::endl;
      return;
    }  
  }
  else { // GUIApp is running directly
    label.set_text("Label\n\n\n\n");
    label.show();
    vbox_main = new Gtk::VBox();
    vbox_main->pack_start( label );
  }

  b_help->show();
  b_quit->show();

  // buttonbox (BOTTOM)
  button_box.set_spacing(5);
  button_box.pack_start(*b_help);
  button_box.pack_start(*b_save);
  button_box.pack_start(*b_quit);
  button_box.show();
  
  separator = new Gtk::HSeparator();
  separator->show();

  vbox_main->set_spacing(5);
  vbox_main->pack_start ( *separator, true, true, 5 );
  vbox_main->pack_end( button_box );
  vbox_main->set_border_width(10);
  vbox_main->show();

  this->add(*vbox_main);

  set_position();
  this->show();

  // signals
  if ( b_help )
    b_help->signal_clicked().connect ( mem_fun(*this, &GUIApp::show_help ) );

  if ( b_quit )
    b_quit->signal_clicked().connect ( mem_fun(*this, &GUIApp::on_quit ) );

  
  // get Xwindow ID of current GTK window
  Glib::RefPtr<Gdk::Window> gwindow = this->get_window();
  if(gwindow){
	  this_win_id_ = GDK_WINDOW_XID(gwindow->gobj());
  // XXX
//    gwindow->add_filter( client_message_filter, NULL);
    std::cout << "GUIApp::this_win_id_ = " << this_win_id_ << std::endl;
  }
  else
    std::cerr 
      << "Error: GUIApp::GUIApp(): can't get Xwindow ID of current window."
      << std::endl;
}


GUIApp::~GUIApp ()
{
  delete_widget(separator);
  delete_widget(b_help);
  delete_widget(b_quit);
  delete_widget(b_save);
  delete_widget(vbox_main);
}


void GUIApp::on_quit ()
{
  // this can generate error: "window id not found"
  // GUIApp::send_event() always show error dialog
  // x11::send_event() only writes to std::cerr
  x11::send_event(tgt_win_id_, g_display_name, CTRL_ATOM_NAME, ACTION_APP_KILL); 
  Gtk::Main::quit();
}


bool GUIApp::on_delete_event ( GdkEventAny *event )
{
  x11::send_event(tgt_win_id_, g_display_name, CTRL_ATOM_NAME, ACTION_APP_KILL); 
  return Gtk::Window::on_delete_event(event);
}


void GUIApp::set_position()
{
  /*
  Gtk::Window::move(x,y) 
    - posun celeho okna vcetne dekoraci WM
  XGetWindowAttributes(display, tgt_win_id_, &attrs) 
    - informace o okne BEZ dekoraci (dekorace = jine okno)
  Gtk::Window::get_frame_positions(&top,&left,&right,&bottom)
    - informace o celem okne vc. dekoraci
  */
  x11::xwindow_attributes attrs;
  if ( x11::get_xwindow_attributes(tgt_win_id_, g_display_name, attrs)){
    int top, left, right, bottom;
    this->get_frame_dimensions( top, left, right, bottom );
    this->move( /* x */ attrs.width  + attrs.border_width + left, 
                /* y */ top );
  }
  else
    std::cerr << "GUIApp::set_position(): can't get XWindowAttributes." << std::endl;
}


void GUIApp::send_event ( long action, long data )
{
  if ( !x11::send_event( tgt_win_id_, g_display_name, CTRL_ATOM_NAME, action, data ) ){
    std::stringstream ss;
    ss << "Can't send message to window " << tgt_win_id_ << std::endl;
    Gtk::MessageDialog dialog(*this, ss.str(), true, Gtk::MESSAGE_ERROR);
    dialog.run();
    Gtk::Main::quit();
  }
}


void GUIApp::set_help ( Glib::ustring str )
{
  str_help = str;
  if ( help_dialog )
    help_dialog->set_text(str);
};


void GUIApp::show_help ()
{
  if ( !help_dialog ) 
    help_dialog = new HelpDialog(str_help);
  help_dialog->show();
}


bool GUIApp::on_key_press_event ( GdkEventKey *ev )
{
  guint val = ev->keyval;
  send_event( ACTION_CONTROL_KEY, val );
  // only for simple GUI
  if ( label.is_visible() )
    label.set_markup( Glib::ustring::compose("<b>Send: </b> %1", val) );
  return true;
};


/////////////////////////////////////////////////////////////////////

GdkFilterReturn 
client_message_filter ( GdkXEvent *xev, GdkEvent *ev, gpointer data)
{
  if ( !g_display_name ){
//    std::cerr << "client_message_filter: no display name\n";
    return GDK_FILTER_CONTINUE;
  } 
  XClientMessageEvent *cm = (XClientMessageEvent*)ev;
  std::cout << cm->type << std::endl;
  if ( cm->message_type == XInternAtom(XOpenDisplay(g_display_name), CTRL_ATOM_NAME, 0)){
    std::cout << "ClientMessage catched.\n";
    switch ( cm->data.l[0] ){
      case ACTION_CONTROL_KEY:
        std::cout << "key\n";
        break;
      default:
        break;
    } // switch cm->data.l[0]
    return GDK_FILTER_REMOVE;
  } // if ClientMessage
  return GDK_FILTER_CONTINUE;
}

}; // namespace gctrl
