#include "guiapp.h"
#include "gui_show.h"
#include "gui_blend.h"

#include <cstdlib>

/******************************************************************** 

  Can I use C++ exceptions with gtkmm?
  ====================================

  Yes, it is possible but it is not a very good idea. The official answer is that, 
  since plain C doesn't know what a C++ exception is, you can use exceptions in your 
  gtkmm code as long as there are no C functions in your call stack between the 
  thrower and the catcher. This means that you have to catch your exception locally.

  You will be warned at runtime about uncaught exceptions, and you can specify 
  a different handler for uncaught exceptions. Some gtkmm methods do even use exceptions 
  to report errors. The exceptions types that might be thrown are listed in the reference 
  documentation of these methods. 

  SRC: http://live.gnome.org/gtkmm/FAQ
********************************************************************/

Glib::ustring g_base_path;
char *g_display_name;

enum AppMode { DEFAULT, BLEND, SHOW };

void usage ()
{
  std::cout 
    << "USAGE: gshow -wid WINDOW_ID [OPTIONS]\n" 
    << "OPTIONS:\n"
    << "  -blend FILE\n"
    << "  -show FILE\n"
    << "  -display DISPLAY\n"
    << "  -path PATH"
    << std::endl;
}

/////////////////////////////////////////////////////////////////////

int main ( int argc, char **argv )
{
  long win_id;
  char *xml_file = 0;
  AppMode mode = DEFAULT;
  g_display_name = ":0.0";

  if (getenv("GSHOWKILL")) {
    // JK: ugly patch to allow the gshow window to be killed from scripts
    // by running script "kill.sh"
    FILE *id=fopen("kill.sh","wt");
    fprintf(id,"kill %d\n",getpid());
    fclose(id);
  }

  // commandline
  if ( argc < 3 ){
    std::cerr << "Error: Wrong # of arguments." << std::endl;
    usage();
    return 1;
  }
  
  // XXX: doesn't work with os.system(), only from shell
  g_base_path = Glib::path_get_dirname(argv[0]);

  int i = 1;
  while ( i < argc ){
    if ( strcmp(argv[i], "-blend") == 0 ){
      xml_file = argv[++i];
      mode = BLEND;
      std::cout << "Using blend interface.\n"
        << "XML File:   " << xml_file << std::endl;
    }  
    else if ( strcmp(argv[i], "-show") == 0 ){
      xml_file = argv[++i];
      mode = SHOW;
      std::cout << "Using show interface.\n"
        << "XML File:   " << xml_file << std::endl;
    }  
    else if ( strcmp(argv[i], "-wid") == 0 ){
      std::stringstream ss(argv[++i]);
      ss >> win_id;
      std::cout << "Target window ID: " << win_id << std::endl; 
    }  
    else if ( strcmp(argv[i], "-display") == 0 ){
      g_display_name = argv[++i];
      std::cout << "Display:   " << g_display_name << std::endl;
    }  
    else if ( strcmp(argv[i], "-path") == 0 ){
      g_base_path = argv[++i];
    }
    else {
      std::cerr << "Error: Unknown option \"" << argv[i] << "\".\n";
      usage();
      return 1;
    }
    ++i;
  }

  std::cout << "Base path: " << g_base_path << std::endl;

  // Run App
  Gtk::Main kit(argc, argv);
  try {
    if ( mode == SHOW ) {
      gctrl::ShowApp app( win_id, xml_file );
      app.set_help(HELP_SHOW);
      kit.run(app);
    }
    else if ( mode == BLEND ) {
      gctrl::BlendApp app( win_id, xml_file );
      app.set_help(HELP_BLEND);
      kit.run(app);
    }
    else {
      std::cout << "Starting simple interface.\n";
      gctrl::GUIApp app( win_id );
      kit.run(app);
    }
  }
  catch (const Glib::FileError& ex) {
    std::cerr << "FileError: " << ex.what() << std::endl;
    return 1;
  }
  catch (const Gtk::BuilderError& ex) {
    std::cerr << "BuilderError: " << ex.what() << std::endl;
    return 1;
  }
  catch (const std::exception &e ){
    std::cerr << "Error: " << e.what() << std::endl;
    return 1;
  }

  return 0;
}
