#!/bin/bash
if [ $# -lt 1 ]; then
  echo "Install MACSIMUS. Call by:"
  echo "  ./install.sh COMPILER"
  echo "or"
  echo "  bash install.sh COMPILER"
  echo "Where COMPILER is the C compiler name:"
  echo "  gcc = GCC compiler - first choice for linux"
  echo "        the default optimization level is -O2 -ffast-math"
  echo "        -O3 may work, but has not been extensively tested"
  echo "  colorgcc = as above, colorized messages"
  echo "  icc = Intel compiler - faster code than gcc, less reliable on AMD"
  echo "        the default optimization level is -O2"
  echo "        flag -fast is not reliable (Trace/breakpoint trap)"
  echo "        locale must be set, LANG=C is recommended"
  echo "        (bash: use export LANG=C, tcsh: use setenv LANG C)"
  echo "  pgcc = Portland Group, Inc. (PGI) compiler"
  echo "        BUG: RANDOM NUMBERS DO NOT WORK (to be solved later)"
  echo "  cc = \"any\" compiler, edit compile.mmk and check your system for options"
  echo "For implementing a new compiler, see compile.mmk"
  if [ -n "$CC" ] ; then
    echo "Example (based on your \$CC=$CC):"
    echo "  ./install.sh $CC"
  else
    echo "Example:"
    echo "  ./install.sh gcc"
  fi
else

# makemake option: now unix and linux are equivalent
UNIX=linux

# no spaces in the path allowed
if [ "${PWD/ /z}" != "${PWD}" ] ; then
  echo "Sorry, I cannot handle spaces in PWD=\"$PWD\"!" 
  echo "Please install MACSIMUS in a path not containing spaces"
  exit
fi

# check if the selected compiler exists
if ! which $1 ; then
  echo "compiler $1 not found"
  exit
fi

# check bin/makeone.sh
if [ ! -e bin/makeone.sh ]; then
  echo "ERROR bin/makeone.sh not found (MACSIMUS incorrectly unzipped?)"
  exit
fi
chmod +x bin/makeone.sh

# using lib64 if exists (probably no longer needed)
# if [ -d /usr/X11R6/lib64 ] ; then
# cat <<EOF | ed graph.mmk
# ,s:/usr/X11R6/lib:/usr/X11R6/lib64:g
# w
# q
# EOF
# echo /usr/X11R6/lib64 used
# fi

# environment for cook/configure.sh
cat > makemake.sh <<EOF
export MAKEMAKE=$1
# this file is used by script cook/configure.sh
EOF

echo
echo "Do you wish to configure 'show' and 'blend' with a GTK-based"
echo "graphical user interface? (y/N)"
echo "- Required: libgtkmm (2.16 or later), based on gtk 2.40 (not 3.0),"
echo "  with development files and all dependencies, see gui/Makefile."
echo "- If you answer n, only hot-key based interfaces are available,"
echo "  providing the same functionality."
read
if [ "$REPLY" == "y" ] ; then
  GUI=gui
  cd gui
  if ! make ; then
    echo "compilation of GUI for show failed, GUI support removed"
    GUI=""
  fi
  cd ..
else
  GUI=""
fi

# export MACSIMUS home directory for makemake
cat > home.mmk <<EOF
# this file defines the MACSIMUS home directory
# it is !included by all metamake files
# non-! statements will become a part of the generated makefile files

EOF
echo '!home='${PWD} >> home.mmk
echo "home.mmk generated"

echo
echo ">>> compiling the makemake utility"
cd c/
rm *.o
$1 -o makemake makemake.c
if ! ./makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm makemake # will be recompiled

echo
echo ">>> compiling general utilities from subdirectory c/"
for f in \
  makemake \
  tabproc smoothg minmax \
  ev \
  mergetab tab endian start sortcite seltab filttab lemon eqfield sum \
  lattice sumetc blocktab eddata ice histogr runsum \
  runint binrepl liat lc
do
  ../bin/makeone.sh $f
done
cd ..

# support of fitting by plot
cd c/fit4plot
rm *.o *.ol *.ox
../../bin/makemake $UNIX $1 $2 long
mv makefile makefile.long
../../bin/makemake $UNIX $1 $2 high
mv makefile makefile.high
../../bin/makemake $UNIX $1 $2
mv makefile makefile.double
cd ../..

echo
echo ">>> compiling blend (lj version), pdb, and similar from blend/"
cd blend/
if ! ../bin/makemake $UNIX lj $1 $2 $GUI ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi

rm *.o
for f in \
  blend pdb showpro makepept blefilt ramachan pdb2pdb bonds
do
  ../bin/makeone.sh $f
done

echo
echo ">>> installing bin versions of common force fields from blend/data/"
cd data
rm gromos96.bin charmm19.bin charmm21.bin charmm22.bin
export BLENDPATH=.
../../bin/blend -b gromos96.par
../../bin/blend -b charmm19.par
../../bin/blend -b charmm21.par
../../bin/blend -b charmm22.par
cd ../..

# compiling graphical utilities
cd show/
if ! ../bin/makemake $UNIX $1 $2 $GUI; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm *.o
for f in \
  show plot tomoil plbmatch
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling different simulation utilities from util/"
cd util/
if ! ../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm *.o
for f in \
  autocorr cppak molcfg plb2diff plbmsd plbpak rdfg sfourier showcp \
  spectrum staprt stereo \
  atomdist coordn cp2cp cutprt plbcut hbonds density densprof plbfilt \
  plbmerge plbinfo plb2nbr plb2cryst shownear plbstack plbreplicate plbsites \
  plb2asc plbcut plb2plb plbconv asc2plb cfg2atm \
  plb2cfg cfg2plb cfg2atm plbframe plbscale plbbox plbtran mol2mol
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling the raytracer + related software from ray/"
echo "(more utilities available upon request)"
cd ray/
if ! ../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi

rm *.o
for f in \
  ray ppm2ps ppminfo ppmscale ppm2ppm
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling three cook versions, see MACSIMUS/cook/exe/"
cd cook/exe/free
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookfree
mv cookfree ../../../bin/
cd -

cd cook/exe/cutelst
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookce
mv cookce ../../../bin/
cd -

cd cook/exe/ewaldslab
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookewslc
mv cookewslc ../../../bin/cook
cd -

echo
cd bin
# LC_COLLATE=C added by T.Trnka (../bin.lst is of C locale order)
LC_COLLATE=C \ls -1 > ../instbin.lst
# this is because of CygWin
cat <<EOF | ed ../instbin.lst
,s:\.exe::g
w
q
EOF
echo "Comparing (using diff):"
echo "  the contents of ${PWD}/ (instbin.lst)"
echo "  with the expected listing bin.lst"
echo "----------------------------------------------------------------------------"
if diff ../bin.lst ../instbin.lst ; then
  echo "No difference detected - MACSIMUS has been successfully compiled"
else
  echo "----------------------------------------------------------------------------"
  echo "ERROR: components listed above with < have not been successfully compiled"
  echo "(or the following utilities are missing: diff ed)."
fi
cd ..

echo
echo "If you wish to use the MACSIMUS calculator \"ev\" and start utility \"start\","
echo "you must install the definition files, .evdata and .startdata."
echo "Hint: get help by running \"ev\" and \"start\" without parameters"
echo "Copy .evdata and .startdata to your home (y/N)?"
read
if [ "$REPLY" == "y" ] ; then
  cp -ip .evdata .startdata ~/
fi

echo =============================================================================
echo "To finish installation:"
echo "* IF you are using sh or bash,"
echo "  THEN append the following lines to your ~/.bashrc file:"
echo "    export BLENDPATH=${PWD}/blend/data"
if [ "$GUI" = "gui" ]; then
echo "    export GUIPATH=${PWD}/gui"
fi
echo "    export PATH=${PWD}/bin:\$PATH"
echo "* ELSE use the syntax of your shell."
echo
echo "* Three Lennard-Jones versions of cook has been precompiled:"
echo "  - cook = Ewald sums, slab + density profile support, linked-cell list"
echo "  - cookce = cutoff electrostatics, simplest all-neighbor algorithm"
echo "  - cookfree = in vacuum"
echo "* You may customize and compile your own cook* version(s) as follows:"
echo "  - go to directory" ${PWD}/cook
echo "    and run script ./configure.sh"
echo "  - alternatively, customize" ${PWD}/generic/metamake
echo "    and" ${PWD}/generic/simopt.h "by hand"
echo "    and then run makemake $1 and make cook*"
echo
echo "* The basic Lennard-Jones version of blend has been precompiled."
echo "  To compile other blend versions:"
echo "  - follow the instructions in ${PWD}/blend/metamake"
echo
echo "Most utilities print help if run without any option, sometimes -h is needed"
echo "If a utility mentioned after \"See also:\" is not present, please let me know"
echo =============================================================================
echo
echo "Would you like to start a simple test of MACSIMUS now (y/N)?"
read
if [ "$REPLY" == "y" ] ; then
  export BLENDPATH=${PWD}/blend/data
  if [ "$GUI" = "gui" ]; then
      export GUIPATH=${PWD}/gui
  fi
  export PATH=${PWD}/bin:$PATH
  cd examples
  echo "NB: the environment has been set within the script."
  ./test.sh
else
  echo "You can run the test later as test.sh from ${PWD}/examples."
fi

fi
