/*
  non-bonded (atom-atom) potential support + combining rules
  see sitesite.h for more info
*/

#include "ground.h"
#include "vector.h" /* simglob.h not needed */
#include "sitesite.h"
#ifdef COOK
#  include "units.h"
#  define PROMPT
#else
#  define PROMPT "! "
#endif

/*** combining rules ***/

void combruleinfo(int comb_rule) /***************************** combruleinfo */
{
  if (comb_rule==0) prt(PROMPT "The Waldman-Hagler combing rule:");
  if (comb_rule==15) prt(PROMPT "The Busing-equivalent combing rule:");
  prt_(PROMPT "sqrt_rule&3: B_ij=");
  switch (comb_rule&3) {
    case  0: prt("[2/(1/B_ii^6+1/B_jj^6)]^(1/6)"); break;
    case  1: prt("(B_ii B_jj)^1/2 (geometric)"); break;
    case  2: prt("(B_ii+B_jj)/2 (arithmetic)"); break;
    case  3: prt("2/(1/B_ii+1/B_jj) (harmonic)"); break;
  }

  prt_(PROMPT "sqrt_rule&12: A_ij=");
  switch (comb_rule&12) {
    case  0: prt("(A_ii A_jj)^1/2 B_ij^6/(B_ii B_jj)^3"); break;
    case  4: prt("(A_ii A_jj)^1/2 (geometric)"); break;
    case  8: prt("(A_ii A_jj)^1/2 (arithmetic)"); break;
    case 12: prt("A_ij=(A_ii B_ii)^qi (A_jj B_jj)^qj / B_ij (busing)\n\
             qi=1/B_ii/(1/B_ii+1/B_jj), qj=1/B_jj/(1/B_ii+1/B_jj)"); break;
  }
  prt(PROMPT "C_ij/r^6: geometric combining rule: C_ij=(C_i C_j)^1/2");
  prt(PROMPT "D_ij/r^12: geometric combining rule: D_ij=(D_i D_j)^1/2");
}

void initcombrule(siteparm_t *sp,int comb_rule) /************** initcombrule */
/* 
   for historical reasons, site parameters have the following names:
     sp->EvdW = A_ii, assumed in kcal/mol, converted to K
     sp->RvdW = B_ii, assumed in 1/A
     sp->parm[0] = C_ii, assumed in kcal/mol A^6, converted to K A^6
     sp->parm[1] = D_ii, assumed in kcal/mol A^12, converted to K A^12
  */
{
#ifdef COOK
  sp->EvdW*=(kcal/Eunit);    /* A_ii */
  sp->parm[0]*=(kcal/Eunit); /* C_ii */
  sp->parm[1]*=(kcal/Eunit); /* D_ii */
#endif
  sp->sqA=sqrt(sp->EvdW);
  sp->sqB=sqrt(sp->RvdW);
  sp->Bi6=0.5/Pow6(sp->RvdW);
  sp->Bi=0.5/sp->RvdW;
  sp->sqABi3=sp->sqA/Cub(sp->RvdW);
  sp->AB=sp->EvdW*sp->RvdW;
  sp->sqC=sqrt(sp->parm[0]);
#ifdef REP12
  sp->sqD=sqrt(sp->parm[1]);
#endif  /*# REP12 */
  sp->comb_rule=comb_rule;
}

void combrule(pairparm_t *pp,siteparm_t *lj0,siteparm_t *lj1,char *info)
{
  double B6,qi;
  /* 
     for historical reasons, site-site parameters have the following names:
     pp->eps = A_ij
     pp->sig = B_ij
     pp->parm[0] = C_ij
     pp->parm[1] = D_ij 
  */
  if (lj0->comb_rule!=lj1->comb_rule)
    ERROR(("atom combining rules differ (%d vs. %d)",lj0->comb_rule,lj1->comb_rule))

  if ((lj0->comb_rule&3)!=2 && lj0->RvdW*lj1->RvdW==0) pp->sig=0; 
  else switch (lj0->comb_rule&3) {
    case 0: B6=lj0->Bi6+lj1->Bi6;
            pp->sig=pow(B6,-1./6); break;
    case 1: pp->sig=lj0->sqB*lj1->sqB; break;
    case 2: pp->sig=(lj0->RvdW+lj1->RvdW)/2; break;
    case 3: pp->sig=1/(lj0->Bi+lj1->Bi); break; }

  if (lj0->EvdW*lj1->EvdW==0) pp->eps=0;
  else switch (lj0->comb_rule&12) {
    case 0: pp->eps=lj0->sqABi3*lj1->sqABi3/B6; break;
    case 4: pp->eps=lj0->sqA*lj1->sqA; break;
    case 8: pp->eps=(lj0->EvdW+lj1->EvdW)/2; break;
    case 12:qi=lj0->Bi/(lj0->Bi+lj1->Bi);
            pp->eps=pow(lj0->AB,qi)*pow(lj1->AB,1-qi)/pp->sig; break; }

  pp->parm[0]=lj0->sqC*lj1->sqC;
#ifdef REP12
  pp->parm[1]=lj0->sqD*lj1->sqD;
  if (info) prt("%s: eps=%g sig=%g C=%g D=%g",info,pp->eps,pp->sig,pp->parm[0],pp->parm[1]);
#else
  if (info) prt("%s: eps=%g sig=%g C=%g",info,pp->eps,pp->sig,pp->parm[0]);
#endif
}

/* calculate auxiliary constants for site-site interactions */
void initssaux(pairaux_t *a,pairparm_t *pp)
{
  a->Aij=pp->eps;
  a->Bij=pp->sig;
  a->C=pp->parm[0];
  a->C6=pp->parm[0]*6;
#ifdef REP12
  a->D=pp->parm[1]; \
  a->D12=pp->parm[1]*12;
#endif /*# REP12 */
}
