/* Exp-6 potential with optional r^-12 term

    u_rep(r) = A_ij exp(-B_ij r)
    u(r) = u_rep(r) - C_ij/r^6 [ + D_ij/r^12 ]

  The site parameters (X_i, X={A,B,C,D}) appearing in the table
  are the same as the site-site parameters X_ii

  Version:
    #define REP12: the extra repulsive term  D_ij/r^12  is added

  Compatible with:
    not POLAR
    POLAR&1==0 (standard linear polarizability)
    POLAR&1==1 (shell-core model, cf. Busing version)

  Parameters:
    alpha   -- unused
    EvdW    -- Aii  [kcal/mol], converted to K internally
    RvdW    -- Bii  [1/A]
    parm[0] -- Cii  [kcal/mol.A^6], converted to K A^6 internally
    parm[1] -- Dii  [kcal/mol.A^12], (converted to K A^12 internally

  Combining rules:

  combrule&3:
    0: B_ij=[2/(1/B_ii^6+1/B_jj^6)]^(1/6) (Waldman-Hagler)
    1: B_ij=(B_ii B_jj)^1/2               (sqrt = geometric)
    2: B_ij=(B_ii+B_jj)/2                 (arithmetic)
    3: B_ij=2/(1/B_ii+1/B_jj)             (harmonic = additive diam.; Busing)
   0,1,3: B_ij=0 if B_ii=0 or B_jj=0

  combrule&12:
    0: A_ij=(A_ii A_jj)^1/2 B_ij^6/(B_ii B_jj)^3  (Waldman-Hagler)
    4: A_ij=(A_ii A_jj)^1/2                       (sqrt = geometric)
    8: A_ij=(A_ii+A_jj)/2                         (arithmetic, not useful)
   12: A_ij=(A_ii B_ii)^qi (A_jj B_jj)^qj / B_ij  (Busing)
         qi=1/B_ii/(1/B_ii+1/B_jj)
         qj=1/B_jj/(1/B_ii+1/B_jj)
   A_ij=0 if A_ii=0 or A_jj=0

  r^-6 and r^-12 terms:
    C_ij=(C_i C_j)^1/2
    D_ij=(D_i D_j)^1/2
    NOTE: some other modules may use C_ij=C_i C_j and/or D_ij=D_i D_j
*/

#ifndef SITE_SITE
#  define SITE_SITE

#  ifdef REP12
#    define SS_TABLE "Buckingham+12"
#  else /*# REP12 */
#    define SS_TABLE "Buckingham"
#  endif /*#!REP12 */

#  include "simopt.h"
#  include "units.h"
#  include "int4.h"

/*** exp6 tables and combining rules ***/
#  ifdef REP12
#    define SS_PARMS      2
#    define DECLARE_INITCOMBRULE \
  double sqA,sqB,Bi6,Bi,sqABi3,AB,sqC, sqD;
typedef struct pairaux_s {
  real Aij,Bij,C,C6,D,D12;
} pairaux_t;
#  else /*# REP12 */
#    define SS_PARMS      1
#    define DECLARE_INITCOMBRULE \
  double sqA,sqB,Bi6,Bi,sqABi3,AB,sqC;
typedef struct pairaux_s {
  real Aij,Bij,C,C6;
} pairaux_t;
#  endif /*#!REP12 */

/* auxiliary constants for site-site interactions */
/* will be accessible via sitesite_t->a */

#  include "ssdecl.h"

#  if defined(POLAR) && (POLAR&1)

/* Urep'', used by blend, r=sqrt(rr) must be passed */
#    define SS_ffrep if (ss->a.Aij==0) ffrep=0; else {     \
  y=ss->a.Aij*exp(-ss->a.Bij*r);    \
  ffrep=Sqr(ss->a.Bij)*y; }

/* site-site repulsive energy and forces */
#    define SS_MEASURE_rep if (ss->a.Aij==0) Urep=frep=0; else \
  x=sqrt(rr); /* double eval? */       \
  y=ss->a.Aij*exp(-ss->a.Bij*x);       \
  Urep=y;  \
  frep=ss->a.Bij*y/x;
/* site-site repulsive forces only */
#    define SS_NOMEASURE_rep           \
  x=sqrt(rr); /* double eval? */       \
  y=ss->a.Aij*exp(-ss->a.Bij*x);       \
  frep=ss->a.Bij*y/x;

/* site-site energy and forces (called after SS_[NO]MEASURE_rep) */
#    ifdef REP12
#      define SS_MEASURE { double zz;         \
  z=Cub(rr);                                  \
  zz=ss->a.D/Sqr(z);                          \
  z=ss->a.C/z;                                \
  U=Urep-z+zz;                                \
  f=frep+(12*zz-6*z)/rr; }
#      define SS_NOMEASURE(OP) {              \
  z=Cub(rr);                                  \
  f OP frep+(ss->a.D12/z-ss->a.C6)/(rr*z); }
#    else /*# REP12 */
#      define SS_MEASURE {    \
  z=ss->a.C/Cub(rr);          \
  U=Urep-z;                   \
  f=frep-6*z/rr; }
#      define SS_NOMEASURE(OP) { \
  f OP frep-ss->a.C6/Pow4(rr); }
#    endif /*#!REP12 */

#  else /*# defined(POLAR) && (POLAR&1) */

/* standard (POLAR not shell-core or nonPOLAR) */
#    ifdef REP12
#      define SS_MEASURE { double zz; \
  x=sqrt(rr); /* double eval? */  \
  y=ss->a.Aij*exp(-ss->a.Bij*x);  \
  z=Cub(rr);                      \
  zz=ss->a.D/Sqr(z);              \
  z=ss->a.C/z;                    \
  U=y-z+zz;                       \
  f=ss->a.Bij*y/x+(12*zz-6*z)/rr; }
#      define SS_NOMEASURE(OP) {                \
  x=sqrt(rr); /* double eval? */                \
  y=ss->a.Aij*exp(-ss->a.Bij*x);                \
  frep=ss->a.Bij*y/x;                           \
  z=Cub(rr);                                    \
  f OP frep+(ss->a.D12/z-ss->a.C6)/(rr*z); }
#    else
/* attractive (or LJ) part */
#      define SS_MEASURE {        \
  x=sqrt(rr); /* double eval? */  \
  y=ss->a.Aij*exp(-ss->a.Bij*x);  \
  z=ss->a.C/Cub(rr);              \
  U=y-z;                          \
  f=ss->a.Bij*y/x-6*z/rr; }
#      define SS_NOMEASURE(OP) {       \
  x=sqrt(rr); /* double eval? */       \
  y=ss->a.Aij*exp(-ss->a.Bij*x);       \
  f OP ss->a.Bij*y/x-ss->a.C6/Pow4(rr); }
#    endif
#  endif  /*#!defined(POLAR) && (POLAR&1) */

#endif /*# SITE_SITE */
