#define ermacro

/*  J.Kolafa 1991, 1992(2D)                                                erfc

General interface for COULOMB approximation, specifically:

COULOMB = -1,-2: erfc functions for r-space part of Ewald summation
COULOMB = 0: MACSIMUS cut-off electrostatics (directly - no splines).
COULOMB = 1: (reserved)
COULOMB = 2: MACSIMUS cut-off electrostatics by quadratic splines
COULOMB = 3: Fennell and Gezelter [JCP 124, 234104 (2006)] cut-and-shift
      potential by quadratic splines.

*/

#include "simopt.h"

#ifdef FLOAT
typedef float erreal;
#else
typedef double erreal;
#endif

#if !defined(COULOMB) || defined(FREEBC) || defined(NIBC)
#error "The electrostatic module expects #defined COULOMB and undefined FREEBC and NIBC.  Check metamake, makefile, simopt.h"
#endif

typedef struct { erreal Au,Bu,Cu,Ad,Bd,Cd; } ertab_t,*ertab_p;

#if COULOMB==-2
#ifndef erexact_eps
#ifdef POLAR
#define erexact_eps 1e-15
#else
#define erexact_eps 1e-10
#endif
#endif
#endif

extern struct Erfc_s {
  int key;          /* 0=1/r cut, 2=2D, 3=3D erfc, ... */
  int ndiv;         /* to def integration step; must be at least 4 */
  int grid;         /* d(r^2)=1/grid */
#if COULOMB!=0
  int shift;        /* 1 = shifted to be continuous (tiny error) */
#endif
  ertab_p tab;      /* table offset */
  ertab_p from,to;  /* valid range of the table */
  erreal sgrid;
#ifdef erminmax
#undef ermacro
  erreal min,max;
#endif
#ifdef errange
#undef ermacro
#endif
#ifdef erexact_eps
  erreal alphaq,A,B;
#endif
  erreal minr,maxr,alpha,beta;
#if COULOMB==0
  erreal shift,A,B,A3,B3,r0;
#endif
} Erfc;

extern erreal byerd; /* by-product result of erud
  NOTE: if ermacro is not #defined then this byerd is used,
  if #ifdef ermacro then a local byerd may be faster, however,
  you must not declare local byerd if ermacro is not #defined !
  macro ermacrodcl may be used within the function's declarations ! */

#ifdef ermacro

#if COULOMB<0
#define eru(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Bu/(er_p->Cu+(x))+er_p->Au)
#define erd(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Bd/(er_p->Cd+(x))+er_p->Ad)
#define erud(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  byerd = er_p->Bd/(er_p->Cd+(x)) + er_p->Ad,  \
  er_p->Bu/(er_p->Cu+(x)) + er_p->Au)
#elif COULOMB>=2
/* cut 1/r, with cutelst.c or fgelst.c */
#define eru(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Au+(x)*(er_p->Bu+er_p->Cu*(x)))
#define erd(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Ad+(x)*(er_p->Bd+er_p->Cd*(x)))
#define erud(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  byerd=er_p->Ad+(x)*(er_p->Bd+er_p->Cd*(x)), \
  er_p->Au+(x)*(er_p->Bu+er_p->Cu*(x)))
#elif COULOMB==0
/* cut 1/r directly (wrong design to have it HERE), with cutelstd.c */
#define eru(x) (er_x=sqrt(x), \
  er_x<Erfc.r0 ? \
  1/er_x-Erfc.shift : \
  Cub(er_x-box.cutoff)*(Erfc.A+Erfc.B*er_x))
#define erd(x) (er_x=sqrt(x), \
  er_x<Erfc.r0 ? \
  er_x/Sqr(x) : \
  Sqr(er_x-box.cutoff)*(Erfc.A3+Erfc.B3*er_x-(er_x-box.cutoff)*Erfc.B)/er_x)
#define erud(x) (er_x=sqrt(x), \
  er_x<Erfc.r0 ? \
  (byerd=er_x/Sqr(x),1/er_x-Erfc.shift) : \
  (byerd=Sqr(er_x-box.cutoff)*(Erfc.A3+Erfc.B3*er_x-(er_x-box.cutoff)*Erfc.B)/er_x,\
    Cub(er_x-box.cutoff)*(Erfc.A+Erfc.B*er_x)))
#else
#error invalid COULOMB
ERROR invalid COULOMB
#endif

#if COULOMB==0
#define ermacrodcl erreal byerd,er_x;
#else
#define ermacrodcl ertab_p er_p; erreal byerd;
#endif

#else

double eru(double x);
double erd(double x);
double erud(double x);
#define ermacrodcl /* empty */

#endif

void initerfc(int Grid, double minr, double maxr, double cutoff, double alpha, double beta, int shift);

#ifdef erexact_eps
double exacterud_sqrt(double xx,erreal *erd);
double exacterud_sqrt_1(double xx,erreal *erd);
#endif
