/*****************************************************************************
  fast (but memory consuming) method to calculate the following functions:

  eru(x) = beta*alpha*e(sqrt(x)*alpha)
  erd(x) = -beta*alpha^3*e'(sqrt(x)*alpha)/(sqrt(x)*alpha)

where

  e(y) = erfc(y)/y, e'(y)=de(y)/dy,

                         infinity
  erfc(y) = 2/sqrt(PI) * integral exp(-t**2) dt,
                            y

and alpha and beta are scaling parameters

method:
  hyperbolic spline (functions A+B/(x+C), no jumps in the derivative)

interface:
  #ifdef erminmax  then min and max arguments of eru(x), erd(x), and erud(x)
    will be recorded in global variables Erfc.min, Erfc.max
  #ifdef errange  then overflow of the tables is checked
  #ifdef ermacro  then eru(x), erd(x), and erud(x) are macroized. The argument
    is used twice (3x in erud) => do not use things like eru(x+=1) !
    variable  er_p  of type  ertab_p  must be known
    using erminmax or errange undefines this conditional
  #define FLOAT: precision for tables and macros (see simopt.h)
( #ifdef eropt  then accuracy will be optimized, however, this procedure
    is very slow and for wide range of arguments leads only to a
    negligible improvement
  #ifdef erinterr integration error checked - slow! )
  if (tr==0) then a protocol is printed to stream out
  requires #include "alloc.c" and "error.c"

initialization:
  initerfc(int Grid, double minr,maxr,alpha,beta)
where:
  [minr,maxr) defines the range of r, the range of the argument x of eru, erd,
    and erud is then [minr^2,maxr^2)
  Grid is a number of grid points per unity
  (negative Grid means only that some info prints are suppressed)

note: using argument x < minr leads to a low-accuracy results (this is NOT
  checked #ifdef errange) while x > maxr is out of the table (checked
  #ifdef errange)

2D version:
^^^^^^^^^^^
if TWODIM is #defined, the following functions are approximated by the same
method:

  eru(x) = beta*E1(alpha^2*x)/2
  erd(x) = beta*exp(-alpha^2*x)/x

where

          infinity
  E1(y) = integral dt/(t exp(t))
              y

This is the exponential integral.  It is, of course, silly to call the module
erfc, however, compatibility is compatibility...

BUG: erfc is calculated from scratch by integrals, this is not needed
with modern math libraries...
*****************************************************************************/

#if COULOMB>=0
#error "COULOMB<0 expected"
#endif

#ifdef TWODIM
#define DIM 2
#else
#define DIM 3
#endif

struct Erfc_s Erfc = { DIM,4 };

erreal byerd; /* by-product result of erud
  NOTE: if ermacro is not #defined then this byerd is used,
  if #ifdef ermacro then a local byerd may be faster, however,
  you must not declare local byerd if ermacro is not #defined ! */

static void erfcs(long to,double *e,double *z) /********************** erfcs */
/*
  e=eru(h*to) and z=erd(h*to) are computed.
  Must be called with decreasing sequence of parameters to, starting
  with a sufficiently high value because the integral is for this to
  approximated by the asymptotic formula
*/
{
#ifdef TWODIM
#define f(X) ((X)*exp(X))
#else
#define f(X) (sqrt(X)*exp(X))
#define SQRTPI 1.772453850905516
#endif
  static long t= -1;
  static double sum;
  double
    h=1e0/(Erfc.ndiv*Erfc.sgrid),
    w=h/2, ha=h*0.2113248654051871, hb=h-ha,
    x=to*h;

  if (x>706) ERROR(("erfcs: range too large - cannot calculate splines"))
  else if (x>67) {
    static int pass;
    if (pass==0) {
      WARNING(("erfcs: range unnecessarily large (er<1e-30)\n\
*** (more warnings suppresed)"))
      pass=1; } }

  if (to>t) {
    t=to; x=to*h;
    if (x<1) ERROR(("erfcs: range too short"))
#ifdef TWODIM
    sum=exp(-x)/(x+1-0.9/x);
#else
    /* the correct asymptotic formula is
       exp(-x)/sqrt(x+1-0.75/x+2/x^2-7.75/x^3+-... */
    sum=exp(-x)/sqrt(x+1-0.7/x);
#endif
  }

  while (t>to) {
    x=h*t;
    sum += w/f(x-ha)+w/f(x-hb);
    t--; }


  x=t*h+1e-20;
#ifdef TWODIM
  *e=0.5*sum;
  *z=exp(-x)/x;
#else
  *e=sum/sqrt(x);
  *z=(*e+2/exp(x))/x/SQRTPI;
  *e /= SQRTPI;
#endif
#undef f
} /* erfcs */

#ifdef ermacro

#define eru(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Bu/(er_p->Cu+(x))+er_p->Au)
#define erd(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  er_p->Bd/(er_p->Cd+(x))+er_p->Ad)
#define erud(x) (er_p=Erfc.tab+(int)((x)*Erfc.sgrid), \
  byerd = er_p->Bd/(er_p->Cd+(x)) + er_p->Ad,  \
  er_p->Bu/(er_p->Cu+(x)) + er_p->Au)

#else

double eru(double x) /************************************************** eru */
{
  ertab_p er_p=Erfc.tab+(int)(x*Erfc.sgrid);
#ifdef erminmax
  if (x<Erfc.min) Erfc.min=x;
  if (x>Erfc.max) Erfc.max=x;
#endif
#ifdef errange
  /*.....if (er_p<Erfc.from || er_p>=Erfc.to) Error("eru:range");*/
  if (er_p>=Erfc.to) Error("eru:range");
#endif
  return er_p->Bu/(er_p->Cu+x) + er_p->Au;
} /* eru */

double erd(double x) /************************************************** erd */
{
  ertab_p er_p=Erfc.tab+(int)(x*Erfc.sgrid);
#ifdef erminmax
  if (x<Erfc.min) Erfc.min=x;
  if (x>Erfc.max) Erfc.max=x;
#endif
#ifdef errange
  /*.....if (er_p<Erfc.from || er_p>=Erfc.to) Error("erd:range");*/
  if (er_p>=Erfc.to) Error("erd:range");
#endif
  return er_p->Bd/(er_p->Cd+x) + er_p->Ad;
} /* erd */

double erud(double x) /************************************************ erud */
/* erud computes eru as above and gives erd as a by-product in variable byerd */
{
  ertab_p er_p=Erfc.tab+(int)(x*Erfc.sgrid);
#ifdef erminmax
  if (x<Erfc.min) Erfc.min=x;
  if (x>Erfc.max) Erfc.max=x;
#endif
#ifdef errange
  /*.....if (er_p<Erfc.from || er_p>=Erfc.to) Error("erud:range");*/
  if (er_p>=Erfc.to) Error("erud:range");
#endif
  byerd = er_p->Bd/(er_p->Cd+x) + er_p->Ad;
  return er_p->Bu/(er_p->Cu+x) + er_p->Au;
} /* eru */
#endif

#ifdef eropt
static double eerfcs(double q, long ifrom, long ito) /*************** eerfcs */
#else
static void eerfcs(double q, long ito) /***************************** eerfcs */
#endif
/* ito,ifrom in units of step */
{
  long x;
  double e,z,eh,zh,ee,zz,ed,zd,xh,xxh;
  double step=1e0/Erfc.sgrid, h=step/Erfc.ndiv;
  ertab_p p;

#ifdef eropt
  double sq;
  ertab_p er_p;
  long i;
#endif

  x=ito*Erfc.ndiv; /* in units of h */
  erfcs(x+1,&eh,&zh);
  erfcs(x  ,&e, &z);
  erfcs(x-1,&ee,&zz);
  ed=(eh-ee)/(2*h)*q; zd=(zh-zz)/(2*h)*q;
  xh=x*h;

  while (ito) { ito--;
    xxh=xh; ee=e; zz=z;
    erfcs(x=ito*Erfc.ndiv,&e,&z);
    xh=x*h;

    p=Erfc.tab+ito;
    p->Cu=(xh*(ee-e)-xxh*ed*step) / (ed*step-(ee-e));
    p->Bu= -ed*(p->Cu+xxh)*(p->Cu+xxh);
    p->Au=ee+ed*(p->Cu+xxh);
    ed= -p->Bu/(p->Cu+xh)/(p->Cu+xh);

    p->Cd=(xh*(zz-z)-xxh*zd*step) / (zd*step-(zz-z));
    p->Bd= -zd*(p->Cd+xxh)*(p->Cd+xxh);
    p->Ad=zz+zd*(p->Cd+xxh);
    zd= -p->Bd/(p->Cd+xh)/(p->Cd+xh); }

#ifdef eropt /* inefficient ! */
  ifrom *= Erfc.ndiv; sq=0;
  for (i=ifrom+Erfc.ndiv; i>ifrom; i--) {
    erfcs(ito*Erfc.ndiv,&e,&z);
    erfcs(i,&e,&z);
    sq += sqr(e-eru(i*h)) + sqr(z-erd(i*h)); }

  prt("q=%f  sqr err=%e",q,sqrt(sq/Erfc.ndiv));

  return sq;
#endif
} /* eerfcs */

void initerfc(int Grid, double minr, double maxr, double cutoff, double alpha, double beta, int shift)
/* NEW: Grid<0 means suppressed info */
{
  alpha=fabs(alpha); /* negative = verbose */
  {
    long ifrom,ito,i;
    int grid=abs(Grid),verbose=Grid>0;
#ifdef eropt
    double r0=0.0001,r,q,f,ff;
#endif
    double x,e,ee,z,zz;
    double h=1e0/(Erfc.ndiv*grid);
    double mine=0,maxe=0,minz=0,maxz=0,minre=0,maxre=0,minrz=0,maxrz=0;
#ifdef TWODIM
    double Ad=beta*Sqr(alpha), AuBd=beta, Bu=beta/Sqr(alpha), Cud=1/Sqr(alpha);
#else
    double Ad=beta*Cub(alpha), AuBd=beta*alpha, Bu=beta/alpha, Cud=1/Sqr(alpha);
#endif
    ertab_p er_p;

    ifrom=(int)(Sqr(minr*alpha)*grid); ito=(int)(Sqr(maxr*alpha)*grid)+1;
    /* ifrom,ito are in units of 1/grid */

    if (Erfc.grid==grid && Erfc.shift==shift 
        &&Erfc.minr==minr && Erfc.maxr>=maxr
        && Erfc.alpha==alpha && Erfc.beta==beta) {
      Erfc.to=Erfc.tab+ito;
      prt(":::::: erfc already initialized ::::::");
      return; }

    Erfc.shift=shift;
    Erfc.sgrid=(erreal)grid;
    Erfc.grid=grid; Erfc.minr=minr; Erfc.maxr=maxr;
    Erfc.alpha=alpha; Erfc.beta=beta;

#ifdef erexact_eps
    Erfc.A=beta*alpha*(-2/SQRTPI);
    Erfc.alphaq=Sqr(alpha);
    Erfc.B=Erfc.A*Erfc.alphaq*2;
#endif

    if (Erfc.tab!=NULL) free(Erfc.tab);
    allocarray(Erfc.tab,ito);

    prt("\n:::::: erfc :::::: tab 0..%ld..%ld = %ld B  grid=%d", ifrom,ito,ito*sizeof(ertab_t),grid);

    if (verbose) prt("subgrid=%d  alpha=%g  beta=%g  range=[%g,%g)  cutoff=%g",
                     Erfc.ndiv,alpha,beta,minr,maxr,cutoff);

    Erfc.from=Erfc.tab+ifrom; Erfc.to=Erfc.tab+ito;

#ifdef eropt
    /* very inefficient way how to improve the h-spline accuracy
       useless for large range !!! */
    do {
      r=r0;
      q=1; f=eerfcs(q,ifrom,ito);

      while (fabs(r)>1e-7) {
        r *= -0.6;
        if ( (ff=eerfcs(q+r,ifrom,ito)) < f ) { q += r; f = ff; } }
      r0 *= 10; if (r0>1) Error("eropt failed");
    } while (fabs(q-1)>r0);
    if (f != ff) f=eerfcs(q,ifrom,ito);
#else
    eerfcs(1.0,ito);
#endif

    /***** check accuracy *****/
    ifrom *= Erfc.ndiv; ito *= Erfc.ndiv;
    /* ifrom,ito are in units of h */
    for (i=ito-1; i>=ifrom; i--) {
      erfcs(i,&e,&z);
      x=i*h; ee=eru(x)-e; zz=erd(x)-z;
      if (i==ito-1)
        if (verbose) prt("eru(%f)=%e  erd(%f)=%e",x,e*AuBd,x,z*Ad);
      if (ee<mine) mine=ee; if (ee>maxe) maxe=ee;
      if (zz<minz) minz=zz; if (zz>maxz) maxz=zz;
      ee /= e; zz /= z;
      if (ee<minre) minre=ee; if (ee>maxre) maxre=ee;
      if (zz<minrz) minrz=zz; if (zz>maxrz) maxrz=zz; }

    if (verbose) {
      prt("abs err:  eru in (%.3e,%.3e)  erd in (%.3e,%.3e)",
          mine*AuBd,maxe*AuBd, minz*Ad,maxz*Ad);
      prt("rel err:  eru in (%.3e,%.3e)  erd in (%.3e,%.3e)",
          minre,maxre, minrz,maxrz); }

#ifdef erinterr
    /***** error of integration - awfully slow *****/
    Erfc.ndiv *= 2;
    erfcs(ito*2,&ee,&zz);
    erfcs(ifrom*2,&ee,&zz);
    prt("halving int step: rel err of eru = %.3e   of erf = %.3e",
        (e-ee)/ee,(z-zz)/zz);
    Erfc.ndiv /= 2;
#endif

#ifdef erminmax
    Erfc.min=1e30,Erfc.max=0;
#endif

    /* rescaling */
    loop (er_p,Erfc.tab,Erfc.to) {
      er_p->Ad *= Ad;   er_p->Au *= AuBd;
      er_p->Bd *= AuBd; er_p->Bu *= Bu;
      er_p->Cd *= Cud;  er_p->Cu *= Cud; }

    Erfc.sgrid *= Sqr(alpha);

    z=Sqr(cutoff);

    if (shift&1)  {
      x=eru(z);
      loop (er_p,Erfc.tab,Erfc.to) er_p->Au -= x; 
      prt("eru(r) shifted by %g to avoid jump in elst energy",x); }

    if (shift&2)  {
      x=erd(z);
      loop (er_p,Erfc.tab,Erfc.to) er_p->Ad -= x; 
      prt("erd(r) shifted by %g to avoid jump in elst forces",x); }

    if (shift&4) 
#  include "ertest.c"
  }
} /* initerfc */

#ifdef erexact_eps
/***
    exact but slow variant of erud w/o direct elst term:
    - value of eru minus direct elst term returned,
    - erd minus direct elst term passed,

    exacterud_sqrt(x) = beta*alpha*[e(sqrt(x)*alpha)-1/(sqrt(x)*alpha)]
    exacterud_sqrt_1(x) = beta*alpha*[e(sqrt(x)*alpha)-1/(sqrt(x)*alpha)-2/sqrt(pi)]

    erd = -beta*alpha^3*[e'(sqrt(x)*alpha)+1/(sqrt(x)*alpha)^2]/(sqrt(x)*alpha)

    see above for e, e'

    efficient for low xx only -- to be used as Ewald correction for
    intramolecular charges close together, e.g. if a dipole is
    approximated by two point charges very close together.  Then the
    number of terms in the series is low and exacterud works reasonably
    fast.
***/

double exacterud_sqrt(double xx,erreal *erd)
{
  int n=1;
  double su=0,sd=0,q,x;

  xx*=Erfc.alphaq;

  q=1;

  while (fabs(q)>erexact_eps) {
    /*.....  if (fabs(q)>1e5) { Error("exacterud"); }*/
    x=q/(2*n+1);
    sd+=x*n;
    su+=x;
    q*=-xx/++n; }

  *erd=Erfc.B*sd;

  return Erfc.A*(1-su*xx);
}

double exacterud_sqrt_1(double xx,erreal *erd)
{
  int n=1;
  double su=0,sd=0,q,x;

  xx*=Erfc.alphaq;

  q=1;

  while (fabs(q)>erexact_eps) {
    /*.....  if (fabs(q)>1e5) { Error("exacterud"); }*/
    x=q/(2*n+1);
    sd+=x*n;
    su+=x;
    q*=-xx/++n; }

  *erd=Erfc.B*sd;

  return Erfc.A*(-su*xx);
}
#endif
