/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
              initializing tables of species, sites, potentials
             molecule descriptors, configuration, and matrices M
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include "ground.h"
#define SDS
#include "alloc.h"
#include "simglob.h"
#include "interpot.h"
#include "simdef.h"
#include "siminit.h"
#include "simils.h"
#include "simpot.h"
#include "units.h"
#include "simcg.h"
#include "cputime.h"

static double machprec(void)
/*
  worst case one FPU operation relative resolution (error=1/2)
  see also soft/nsk/nsksys.c
*/
{
#define NN 999
  real x[NN];
  real err=0,d;
  int i;

  loop (i,0,NN) x[i]=(real)(i+2)/(2*i+3);
  loop (i,0,NN) {
    d=fabs(x[i]*(2*i+3)-(i+2))/(2*i+3);
    Max(err,d) }

  i=-log(d)/log(2);
  return powi(2,-i);
#undef NN
}

void initNo(void) /************************************************** initNo */
/*
  calculates No.s, No.N, and No.mass
  to be called (in pass=1) after spec[sp]->N and spec[sp]->ns have been read
  from sysname.def
*/
{
  int sp,nd;
  int downN=option('j'),jN,nq=0,nnq;
  double molmass,molcharge,sumq=0,molq,err;

#  ifdef POLAR
  double molalpha;
  int npol;

  No.A=0;
#  endif /*# POLAR */

  No.s=No.free_s=No.N=No.free_N=No.depend[0]=No.free_depend=0;
  No.mass=No.free_mass=0;
  No.charge=0;

  No.eps=machprec();
  No.epsq=pow(No.eps,1.5);
  prt("machine precision=%.3g, charge neutrality precision=%.3g",No.eps,sqrt(No.epsq));

  loop (sp,0,nspec) {
    int ns=spec[sp]->ns;
    int N=spec[sp]->N,i;
    depend_t *d;

#  ifdef POLAR
    molalpha=0;
    npol=0;
#  endif /*# POLAR */
    molmass=molcharge=molq=0;

    loop (i,0,ns) {
      siteinfo_t *si=spec[sp]->si+i;
#  ifdef POLAR
      sitedef_t *sd=sitedef+si->st;
      molalpha += sd->alphapol;
      if (sd->alphapol) npol++;
      molcharge += si->chargepol;
#  endif /*# POLAR */
      molcharge += si->charge;
      if (si->charge) nq+=N;
      molq += Sqr(si->charge);
      //      molmass += sd->mass; 
      molmass += si->mass; 
    }

#  ifdef POLAR
    No.A += molalpha*N;
    No.pol += npol*N;
#  endif /*# POLAR */
    No.mass += molmass*N;
    No.charge += molcharge*N;
    sumq += N*molq;
    No.s += ns*N;
    No.N += N;

    jN = N<downN ? 0 : downN>0 ? N-downN : N;
    No.free_N += jN;
#  ifdef POLAR
    No.free_pol += npol*N;
    /* there should be also  No.free_A ... */
#  endif /*# POLAR */
    No.free_s += ns*jN;
    No.free_mass += molmass*jN;

    /* NEW: dependants */
    nd=0;
    for (d=spec[sp]->dependants; d; d=d->next) nd++;
    if (nd && fixsites0) WARNING(("dependants + fixed sites"))
    No.depend[0] += nd*N;
    No.free_depend += nd*jN;

    downN -= N; } /* loop sp */

  prt("No.N=%d molecules  No.s=%d sites (of these, %d dependants)",
            No.N,              No.s,               No.depend[0]);
  prt("total mass of configuration M = %.4f g/mol = %g kg",
                                   No.mass*munit,No.mass*massunit);
  prt("total charge Q = %g prog.u. = %g e",No.charge,No.charge/electron);

  put3(No.depend[1],No.depend[2],No.depend[3])

  err=Sqr(No.charge)/sumq;
  if (err>No.epsq && option('q')>0)
    WARNING(("the system is charged (%g e, rel.err=%g): consider option -q-%d",
             No.charge/electron,sqrt(err),option('q')))

  if (option('q')<0 && nq) {
    double dq=-No.charge/nq;
    
    nnq=0;

    prt("WARNING: neutralizing system charge: %d charges shifted by %g e",
      nq, dq/electron);

    loop (sp,0,nspec) {
      int ns=spec[sp]->ns,i;

      loop (i,0,ns) if (spec[sp]->si[i].charge) {
        spec[sp]->si[i].charge+=dq; 
        nnq+=spec[sp]->N; } }
    if (nq!=nnq) ERROR(("internal: # of charges changed %d -> %d",nq,nnq)) }

  /* 
     counting ions (after optional neutralizing system charge)
     the original No.charge is kept (the new should be zero)
     BUG: zero charge and nonzero Drude charge not treated correctly
  */
  nnq=0;
  No.ion=0;
  molq=0; /* used for total charge */
  loop (sp,0,nspec) {
    int ns=spec[sp]->ns;
    int N=spec[sp]->N,i;

    molcharge=0;

    loop (i,0,ns) {
      siteinfo_t *si=spec[sp]->si+i;
#  ifdef POLAR
      molcharge += si->chargepol;
#  endif /*# POLAR */
      molcharge += si->charge;
      if (si->charge) nnq+=N; }

    molq+=N*molcharge;

    if (fabs(molcharge)>=0.01*electron) No.ion+=N; }

  prt("number of ions No.ion=%d",No.ion);

  if (nq!=nnq) ERROR(("internal: # of charges changed %d -> %d",nq,nnq))

  if (Sqr(molq)/sumq>Sqr(No.epsq) && option('q')<=0)
    WARNING(("the system is charged (%g e) (after fixup)",No.charge/electron))

  if (FROM>0)
    prt("free %d molecules  %d sites (%d dependants) %.4f g/mol",
	No.N-option('j'),No.free_s,No.free_depend,No.free_mass*munit);
  if (No.N<=0) { ERROR(("no molecule - nothing to do")) exit(-1); }
  if (option('j')>=No.N) {
    ERROR(("option -j%d > # of molecules N=%d",option('j'),No.N))
    option('j')=No.N; }
#  ifdef POLAR
  prt("total polarizability = %g A^3",No.A);
#  endif /*# POLAR */
}

void makemolgol(int nmol) /************************************** makemolgol */
{
  int sp,err=0,nn;
  char cmd[2048];
  char *nm=simils.simname;

  if (nmol<=0) return;
  Min(nmol,No.N)

  strcpy(cmd,"molcfg ");
  loop (sp,0,nspec) {
    if (!strcmp(spec[sp]->name,simils.simname)) err++;
    nn=min(nmol,spec[sp]->N);
    if (nn) sprintf(strend(cmd),"-%d:%c%%03d %s ",
                    nn, sp+'a', spec[sp]->name);
    nmol-=nn; }

  if (err) {
    WARNING(("simulation name %s matches (one of) species name(s)\n\
*** CONFIG.mol and .gol will be generated instead of %s.mol",simils.simname,simils.simname))
    nm="CONFIG"; }

  strcat(cmd,nm);
  underline("show support");
  prt("Command to generate .mol and .gol files for show will be called:\n  %s",cmd);
  putenv("MOLCFG=1");
  sp=system(cmd);
  if (sp) WARNING(("%s.{mol,gol} not generated\n\
*** (missing .mol and .gol files for species?)",nm))
  else
    prt("Files %s.{mol,gol} have been generated.\n\
To show the trajectory, use command\n  show %s", nm,nm);
}

void initpot(void) /*********************************************** initpot */
{
  int i,j;

#ifdef LINKCELL
  ralloc(pot1,nspec*sizeof(pot1_t*));
#endif /*# LINKCELL */
  ralloc(pot,nspec*sizeof(pot2_t**));
  loop (i,0,nspec) {
#ifdef LINKCELL
    pot1[i]=charmm1; /* undefined for PEO */
#endif /*# LINKCELL */
    ralloc(pot[i],nspec*sizeof(pot2_t*));
    /* old COOK: charmm2; */
    loop (j,0,nspec) pot[i][j]=Potential(i,j); }
}

void initmolecules(void) /*********************************** initmolecules */
/*
  allocate configurations a[] with pointers to molecules
  allocate M for constraint dynamics
  initialize degrees of freedom
*/
{
  int sp,n=0,irp=0,i,Mlen=0,sub,tsub;

  if (RPOFFSET!=(real*)(cfg[0]->rp)-(real*)&(cfg[0]->logs))
    ERROR(("wrong RPOFFSET, real padding in structure ToInt, or other declaration problems"))

  No.c=No.maxc=No.maxs=No.free_c=0;

  rallocarrayzero(molec,No.N); /* 11/2004 because of ANCHOR */
  ralloc(Ms,(option('c')&1?No.N:nspec)*sizeof(M_t*));

  underline("initializing molecules");
  if (option('v')&2) header("species     N     ns     nc   Mlen");
  loop (sp,0,nspec) {
    specinfo_t *s=spec[sp];

    if (s->N) {
      if (!(option('c')&1)) Ms[sp]=defineM(s,&Mlen);

      loop (i,0,s->N) {
        molecule_t *m=molec+n;

        /* INEFFICIENT! should allocate (by molecules) in advance/bigger chunks */
        if (option('c')&1) Ms[n]=defineM(s,&Mlen);
        m->sp=sp;
        m->ns=s->ns;
        m->nc=s->nc;
        m->ir=sizeof(vector)*irp;
        m->ig=No.c;

        Max(No.maxs,m->ns)
        irp += m->ns;
        Max(No.maxc,m->nc)
        No.c += m->nc;
        if (n>=option('j')) No.free_c += m->nc;
        n++; } }
    if (option('v')&2)
      prt("%6d %6d %6d %6d %6d", sp,s->N,s->ns,s->nc,Mlen); }
  if (option('v')&2) header("");

  if (irp!=No.s) ERROR(("internal: irp=%d != No.s=%d",irp,No.s))
  No.nreal=No.s*DIM;
#ifdef POLAR
  No.nreal*=2;
#endif /*# POLAR */

  No.eq=No.s*DIM+RPOFFSET; /* # of eqs of motion (incl. logs, lambda) */ 

/* DEGREES OF FREEDOM
   ^^^^^^^^^^^^^^^^^^
   No.f = DIM*No.s
         -No.c  constraints
         +1     additional degree of freedom for Nose
         -1     energy (Hamiltonian) conservation
         -DIM*(# of dependants) for all sites lin. dependent on other sites
         -(fixed degrees from drift: lin and ang momentum components)

   WARNING: not correct for a few fixed sites

   Intramolecular degrees of freedom (No.f_in) contain all moves relative to
   the center of mass (incl. rotations).

   Intermolecular degrees of freedom (No.f_tr) contain only translational
   moves of centers of mass and the above corrections apply BUT -1 for energy
   conservation if no Nose thermostat is used.

   The Maxwell-Boltzmann, Andersen, (Langevin) thermostats sample also translations and rotations.
*/

  sub=0;
  if ((drift&DRIFT_WHEN)!=DRIFT_START)
    for (i=DRIFT_VX; i<=DRIFT_AZ; i*=2) if (drift&i) sub++;
  prt("Number of conserved degrees of freedom = %d\n\
(includes conserved momenta and angular momenta, as calculated from drift=%d;\n\
Hamiltonian conservation is treated separately).",sub,drift);
  if (conserved<0)
    prt("\
This value will be subtracted from the total number of degrees of freedom as\n\
well as intermolecular (based on center-of-mass) number of degrees of freedom.");
  else {
    if (sub!=conserved)
      prt("WARNING: specified conserved=%d does not match the calculated value");
    sub=conserved; }

  tsub=0; /* for Nose, Maxwell/Andersen */
  if (thermostat<=T_BERENDSEN || (thermostat>=T_TR && thermostat<=T_FRICTIONS)) {
    tsub=1; /* energy conservation */
    prt("One degree of freedom subtracted from total because of energy conservation.");
    if (thermostat)
      prt("In Berendsen and other friction-like thermostats, this is approximate.");
    prt("\
The translational temperature (from centers-of-mass) Ttr should be the same as\n\
the intramolecular and rotational temperature Tin for an equilibrated system,\n\
but both may sligtly differ from the total kinetic temperature T"); }

  No.f4P=DIM*(No.free_s-No.free_depend)-No.free_c;
#ifdef POLAR
  if (option('p')/10%10==0) {
    /* Car-Parrinello-like */
    No.f4P+=DIM*No.free_pol;
    prt("Car-Parrinello-like: %d*%d mechanical degrees of freedom for dipoles added",
	No.free_pol,DIM); }
#endif /*# POLAR */
  No.f=No.f4P-sub;
  /* No.f=degrees of freedom without energy conservation */

  put3(No.free_s,No.free_c,No.free_depend)
  if (No.f<=0) {
    WARNING(("%d degrees of freedom: replaced by 1",No.f))
    No.f=1; }

  /* inter- and intramolecular degrees of freedom (without energy conservation) */
  No.f4P_tr=DIM*(No.free_N);
  No.f_tr=No.f4P_tr-sub;
  No.f_in=No.f-No.f_tr;

  No.f-=tsub; /* (rather heuristic or approximate) correction for energy conservation */

  No.conserved=sub+tsub;

  put3(No.s,No.c,No.f)
  put3(No.s*3,No.eq,No.nreal)
  put3(No.free_N,No.f_tr,No.f_in)
  put3(No.conserved,No.f4P,No.f4P_tr)

  No.NPT=1+3./No.f;
  No.fx=No.f;

  if (thermostat==T_NPT) {
    if (rescale&RESCALE_CM) {
      ERROR(("rescale=%d includes CM-based rescaling, which is not supported with NPT/MTK\n\
***   for the isotropic barostat: use rescale=\"xyz\"\n\
***   to maintain Pzz=P:  use rescale=\"Z\", etc.",rescale))
      rescale-=RESCALE_CM; }
    if (tau.P<.999e6) {
      if (rescale&RESCALE_PT) No.fx+=No.ncoord;
      else No.fx++; }
    else
      WARNING(("large tau.P>=1e6 forces Nose (no extra degrees of freedom added)\n\
*** but term PV is still included in total energy"))
    No.M_T=Sqr(tau.T)*T*No.f;
    No.M_Th=No.M_T/2;
    No.M_P=Sqr(tau.P)*T*(No.f+3)*No.ncoord;
    No.M_Ph=No.M_P/2;
    put3(No.ncoord,No.fx,No.NPT)
    put2(No.M_T,No.M_P) }

/*** physical configuration allocation ***/
#  ifdef POLAR
  i=sizeof(ToInt)-sizeof(vector)+No.s*2*sizeof(vector);
  polar_off=No.s*sizeof(vector);
#  else /*# POLAR */
  i=sizeof(ToInt)-sizeof(vector)+No.s*sizeof(vector);
#  endif /*#!POLAR */
  loop (n,0,option('m')+3) sdsralloczero(cfg[n],i)
} /* initmolecules */


void initgammas(void) /****************************************** initgammas */
{
  int gssize=No.c*sizeof(gs[0][0])+(No.c==0);

  if (No.pred) {
    int i; int b= -1;

    ralloc(gs,No.pred*sizeof(gs[0]));
    ralloc(binom,No.pred*sizeof(binom[0]));
    loop (i,0,No.pred) {
      ralloczero(gs[i],gssize);
      binom[i] = b= -b*(No.pred-i)/(i+1); } }
} /* initgammas */


double cutcorr(double LJcutoff) /*********************************** cutcorr */
/***
    Standard bulk fluid cutoff correction of pair Lennard-Jones-like
    forces over all intermolecular pairs is calculated.
    Intramolecular pairs are not included!
    The result should be divided by V to obtain cutoff correction of Epot
    The result should be divided by V^2 to obtain cutoff correction of P
    In addition, # of pairs for rdf histograms are calculated
    (incl. intramolecular pairs more distant than 1--4)
    (elst forces not included)
***/
{
  int n,m,i,j,si,sj;
  double np;
  double c,corr=0;
  rdf_t *rdfij;

  if (LJcutoff==0) return 0;

  loop (n,0,nspec) {
    loopto (m,0,n) {
      if (m==n) np=(double)spec[n]->N*(double)(spec[n]->N-1)/2;
      else np=(double)spec[n]->N*spec[m]->N;
      c=0;
      loop (i,0,spec[n]->ns) loop (j,0,spec[m]->ns) {
#ifdef COOK
	c += sstab [si=spec[n]->si[i].st] [sj=spec[m]->si[j].st] .corr;
#else /*# COOK */
	{
	  sitesite_t dummy;
	  c += setss(&dummy,si=Site(n,i),sj=Site(m,j),LJcutoff);
	}
#endif /*#!COOK */
	if ( (rdfij=rdf[si][sj]) ) rdfij->npair+=np;
        /* this is OK because rdf[si][sj] and rdf[sj][si]
           point to the same rdf_t */ }
      corr+=c*np; }

#ifdef COOK
    /* now intramolecular pairs but 1-2,1-3, and 1-4 */
    /* cutoff corr not included because it is not clear how */

    loop (i,0,spec[n]->ns) {
      exception_t *exc=spec[n]->si[i].exc;
      int j0=0,j1;

      do {
	j1=exc->indx;

	if (exc->type==ONEFOUR)
	  if ( (rdfij=rdf[spec[n]->si[i].st][spec[n]->si[j1].st]) )
	    rdfij->npair += spec[n]->N;

	loop (j,j0,j1)
	  if ( (rdfij=rdf[spec[n]->si[i].st][spec[n]->si[j].st]) )
	    rdfij->npair += spec[n]->N;

	exc++;
      } while ( (j0=j1+1)<i );
    }
#endif /*# COOK */
  } /* n */

#ifdef DEBUG
  {
    int n,m,sp;
    molecule_t *mn,*mm;
    double debugcorr=0;

    measure=2;

    loop (n,0,No.N) {
      mn=molec+n;
      sp=mn->sp;

      loop (m,0,n) {
	mm=molec+m;
	debugcorr += (*pot[sp][mm->sp]) (NULL,NULL,mn,mm,a[0]->rp); }

      (*pot[sp][mn->sp]) (NULL,NULL,mn,mn,a[0]->rp); /* ugly, but it works */ }

    put2(corr,debugcorr)
    if (fabs(corr/debugcorr-1)>1e-10) Error("");
  }
#endif /*# DEBUG */
  return corr;
} /* cutcorr */

#ifdef WIDOM
double Widomcutcorr(int spreal,int spvirt,double LJcutoff) /****************** Widomcutcorr */
/***
    Standard bulk fluid cutoff correction of pair Lennard-Jones-like
    forces over intermolecular pairs of virtually inserted molecule of
    species spvirt is calculated.
    If spreal>=0, the analogous term for species spreal is subtracted
    so that the correction for spreal->spvirt identity change is returned.
    Intramolecular pairs are not included!
    The result should be divided by V to obtain the cutoff correction
    of the chemical potential by the Widom method.
***/
{
  int m,i,j,si,sj,np;
  double c,corr=0;

  if (spvirt>=nspec) ERROR(("widom.sp undefined (>=nspec)"))
  if (spreal>=nspec) ERROR(("widom.spreal undefined (>=nspec)"))

  loop (m,0,nspec) {
    np=spec[m]->N;
    c=0;
    loop (i,0,spec[spvirt]->ns) loop (j,0,spec[m]->ns)
      c += sstab [si=spec[spvirt]->si[i].st] [sj=spec[m]->si[j].st] .corr;
    corr+=c*np; }

  if (spreal>=0) loop (m,0,nspec) {
    np=spec[m]->N;
    if (m==spreal) np--;
    c=0;
    loop (i,0,spec[spreal]->ns) loop (j,0,spec[m]->ns)
      c += sstab [si=spec[spvirt]->si[i].st] [sj=spec[m]->si[j].st] .corr;
    corr-=c*np; }

  if (spreal>=0)
    prt("\nWidom cutoff corection for species id-change [%d]->[%d] for a homogeneous box",
	spreal,spvirt);
  else
    prt("\nWidom cutoff corection for species [%d] for a homogeneous box",spvirt);

  prt("corr = %g/V = %g\n\
Widom correcting Boltzmann factor = %g\n\
(the latter two values will change if the box size changes)\n",
      corr,corr/(L[0]*L[1]*L[2]),
      exp(-corr/(L[0]*L[1]*L[2]*T)));

  return corr;
} /* Widomcutcorr */
#endif /*# WIDOM */

double setL(vector L,double rho) /************************************* setL */
/*
  rescale L, or calculate missing L[] from rho given
  returns final rho
*/
{
  int i,nzero=0;
  double x,p=1;

  loop (i,0,3) if (L[i]<0) ERROR(("negative L[%d]=%g",i,L[i]))

  if (rho*(L[0]*L[1]*L[2])!=0) {
    /* both rho[initrho] and L given: scaling */
    x=cubrt(No.mass*rhounit/(L[0]*L[1]*L[2])/rho);
    if (fabs(x-1)>5e-16) {
      prt("Current L=(%g %g %g) will be rescaled to reach %g kg/m3.",VARG(L),rho);
      VO(L,*=x) } }
  else {
    loop (i,0,3) if (L[i]==0) nzero++; else p*=L[i];

    if (nzero==0) {
      /* all L[] given ==> calculate rho */
      rho=No.mass*rhounit/(L[0]*L[1]*L[2]);
      prt("Density %g kg/m3 was calculated from box sides",rho); }
    else {
      if (rho==0) {
#ifdef FREEBC
        WARNING(("The reference density was set to rho=1000\n\
*** It is used by the initializer only and does not affect calculations"))
#else /*# FREEBC */
        ERROR(("I cannot calculate the reference box because both density (rho) and\n\
*** %d box sides (L[]) are undefined (zero).\n\
*** (This is no longer warning to force users to specify the density;\n\
*** nevertheless, ignoring this error will continue as if rho=1000)",nzero))
#endif /*#!FREEBC */
        rho=1000; }
      /* calculate missing L[] from rho */
      x=pow(No.mass*rhounit/rho/p,1./nzero);
      loop (i,0,3) if (L[i]==0) L[i]=x;

      prt("%d missing box sides calculated from density rho=%g kg/m3",
           nzero,                                           rho); } }

  prt("Density doublecheck (calculated from new L[]): current rho=%g kg/m3",
    No.mass*rhounit/(L[0]*L[1]*L[2]));

  return rho;
}

double initcutoff(double cutoff,vector L) /********************** initcutoff */
/*
  L is the reference L
*/
{
  double minL=9e99;
  int i;

  prt("Truncation of electrostatics forces and maximum cutoff for LJ and RDF:");

#ifdef FREEBC
  if (cutoff<=0) cutoff=100;
  prt("cutoff = %g A",cutoff);
#else /*# FREEBC */

  loop (i,0,DIM) Min(minL,L[i])

  if (cutoff<=0) {
#  ifdef NIBC
    cutoff=9e9;
#  elif COULOMB<0
    /* Ewald: L/2 break-even for 1000 atoms (changed from 2000 in V2.7d)*/
    cutoff=cubrt(L[0]*L[1]*L[2]/8)*pow(No.s/1000.0,-1.0/6);
    prt("default cutoff for Ewald summation estimated:\n\
WARNING: may be far from optimum! (consider the test module, see el.test)");
#  else /*#!NIBC!COULOMB<0 */
    cutoff=12;
    prt("default cutoff for short-range electrostatics set:");
#  endif /*#!NIBC!COULOMB<0 */
    Min(cutoff,minL/2) }

  prt("cutoff = %.8f  cutoff/min(L)=%g",cutoff,cutoff/minL);

  loop (i,0,DIM)
    if (cutoff>L[i]/2)
      prt("WARNING: cutoff=%g > L%c/2=%g (by %.3g%%)",
                       cutoff,'x'+i,L[i]/2,(cutoff/(L[i]/2)-1)*100.);
#endif /*#!FREEBC */
  return cutoff;
}

/*+++++++++++++++++++++++++ fixed sites support ++++++++++++++++++++++++++++*/

int initfix(char *fn) /******************************************** initfix */
{
  static int from;

  if (option('k')>0) {
    FILE *fs=fopen(fn,"rt");
    char line[128],*s;
    fixsites_t *f;
    int i,to,isum=0;

    if (!fs) {
      WARNING(("no %s file -- no sites kept fixed, option -k ignored",lastFn))
      option('k')=-9;
      return 0; }

    if (FROM)
      ERROR(("combination of option -j and file %s not supported",lastFn))

    while (fgets(line,128,fs)) if (!strchr("!#",line[0])) {
      s=strtok(line," \t\n"); if (!s) continue;
      if (!isdigit(s[0])) ERROR(("%s: %s not a number",lastFn,s))
      from=atoi(s);
      to=from;
      s=strchr(s,'-');
      if (s) sscanf(s+1,"%d",&to);
      if (to<from || from<0)
        ERROR(("%s: %d-%d bad order or negative site",lastFn,from,to))

      alloczero(f,sizeof(fixsites_t));
      f->from=from;
      to++; /* last not incl. */
      f->to=to;

      while ( (s=strtok(NULL," \t\n")) && f->isr<3)
	f->r[f->isr++]=atof(s);
      if (f->isr%DIM)
	ERROR(("%s: format of optional site %d position",lastFn,from))

      loop (i,from,to)
        if (isfixed(i)) prt("%s: %d multiply listed",lastFn,i);
      else
        isum++;
      f->next=fixsites0; fixsites0=f; }

    fclose(fs);
    if (!isum) {
      WARNING(("%s: no site kept fixed",lastFn))
      from=-1;
      return 0; }
    prt("%s: %d sites kept fixed",lastFn,isum); }

  return from;
}

int isfixed(int i) /*********************************************** isfixed */
{
  fixsites_t *f;

  for (f=fixsites0; f; f=f->next) if (i>=f->from && i<f->to) return 1;

  return 0;
}

void checkfixed(void) /***************************************** checkfixed */
{
  fixsites_t *f;

  for (f=fixsites0; f; f=f->next) if (f->to>No.s)
    ERROR(("fixed site %d out of range",f->to))
}

#ifdef ANCHOR
#  define ALINELEN 1024
void initanchor(char *fn,int init,double drmax) /*************** initanchor */
{
  FILE *fs;
  char line[ALINELEN];
  int n,iaxis[3];
  vector r;
  molecule_t *mn;
  static int keep=0;
  int STAT[32],i;

  loop (i,0,32) STAT[i]=0;

  prt("reading %s",fn);

  loop (n,0,No.N) molec[n].anchor=0;

  if (option('k')!=0) return;

  if ( !(fs=fopen(fn,"rt")) ) {
    WARNING(("no %s file -- no sites kept fixed, option -k ignored",lastFn))
    option('k')=-9;
    return; }

  while (fgets(line,ALINELEN,fs)) if (isalpha(line[0])) {
    char *c=line;
    int anchor=0;

    while (isalpha(*c)) c++;

    iaxis[0]=iaxis[1]=iaxis[2]=0;

    STAT[((unsigned char*)line)[0]&31]++;

    switch (line[0]) {
      case 'v': anchor=ANCHOR_v; goto readn;
      case 'r': anchor=ANCHOR_r; goto readn;
      case 'x': anchor=ANCHOR_x; goto readn;
      case 'f': anchor=ANCHOR_f; goto readn;
      case 'm':	anchor=ANCHOR_m;
      readn:
        if (1!=sscanf(c,"%d",&n))
	  ERROR(("%s: missing molecule number in line:\n%s",lastFn,line))
	break;
      case 's':
        anchor=ANCHOR_s; goto s;
      case 'a':
        anchor=ANCHOR_a;
      s:
        if (5!=sscanf(c,"%d%d%lf%lf%lf",&n,iaxis,r,r+1,r+2))
          ERROR(("%s: not enough data in line:\n%s",lastFn,line))
        break;
      case 'p':
        anchor=ANCHOR_a;
        if (6!=sscanf(c,"%d%d%d%lf%lf%lf",&n,iaxis,iaxis+1,r,r+1,r+2))
          ERROR(("%s: not enough data in line:\n%s",lastFn,line))
        break;
      case 't':
        anchor=ANCHOR_t;
        if (7!=sscanf(c,"%d%d%d%d%lf%lf%lf",&n,iaxis,iaxis+1,iaxis+2,r,r+1,r+2))
          ERROR(("%s: not enough data in line:\n%s",lastFn,line))
        break;
      case 'c':
        anchor=ANCHOR_c; goto c;
      case 'i':
        anchor=ANCHOR_i;
      c:
        if (4!=sscanf(c,"%d%lf%lf%lf",&n,r,r+1,r+2))
        ERROR(("%s: not enough data in line:\n%s",lastFn,line))
        break;
      case 'g': {
        struct anchorgroup_s *group;
        char *tok=strtok(line," \t\n");
        int i,ns;

        tok=strtok(NULL," \t\n"); if (!tok) ERROR(("%s: group: not enough data",lastFn))
        n=atoi(tok);
        if (n<0 || n>=No.N)
          ERROR(("%s: group: %d is invalid molecule number",lastFn,n))
        mn=molec+n;
        tok=strtok(NULL," \t\n"); if (!tok) ERROR(("%s: group: not enough data",lastFn))
        ns=atoi(tok);
        if (ns<1 || ns>=mn->ns)
          ERROR(("%s: group: ns=%d out of range",lastFn,ns))

        anchor=ANCHOR_g;

        ralloc(group,sizeof(*group)+(ns-1)*sizeof(group->site[0]));
	group->ns=ns;
        loop (i,0,ns) {
          tok=strtok(NULL," \t\n"); if (!tok) ERROR(("%s: group: not enough data",lastFn))
          group->site[i]=atoi(tok);
          if (group->site[i]<0 || group->site[i]>=mn->ns)
            ERROR(("%s: group: site %d out of range",lastFn,group->site[i])) }

        loop (i,0,DIM) {
          tok=strtok(NULL," \t\n"); if (!tok) ERROR(("%s: group: not enough data",lastFn))
          group->r0[i]=atof(tok); }

        group->next=mn->group;
        mn->group=group; }
	break;
      default:
        ERROR(("%s: wrong key in line:\n%s",lastFn,line)) }

    if (n<0 || n>=No.N)
      ERROR(("%s: n=%d is invalid molecule number",lastFn,n))

    mn=molec+n;

    if ( (anchor & ANCHOR_g) && (mn->anchor & (ANCHOR_pos|ANCHOR_axis))
     ||  (mn->anchor & ANCHOR_g) && (anchor & (ANCHOR_pos|ANCHOR_axis)) )
      ERROR(("%s: constrain group and axis/cm/site combined for molecule %d",lastFn,n))
    if ( (anchor & ANCHOR_pos) && (mn->anchor & ANCHOR_pos) )
      ERROR(("%s: constrain position doubly defined for molecule %d",lastFn,n))
    if ( (anchor & ANCHOR_axis) && (mn->anchor & ANCHOR_axis) )
      ERROR(("%s: constrain direction doubly defined for molecule %d",lastFn,n))

    mn->anchor|=anchor;

    if ( (mn->anchor & ANCHOR_s) && (mn->anchor & ANCHOR_axis) )
      ERROR(("%s: combination of site and axis constraining for molecule %d",lastFn,n))

    if (anchor & ANCHOR_pos) {
      if (ANCHOR_s) mn->iaxis[0]=iaxis[0];
      VV(mn->r0,=r) }
    else if (anchor & ANCHOR_axis) {
      double x=sqrt(SQR(r));
      if (x==0)
        ERROR(("%s: zero vector for molecule %d",lastFn,n))
      copy(mn->iaxis,iaxis,sizeof(iaxis));
      VVO(mn->axis,=r,/x) }
    }

  fclose(fs);

  /* setting bounds to moves caused by constraints - cheap */
  anchor.rr=Sqr(drmax)/40;
  anchor.sin=sqrt(anchor.rr);
  anchor.cos=1-anchor.rr;

  if (drmax==0 || (drmax>0 && init<=2)) {
    anchor.rr=9e99;
    anchor.sin=2; anchor.cos=-2; }

  n=0;
  loop (i,0,32) n+=STAT[i];
  if (n) {
    prt_("ANCHOR:");
    loop (i,0,32) if (STAT[i]) prt_(" %c=%d",i+'`',STAT[i]);
    _n }

  anchor.f=fopen(Fn("anc"),init<2?"at":"wt");
  if (!anchor.f) ERROR(("cannot write to %s",lastFn))

  if (init>=2) {
    int n,col=0;

    fprintf(anchor.f,"# UNITS: [force]=N, [moment of force (torque)]=N.m, other based on [length]=A, [time]=ps\n#");
    loop (n,0,No.N) {
      molecule_t *mn=molec+n;

      if (mn->anchor & ANCHOR_g) {
	struct anchorgroup_s *group;
	looplist (group,mn->group) {
	  fprintf(anchor.f,"--constr.f. mol.%3d group ns=%d(%2d..)---|",n,group->ns,group->site[0]);
	  col+=3; } }
      if (mn->anchor & ANCHOR_r) fprintf(anchor.f,"-------- position molecule%3d ---------|",n),col+=3;
      if (mn->anchor & ANCHOR_v) fprintf(anchor.f,"-------- velocity molecule%3d ---------|",n),col+=3;
      if (mn->anchor & ANCHOR_x) fprintf(anchor.f,"------ acceleration molecule%3d -------|",n),col+=3;
      if (mn->anchor & ANCHOR_f) fprintf(anchor.f,"---------- force molecule%3d ----------|",n),col+=3;
      if (mn->anchor & ANCHOR_m) fprintf(anchor.f,"-------- momentum molecule%3d ---------|",n),col+=3;

      if (mn->anchor & ANCHOR_s) fprintf(anchor.f,"---- constr.force mol.%3d site %3d ---|",n,mn->iaxis[0]),col+=3;
      else if (mn->anchor & ANCHOR_c)
        fprintf(anchor.f,"- constr.force mol.%3d center-of-mass -|",n),col+=3;
      if (mn->anchor & ANCHOR_axis)
        fprintf(anchor.f,"- constr.moment mol.%3d center-of-mass |",n),col+=3;
      if (mn->anchor & (ANCHOR_c | ANCHOR_s | ANCHOR_g)) keep++; }

    fprintf(anchor.f,"  time/ps\n#");
    loop (n,1,col+2) fprintf(anchor.f,"     #%-7d%s",n,n%3?"":"|");
    fprintf(anchor.f,"\n"); }

  if (keep)
    WARNING(("check variables `drift' and `conserved'\n\
*** (automatical setup has not been checked/finished)"))
}
#endif /*# ANCHOR */

#if PARALLEL
void initparallel(void) /************************************** initparallel */
{
  int maxth;
  static int pass;

#  if PARALLEL==1
#    define MAXTHREADS 8
  if (No.th<1 || No.th>MAXTHREADS || No.lcth<1 || No.lcth>MAXTHREADS)
    ERROR(("parallel: number of Ewald threads (%d) or x-slab threads (%d)\n\
*** is out of permitted range [1,%d]",No.th,No.lcth,MAXTHREADS))
  prt("PARALLEL=%d (linked-cell list parallelized)\n\
number of Ewald threads=%d  number of x-cell threads=%d",
      PARALLEL,No.th,No.lcth);

  if (No.lcth!=No.th) WARNING(("The number of r-space x-slabs (%d)\n\
*** is not an integer multiple of the number of parallel threads (%d)",No.lcth,No.th))
  maxth=max(No.th,No.lcth);
#  endif /*# PARALLEL==1 */

#  if PARALLEL==2
  if (No.th!=2) ERROR(("PARALLEL==2: the number of threads (%d) is not 2",No.th))
  maxth=2;
#  endif /*# PARALLEL==2 */

/* Thread initialization. Taken from:
**  GNU Pth - The GNU Portable Threads
**  Copyright (c) 1999-2003 Ralf S. Engelschall <rse@engelschall.com>
*/
  if (pass==0) {
    if (pthread_attr_init(&No.thread_attr))
      ERROR(("parallel: pthread_attr_init"))
    if (pthread_attr_setdetachstate(&No.thread_attr, PTHREAD_CREATE_JOINABLE))
      ERROR(("parallel: pthread_attr_setdetachstate")) }

  rallocarray(No.thread,maxth);
  fprintf(stderr,"initparallel: %d threads allocated\n",maxth);
#  ifdef SERIAL
  WARNING(("SERIAL: this is debugging version with all threads run sequentially"))
#  endif /*# SERIAL */
  pass=1;

  if (option('t')) {
#  if PARALLEL==2
    partimes.nrspace=2;
#  else /*# PARALLEL==2 */
    partimes.nrspace=No.lcth;
#  endif /*#!PARALLEL==2 */
    allocarrayzero(partimes.rspace,partimes.nrspace);
    partimes.nkspace=No.th;
    allocarrayzero(partimes.kspace,partimes.nkspace); }
}

#endif /*# PARALLEL */
