/*
  measuring: CP, RDF and dihedral angle distribution (#ifdef DIHHIST)
  dipole moments, radius of gyration
*/

#include <time.h>
#include "ground.h"
#define SDS
#include "alloc.h"
#include "varfile.h"
#include "statics.h"
#include "simglob.h"
#include "simmeas.h"
#include "simdef.h"
#include "simils.h"
#include "pakcp.h"
#include "units.h"
#include "forces.h"
#include "interpot.h"

#ifndef SS_MEASURE_rep
#  define SS_MEASURE_rep /**/
#endif /*# SS_MEASURE_rep */

/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                              convergence profile
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include "cpmark.h"

int NCP;      /* # of items in a record - to grow */

/* OLD VERSION: typedef struct { float Etot,T,Epot,Eintra,P; } CPtype; */

float *CPbuf; /* convergence profile is buffered and stored at once when
                 also .sta and .cfg are stored.  This is useful for
                 recovering the simulations in the case of crash */

/*
  list of variables always recorded in SIMNAME.cp
  this info MUST match with the series of recordCP in main.c
  each item must be 4 chars long (or it must be padded by ' ' to 4 chars)
  first 2 items (namely Etot and T) are NOT included
*/
static char CPinfo0[]=
/* col3col4col5col6col7 */
  "Epot?   ?   Tin Ttr "
#ifdef COULOMB
  "|M| "
#endif /*# COULOMB */
#ifdef SHEAR
  "Cvel"
#endif /*# SHEAR */
#ifdef POLAR
  "self"
#endif /*# POLAR */
#ifdef DIHHIST
  "gauc"
#endif /*# DIHHIST */
#ifdef RGYR
  "Rgyr"
  "ende"
#endif /*# RGYR */
#ifdef XSECTION
  "Xsec"
#endif /*# XSECTION */
#ifdef WIDOM
  "Widm"
#endif /*# WIDOM */
  ;
static char *CPinfo=NULL;

static int CPnbit;

struct ssd_s *ssd0;

static void prtCPinfo(char *info)
{
  int i;

  prt_(" 1=Etot  2=   T");
  loop (i,0,NCP-2) {
    prt_("%c%2d=%c%c%c%c",
         " \n"[i%10==8],i+3,
         info[4*i],info[4*i+1],info[4*i+2],info[4*i+3]); }
  _n
}

#define NENERGY 17
/* first NENERGY items are energies, they will be converted according to SI */

/*
  names to be listed in SIMNAME.cpi
*/
static struct CPItab_s {
  char *name;  /* name referring to .cpi file */
  char *info;  /* one line of humar-readable info */
  double *var; /* pointer to a double variable with the value */
} CPItab[]={
  /* {"KEY","INFO",&double}, */
  /* ######## carefully set NENERGY ######## */
  {"Epnc","potential energy without cutoff corrections [J/mol]",&En.pot},
  {"Ep0", "potential energy first No.first molecules [J/mol]",&En.pot0},
  {"EpX", "? potential energy excl. No.first molecules [J/mol]",&En.potX},
  {"elst","electrostatic energy [J/mol]",&En.el},
  {"el0", "electrostatic energy first No.first molecules [J/mol]",&En.el0},
   {"elX", "? electrostatic energy excl. No.first molecules [J/mol]",&En.elX},
   {"LJ",  "non-bonded (LJ) energy excl. elst. [J/mol]",&En.LJ}, /* ? */
   {"LJ0", "non-bonded energy excl. elst. first No.first molecules [J/mol]",&En.LJ0},
   {"LJX", "? non-bonded energy excl. elst. excl. No.first molecules [J/mol]",&En.LJX},
   {"bond","bonded energy [J/mol]",&En.bonded},
  {"bon0","bonded energy first No.first molecules [J/mol]",&En.bonded0},
  {"Eext","Extended degrees of freedom energy (Nose, barostat) [J/mol]",&En.ext},
  {"U",   "Internal energy = Epot + Ekin, with cutoff corrections [J/mol]",&En.U},
  {"Unc", "Internal energy = Epot + Ekin, without cutoff corrections [J/mol]",&En.Unc},
  {"H",   "Enthalpy = Epot + Ekin + pV, with cutoff corrections [J/mol]",&En.H},
  {"Hnc", "Enthalpy = Epot + Ekin + pV, without cutoff corrections [J/mol]",&En.Hnc},
  {"fix", "potential energy of forces fixing sites in place (-k) [J/mol]",&En.fix},
  /* ######## no energy conversion from here - see NENERGY above ######## */
  {"Pvir","virial pressure (using electrostatic virial=energy) [Pa]",&En.P},
  {"PdV", "pressure by virtual volume change (also PdVm or PdVa) [Pa]",&En.PdV},
  {"Pnc", "pressure (based on virial) without cutoff corrections [Pa]",&En.Pnc},
  {"cmdr","center-of-mass (to detect drift) [A]",&En.cmdrift},
  {"vdr", "center-of-mass velocity (to detect drift) [A/ps]",&En.vdrift},
  {"andr","angular velocity (to detect drift) in free b.c. [ps-1]",&En.angdrift},
  {"V"   ,"volume [A^3]",&box.V},
#ifdef POLAR
  {"Tpol", "T of Car-Parrinello-like dipoles [K]",&En.Tpol}, /* new in 2.0c: to be used for mech.dip. */
  {"pstd","polar one-step stderr [prog.u.]",&En.polstderr},
  {"pmax","polar one-step max err. [prog.u.]",&En.polmaxerr},
  {"Pstd","Polar stderr [prog.u.]",&En.Polstderr},
  {"Pmax","Polar max err. [prog.u.]",&En.Polmaxerr},
  {"rate","SCF convergence rate",&scf.rate},
  {"X_hf","high-frequency susceptibility by direct response to ext.field",&En.chi_hf},
#endif /*# POLAR */
#if PRESSURETENSOR&PT_VIR
/* virial pressure tensor */
  {"Pvxx","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir},
  {"Pvyy","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir+1},
  {"Pvzz","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir+2},
  {"Hz",  "Enthalpy = Epot + Ekin + p_zz V, without cutoff corrections [J/mol]",&En.Hz},
#  if PRESSURETENSOR&PT_OFF
  {"Pvyz","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir+3},
  {"Pvzx","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir+4},
  {"Pvxy","site-based virial+constraint part of pressure tensor [Pa]",En.Pvir+5},
#  endif /*# PRESSURETENSOR&PT_OFF */
#endif /*# PRESSURETENSOR&PT_VIR */
#if PRESSURETENSOR&PT_KIN
/* site-based kinetic pressure tensor */
  {"Pkxx","site-based kinetic part of pressure tensor [Pa]",En.Pkin},
  {"Pkyy","site-based kinetic part of pressure tensor [Pa]",En.Pkin+1},
  {"Pkzz","site-based kinetic part of pressure tensor [Pa]",En.Pkin+2},
#  if PRESSURETENSOR&PT_OFF
  {"Pkyz","site-based kinetic part of pressure tensor [Pa]",En.Pkin+3},
  {"Pkzx","site-based kinetic part of pressure tensor [Pa]",En.Pkin+4},
  {"Pkxy","site-based kinetic part of pressure tensor [Pa]",En.Pkin+5},
#  endif /*# PRESSURETENSOR&PT_OFF */
#endif /*# PRESSURETENSOR&PT_KIN */
#if PRESSURETENSOR&PT_MOL
/* CM-based kinetic pressure tensor */
  {"PKxx","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin},
  {"PKyy","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin+1},
  {"PKzz","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin+2},
#  if PRESSURETENSOR&PT_OFF
  {"PKyz","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin+3},
  {"PKzx","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin+4},
  {"PKxy","molecular (CM-based) kinetic part of pressure tensor [Pa]",En.PKin+5},
#  endif /*# PRESSURETENSOR&PT_OFF */
#endif /*# PRESSURETENSOR&PT_MOL */
#if (PRESSURETENSOR&PT_ANY) == PT_ANY
/* total pressure tensor */
  {"Ptxx","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens},
  {"Ptyy","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens+1},
  {"Ptzz","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens+2},
  {"trPt","corrected P from pressure tensor: tr(Pt)/3+corr [Pa]",&En.trPt},
#  if PRESSURETENSOR&PT_OFF
  {"Ptyz","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens+3},
  {"Ptzx","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens+4},
  {"Ptxy","pressure tensor w/o LJ cutoff corrections [Pa]",En.Ptens+5},
#  endif /*# PRESSURETENSOR&PT_OFF */
#endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */
  {"Lx","x box size [A]",box.L},
  {"Ly","y box size [A]",box.L+1},
  {"Lz","z box size [A]",box.L+2},
  {"Jx","current density in the x direction [Ampere/m2]",En.J},
  {"Jy","current density in the y direction [Ampere/m2]",En.J+1},
  {"Jz","current density in the z direction [Ampere/m2]",En.J+2},
  {"Mx","dipole moment of the box in x [p.u.=0.0117501081 Debye]",En.M},
  {"My","dipole moment of the box in y [p.u.=0.0117501081 Debye]",En.M+1},
  {"Mz","dipole moment of the box in z [p.u.=0.0117501081 Debye]",En.M+2},
  {"NHxi","Nose-Hoover degree of freedom xi = ln(s)",&En.logs},
  {"NHdx","time derivative of Nose-Hoover degree of freedom xi = ln(s)",&En.dlogs},
#ifdef RGYR
  {"Rgxx","averaged diagonalized moment of inertia of species 0",&En.RGm[0]},
  {"Rgyy","averaged diagonalized moment of inertia of species 0",&En.RGm[1]},
  {"Rgzz","averaged diagonalized moment of inertia of species 0",&En.RGm[2]},
#endif /*# RGYR */
#ifdef SLAB
  {"zsh","slab center as determined by \"autocenter\" [Lz]",&En.zshift},
  {"Pt","tangential pressure (Ptxx+Ptyy-2*Ptzz)/3 [Pa]",&En.Pt},
  {"Ptv","virial part of tangential pressure (Pvxx+Pvyy-2*Pvzz)/3 [Pa]",&En.Ptv},
#  if SLAB & 2
  {"Wcl","dG/dsigma (rev. work) per unit interface area [K/A^2]",&En.Wcleave},
  {"clz0","cleave.z[0] = center of the cleaving wall 0 [Lz]",cleave.z},
  {"clz1","cleave.z[1] = center of the cleaving wall 1 [Lz]",cleave.z+1},
  {"clf0","force to cleaving wall 0 [K/A]",En.fcleave},
  {"clf1","force to cleaving wall 1 [K/A]",En.fcleave+1},
#  endif /*# SLAB & 2 */
#endif /*# SLAB */
#ifdef COULOMB
  {"Ex"  ,"electrostatic field in x-direction",Eext.E},
  {"Ey"  ,"electrostatic field in y-direction",Eext.E+1},
  {"Ez"  ,"electrostatic field in z-direction",Eext.E+2},
  {"einf","el.epsinf, dielectric constant at infinity",&el.epsinf},
  {"xinf","1/(2*el.epsinf+1), ~ constant at M^2",&el.xinf},
#endif
  {"","",NULL} };

void initCP(int no,int nbit,char *col4, char *col5,int corr) /******* initCP */
/*
   nbit=0 : normal write in float[NCP] format
   nbit>0 : packed, nbit words
*/
{
  char *fileinfo=NULL;
  FILE *CP;
  int ncp=0;
  int lengthchanged=0;
  char *corrinfo = corr&1 ? "incl. LJ cutoff corrections " 
                          : "w/o LJ cutoff corrections ";

  if (!option('e')) return;

  if (!CPinfo) {
    /* this initialization (of CPinfo and ssd0->..) is done only once */
    FILE *CPI=fopen(Fn("cpi"),"rt");
    char line[LINELEN];
    int nssd=0,i,col,e;
    struct ssd_s *ssd,**next=&ssd0;
    char *c,*info="?";

    if (nbit<0 || nbit==1 || nbit>18) {
      ERROR(("CPnbit=%d is wrong",nbit))
      nbit=0; }
    CPnbit=nbit;

    if (sizeof(CPinfo0)%4!=1) ERROR(("\"%s\" is bad CPinfo0",CPinfo0))

    strcpy(line,col5); strcat(line,"   "); memcpy(CPinfo0+8,line,4);
    if (No.first) {
      char x[8];
      sprintf(x,"%d  ",No.first);
      copy(CPinfo0+5,x,3); }
    strcpy(line,col4); strcat(line,"   "); memcpy(CPinfo0+4,line,4);
#ifdef COULOMB
    if (Eext.isE<0) copy(CPinfo0+20,string("M%c  ",Eext.isE+3+'x'),4);
#endif /*# COULOMB */

    underline("convergence profile");
    prt("HINT: use \"showcp [OPTIONS] %s.cp\" to show and analyze",simils.simname);
    prt(" 1:Etot total energy or extended Hamiltonian [K]");
    prt(" 2:T    kinetic temperature (cf. Tin, Ttr below) [K]");
    col=2;

    for (c=CPinfo0; *c>' '; c+=4) {
      col++;
      prt_("%2d:%c%c%c%c ",col,c[0],c[1],c[2],c[3]);
      if (!memcmp("Epot",c,4)) prt("total potential energy %s[J/mol]",corrinfo);
      else if (!memcmp("Tin ",c,4)) prt("temperature from internal motions and rotations [K]");
      else if (!memcmp("Ttr ",c,4)) prt("translational temperature from molecular centers of mass [K]");
      else if (!memcmp("Ein ",c,4)) prt("intramolecular potential energy [J/mol]");
      else if (!memcmp("Ep+k",c,4)) prt("potential+kinetic energy (w/o Nose) [J/mol]");
      else if (!memcmp("rho ",c,4)) prt("density [kg/m3]");
      else if (!memcmp("PdVm",c,4)) prt("pressure by CM-based virtual volume change %s[Pa]",corrinfo);
      else if (!memcmp("PdVa",c,4)) prt("pressure by atom-based virtual volume change %s[Pa]",corrinfo);
      else if (!memcmp("P   ",c,4)) prt("pressure (from virial and kin. energy) %s[Pa]",corrinfo);
      else if (!memcmp("|M| ",c,4)) prt("sqrt(<M^2>), simulation cell dipole moment [p.u.=0.0117501077 Debye]");
      else if (!memcmp("Mz  ",c,4)) prt("<Mz>, simulation cell dipole moment in z [p.u.=0.0117501077 Debye]");
#ifdef SHEAR
      else if (!memcmp("Cvel",c,4)) prt("shear-induced velocity [p.u.]");
#endif /*# SHEAR */
      else if (!memcmp("self",c,4)) prt("self-field [p.u.]");
      else if (!memcmp("RvdW",c,4)) prt("van der Waals radius of the site beeing adjusted [A]");
      else if (!memcmp("Rgyr",c,4)) prt("radius of gyration [A]");
      else if (!memcmp("ende",c,4)) prt("end-to-end distance [A], see rg.end[]");
      else _n }

    i=0;
    if (CPI) {
      while (fgets(line,LINELEN,CPI)) {
        i++;
        if (line[0]!='!' && line[0]!='#' && line[0]!='\n') {
          alloczero(ssd,sizeof(*ssd));
          ssd->stat=line[0]=='+';
          e=sscanf(line+ssd->stat,"%d%d%s",&ssd->indx[0],&ssd->indx[1],ssd->name);

          if (e>=2) {
            if (e==2) sprintf(ssd->name,"ss%d",nssd+1);
            loop (e,0,2) if (ssd->indx[e]<0 || ssd->indx[e]>=No.s) {
              ERROR(("%s, line %d: atom index %d out of range <0,%d)",
                     lastFn,i,ssd->indx[e],No.s))
              ssd->indx[e]=0; }
            prt("%2d:%-4s |%d-%d| distance [A]",
              col+i,ssd->name,ssd->indx[0],ssd->indx[1]); }

          else if (e==0) {
            e=sscanf(line+ssd->stat,"%s",ssd->name);
            ssd->indx[0]=-1;

            if (e) {
              int i;

              for (i=0; CPItab[i].var; i++)
                if (!strcmp(CPItab[i].name,ssd->name)) {
                  info=CPItab[i].info;
                  ssd->u.q=CPItab[i].var;
                  //                  if (i>2 && !No.first) No.first=1;
                  if (i>=NENERGY) ssd->indx[0]=-2; /* don't convert Eunit */
                  goto CPIfound; }
              e=0;
              ERROR(("%s: keyword %s unknown",lastFn,ssd->name))
              CPIfound:; }
            prt("%2d:%-4s %s",
              col+i, ssd->name, info); }
          else {
            e=0;
            ERROR(("%s: line %d not enough data",lastFn,i)) }

          if (e) {
            nssd++;
// padding removed!!!
//            while (strlen(ssd->name)<4) strcat(ssd->name," ");
            ssd->next=NULL;
            *next=ssd;
            next=&ssd->next; }
        } /* noncomment line */
      } /* while line of CPI */

      fclose(CPI);
      prt("%s: %d variables and site-site distances recorded",lastFn,nssd);
      alloczero(CPinfo,sizeof(CPinfo0)+4*nssd);
      copy(CPinfo,CPinfo0,sizeof(CPinfo0));

      for (ssd=ssd0,c=CPinfo+(sizeof(CPinfo0)-1); ssd; ssd=ssd->next,c+=4) {
      // padded by SPACE here...
        memset(c,' ',4);
        if (strlen(ssd->name)>4) {
          copy(c,ssd->name,3);
          c[3]=strlast(ssd->name)[0]; }
        else
          copy(c,ssd->name,strlen(ssd->name)); } }
    else {
      CPinfo=CPinfo0;
      prt("no %s file",lastFn); } }

  NCP=2+strlen(CPinfo)/4;

  if (option('e')>0) {
    Min(NCP,option('e'))
    prt(">>> because of option -e, only %d columns recorded",option('e')); }

  if (init<2) {
    float hdr;

    if (!(CP=fopen(Fn(CPnbit?"cpz":"cp"),"rb")))
      ERROR(("open %s (CPnbit=%d)",Fn(CPnbit?"cpz":"cp"),CPnbit))
    else {
      fread(&hdr,sizeof(float),1,CP);
    if (hdr!=CPmark) ERROR(("%s: bad header",lastFn))
      fread(&ncp,sizeof(int4),1,CP);
    if (ncp<2 || ncp>255)
      ERROR(("%s: %d is bad # of fields",lastFn,ncp))
      alloczero(fileinfo,ncp*4-7);
    fread(fileinfo,4,ncp-2,CP);
    if (ncp!=NCP) {
      lengthchanged++;
      WARNING(("%s contains %d fields but %d requested: extra fields %s",
               lastFn,ncp,NCP,ncp<NCP?"truncated":"padded by 0"))
        if (ncp>NCP) {
        char *x;

        alloc(x,ncp*4-7);
        strcpy(x,CPinfo);
        while (ncp>NCP) { strcat(x,"null"); NCP++; }
        /* free(CPinfo); mem leak??? */
        CPinfo=x; }
      NCP=ncp; }
    fclose(CP); } }

  prt("CP record %d floats:",NCP);

  if (no && NCP>=2) ralloc(CPbuf,(int4)no*sizeof(float)*NCP);

  prtCPinfo(CPinfo);
  if (fileinfo) if (!lengthchanged) if (memcmp(fileinfo,CPinfo,NCP*4-8)) {
    prt(" header=\"%s\"\nin file=\"%s\"",CPinfo,fileinfo);
    prt("header in file %s:",lastFn);
    prtCPinfo(fileinfo);
    WARNING(("inconsistent header of CP file\n\
*** this is most often caused by using init=\"cont\" (0) instead of \"start\" (2)")) }
}

void saveCP(int icyc,time_t stoptime) /****************************** saveCP */
{
  static char CPmode[3]="wb";
  static FILE *CP;
  static float hdr=CPmark;
  char *aux;
  int l;
  int4 ncp;

  if (!CPbuf) return;

  if (sizeof(int4)!=sizeof(float)) ERROR(("sizeof(int4)!=sizeof(float))"))
  if (sizeof(float)!=4) ERROR(("sizeof(float)!=4"))

  aux=CPnbit?"cpz":"cp";
  if (init<=1)
    CPmode[0]='a';
  else {
    CPmode[0]='w';
    backup(aux); }

  CP=fopen(Fn(aux),CPmode);
  if (!CP) ERROR(("%s: open",lastFn))
  if (CPmode[0]=='a') prt("appending %s at position %ld", lastFn, ftell(CP));
  if (CPmode[0]=='w') {
    /* create header */
    fwrite(&hdr,sizeof(float),1,CP);
    ncp=NCP;
    fwrite(&ncp,sizeof(int4),1,CP);
    fwrite(CPinfo,4,NCP-2,CP); }
  if (CPnbit) appendpakcp(CP,NCP,CPnbit,(void*)CPbuf,icyc);
  else fwrite(CPbuf,sizeof(float)*NCP,icyc,CP);

  /* ending record with time stamp */
  memset(CPbuf,0,NCP*sizeof(float));
  CPbuf[0]=CPmark;
  /* old version */
  copy((char*)CPbuf+(sizeof(float)*NCP-sizeof(stoptime)),
       &stoptime,sizeof(stoptime));
  /* new version */
  aux=ctime((void*)&stoptime);
  l=min(strlen(aux),sizeof(float)*(NCP-1));
  copy(CPbuf+1,aux,l);
  aux=(char*)(CPbuf+1);
  while (l--) { if (*aux=='\n') *aux=0; aux++; }
  ((char*)(CPbuf+NCP))[-1]=0;

  fwrite(CPbuf,sizeof(float)*NCP,1,CP);

  prt("closing %s at position %ld on %s", lastFn, ftell(CP), (char*)(CPbuf+1));
  if (fclose(CP)) if (!CP) ERROR(("%s: close",lastFn))
}


/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                  site-site radial distribution functions
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

static int global_rdfgrid;

void initrdf(double gr,double cutoff,char *fn) /******************** initrdf */
/*
  gr=grid (subintervals/length unit)
  initializing histograms for rdf's,
  gr=0: none
  gr<0: as |gr| but all site-site histograms merged
  assigned to ss->rdf in setss which is called by initss
*/
{
  int i,j,nss=0;
  int sp;
  int *sitecount;
  int nh=(int)(cutoff*abs(gr)+1.0000001);
  int size=sizeof(rdf_t)-2*sizeof(int4)+sizeof(int4)*nh;
  rdf_t *r;
  FILE *ff=NULL;

  struct ss_s {
    struct ss_s *next;
    char s1[8],s2[8]; } *ss,*ss0=NULL;

  underline("radial distribution functions");
  if (!gr) prt("Radial distribution functions are turned off.");

  if (cutoff<0) ERROR(("rdf.cutoff=%g out of range",cutoff))

  if (gr>0 && !(ff=fopen(fn,"rt")))
    prt("%s not found, RDFs will be recorded for all pairs of site types.",fn);

  if (ff) {
    char line[LINELEN],*s1,*s2;

    prt("reading %s: site type pairs",fn);

    while (fgets(line,LINELEN,ff)) if (!strchr("!#",line[0])) {
      s1=strtok(line," \t\n"); if (!s1) continue;
      if (s1[0]=='*') {
        if (nss) ERROR(("%s: wildcard specifying all pairs comes after %d pairs.",lastFn,nss))
        nss=0; ss0=NULL; /* memory leak */
        fclose(ff); ff=NULL;
        goto escape; }
      if (strlen(s1)==1 && (islower(s1[0])||isdigit(s1[0]))) continue;
      s2=strtok(NULL," \t\n");
      if (!s2) { ERROR(("%s: %s what?",fn,s2)) continue; }

      alloczero(ss,sizeof(struct ss_s));
      copy(ss->s1,s1,5);
      copy(ss->s2,s2,5);
      ss->next=ss0; ss0=ss;
      nss++; }
    prt("%d pairs for rdf selected",nss); }

 escape:

  global_rdfgrid=gr;

  /* now site-type statistics (option -j ignored) */

  allocarrayzero(sitecount,nsites);

  loop (sp,0,nspec) {
    specinfo_t *s=spec[sp];
    loop (i,0,s->ns) sitecount[s->si[i].st] += s->N; }

  ralloc(rdf,nsites*sizeof(rdf_t**));

  loop (i,0,nsites) {
    ralloc(rdf[i],nsites*sizeof(rdf_t*));
    loop (j,0,nsites) rdf[i][j]=NULL; }

  if (option('v')&1) {
    prt("\n _i_site_count  _i_site_count  _i_site_count  _i_site_count  _i_site_count");
    loop (i,0,nsites)
      prt_("%3d %-4s %4d %c",i,sitedef[i].name,sitecount[i],"    \n"[i%5]);
    _n }

  if (gr) loop (i,0,nsites) {

    loopto (j,0,i) {
      if (gr>0 || (i|j)==0) {
        if (ff) {
          for (ss=ss0; ss; ss=ss->next) if (
             (!strcmp(ss->s1,sitedef[i].name) && !strcmp(ss->s2,sitedef[j].name))
           ||(!strcmp(ss->s1,sitedef[j].name) && !strcmp(ss->s2,sitedef[i].name)) )
            goto incl;
          r=NULL;
          goto noincl; /* not include this pair */ }
      incl:
        sdsralloczero(r,size); /* cleared all .. purify! */
        r->nmeas=r->npair=0;
        r->indx[0]=i; r->indx[1]=j;
        if (gr<=0) {
          r->ns[0]=r->ns[1]=No.s;
          strcpy(r->name[0],"*");
          strcpy(r->name[1],"*"); }
        else {
          r->ns[0]=sitecount[i];
          copy(r->name[0],sitedef[i].name,
               min(sizeof(r->name[0]),sizeof(sitedef[0].name)));
          r->ns[1]=sitecount[j];
          copy(r->name[1],sitedef[j].name,
               min(sizeof(r->name[0]),sizeof(sitedef[0].name))); }
        r->grid=(double)abs(gr);
        r->V=0;
        r->nhist=nh; }
      else
        r=rdf[0][0];
    noincl:
      rdf[i][j]=rdf[j][i]=r; } }

  free(sitecount);
  if (ff) fclose(ff);
  for (ss=ss0; ss; ) {
    struct ss_s *new=ss->next;
    free(ss);
    ss=new; }
}

void advancerdf(void) /****************************************** advancerdf */
{
  int i,j;
  rdf_t *r;

  if (global_rdfgrid && measure)
    loop (i,0,nsites) loopto (j,0,i) {
      r=rdf[i][j];
      if (r) { r->nmeas++; r->V+=box.L[0]*box.L[1]*box.L[2]; }
      if (global_rdfgrid<0) return; }
}

void rdfgraph(void) /********************************************** rdfgraph */
/*
  char-like graph of rdf, obsolete feature
*/
{
int i,j,ng=nsites,merged;
#define scale (19.0/76)

merged=nsites>1 && rdf[0][0]==rdf[0][1];
if (merged) ng=1;

loop (i,0,ng) loopto (j,0,i) {
  double dr,q;
  rdf_t *r=rdf[i][j];
  int ir;
  if (r) if (r->npair) {
    dr=1/r->grid;
    prt("grid=1/%.2f=%.f  nhist=%d  npair=%ld  nmeas=%ld  V=%g",
             r->grid,dr,  r->nhist, r->npair,  r->nmeas,  r->V);
    q=r->V/(4*PI*r->npair*Sqr(r->nmeas)*Cub(dr)) * scale;
    if (merged) prts("rdf g(r)");
    else prt("rdf g(r) for sites %s-%s",sitedef[i].name,sitedef[j].name);
    prt_("1="); graph(scale,77);
    loop (ir,0,r->nhist) if (r->hist[ir]) {
      prt_(ir<100 ? "%2i" : "%02i",ir%100);
      graph(
        r->hist[ir]*q/(ir*ir+ir+1.0/3),77); }
    prt_("1="); graph(scale,77); } }
#undef scale
}

void loadrdf(double rdfgrid) /************************************** loadrdf */
{
  int ng = rdfgrid<0 ? 1 : nsites;
  int nsi,i,j;
  double gr;

  if (!rdfgrid) return;

  VarOpen(Fn("rdf"),"r");
  VarRead(&gr,sizeof(gr));
  VarRead(&nsi,sizeof(nsi));
  if (gr!=rdfgrid || nsi!=nsites)
    ERROR(("%s: grid or # of sites inconsistent",lastFn))
  loop (i,0,ng)
    loopto (j,0,i)
      if (rdf[i][j]) {
        VarReadSds(rdf[i][j]);
        if (rdf[i][j]->nhist*sizeof(rdf[i][j]->hist[0])+((char*)rdf[i][j]->hist-(char*)rdf[i][j]) != rdf[i][j]->size)
        ERROR(("Inconsistent RDF file. If you continue with RDF obtained by cook < V2.7f,\n\
*** use  rdfconv %s  first",lastFn)) }

  VarClose();
}

void saverdf(double rdfgrid) /************************************** saverdf */
{
  int ng = rdfgrid<0 ? 1 : nsites, i,j;

  if (!rdfgrid) return;

  backup("rdf");
  VarOpen(Fn("rdf"),"w");
  VarPut(&rdfgrid,sizeof(rdfgrid));
  VarPut(&nsites,sizeof(nsites));
  loop (i,0,ng) loopto (j,0,i)
    if (rdf[i][j]) VarPutSds(rdf[i][j]);
  VarClose();
}

#if PARALLEL==1
void parallelizerdf(int replicate)
/*
  replicate=1: replicate the histograms to be able to sum them concurrently
  replicate=0: sum back to one histogram
*/
{
  int ih,ith,i,j;

  loop (i,0,nsites) loopto (j,0,i) if (replicate) {
    /* replicate histograms, grid, nhist (more not needed) */
    rallocarrayzero(sstab[i][j].rdfs,No.lcth);
    sstab[j][i].rdfs=sstab[i][j].rdfs;
    loop (ith,0,No.lcth)
      if (!rdf) sstab[i][j].rdfs[ith]=NULL; /* should not happen */
      else if (No.measureserial)
        sstab[i][j].rdfs[ith]=sstab[i][j].rdf; /* may be NULL */
      else if (sstab[i][j].rdf) {
        if (ith) {
          /* replica */
          sdsralloczero(sstab[i][j].rdfs[ith],sstab[i][j].rdf->size);
          sstab[i][j].rdfs[ith]->grid=sstab[i][j].rdf->grid;
          sstab[i][j].rdfs[ith]->nhist=sstab[i][j].rdf->nhist; }
        else {
          /* the original structure */
          sstab[i][j].rdfs[ith]=sstab[i][j].rdf; } } }
  else {
    /* sum back */
    if (sstab[i][j].rdf && !No.measureserial)
      loop (ih,0,sstab[i][j].rdf->nhist)
        loop (ith,1,No.lcth)
          sstab[i][j].rdf->hist[ih]+=sstab[i][j].rdfs[ith]->hist[ih];
    }
}
#endif /*# PARALLEL==1 */

/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                           dihedral angle distribution
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef DIHHIST

static int diheq(char type1[4][6],char type2[4][6])
/*
  returns 1 if the dihedral has already been included
  (according to the rules given by #define DIHHIST)
*/
{
#  define EQ(I,J) !strcmp(type1[I],type2[J])

#  if DIHHIST==2
  return (EQ(1,1) && EQ(2,2)) || (EQ(1,2) && EQ(2,1));
#  elif DIHHIST==4
  return
      (EQ(0,0) && EQ(1,1) && EQ(2,2) && EQ(3,3))
   || (EQ(0,3) && EQ(1,2) && EQ(2,1) && EQ(3,0));
#  elif DIHHIST==-1
  return 0; /* all included separately */
#  elif DIHHIST==1
  return 1; /* all dihedrals summed */
#  else /*#!DIHHIST==2!DIHHIST==4!DIHHIST==-1!DIHHIST==1 */
#    error "illegal DIHHIST"
#  endif /*#!DIHHIST==2!DIHHIST==4!DIHHIST==-1!DIHHIST==1 */
#  undef EQ
}

dihhist_t *dihhead;

void includedihhist(torsion_t *t, char type[4][6], int sp) /* includedihhist */
/*
 includes a dihedral provided that it has not been included yet (the rules
 when a dihedral is considered as the same depend on #define DIHHIST)
*/
{
  dihhist_t *dih;

  for (dih=dihhead; dih; dih=dih->next)
    if (diheq(type,dih->type)) goto already;
  alloczero(dih,sizeof(dihhist_t));
  dih->next=dihhead;
#  if DIHHIST==-1
  dih->sp=sp;
#  else /*# DIHHIST==-1 */
  dih->sp=-1; /* species irrelevant */
#  endif /*#!DIHHIST==-1 */
  dihhead=dih;
  copy(dih->type,type,sizeof(dih->type));
 already:
  t->parm.dihhist=dih;
  if (abs(t->parm.n)==3) dih->gauche_trans=2*PI/3;
  else dih->gauche_trans=PI/2;
}

void initdihhist(int dihgrid) /********************************* initdihhist */
{
  dihhist_t *dih;
  double q=dihgrid/(2*PI+No.eps);

  for (dih=dihhead; dih; dih=dih->next) {
    dih->grid=dihgrid;
    dih->q=q;
    if (dihgrid) ralloczero(dih->hist,sizeof(unsigned)*dihgrid);
    else dih->hist=NULL; }
}

double gauchetrans(int key) /*********************************** gauchetrans */
/*
  key=-3: assigns all gauche[cis]/trans statistics to 0
  key=-2: adds gauche/trans statistics to StaAdd and assigns 0
  key=-1: calculates total gauche ratio
  key>0: returns gauche/all for dihedral# key
         (NEW: key==0 not alloweed)
*/
{
  char name[20];
  dihhist_t *dih;
  int sum,idih=0;
  int sumg=0,sumd=0;

  for (dih=dihhead; dih; dih=dih->next) {

    sum=dih->trans+dih->gauche;

    if (key==-2 && sum) {
      sprintf(name,"%s(%d)",dih->gauche_trans<2?"cis":"gauche",idih+1);
      StaAdd(name,dih->gauche/(double)sum); }

    if (key<-1) dih->gauche=dih->trans=0;
    else if (key==-1) {
      sumg+=dih->gauche;
      sumd+=sum; }
    else if (idih+1==key) return dih->gauche/(double)sum;

    idih++; }

  if (key>=0) ERROR(("no dihedral # %d",key))

  return sumg/(sumd+1e-33);
}

void printdihhist(void) /************************************** printdihhist */
{
  dihhist_t *dih;
  unsigned sum,trans,gauche;
  int i,idih=0;
  FILE *f=NULL;
  double d;

  if (!dihhead) return;

  for (dih=dihhead; dih; dih=dih->next) {
    if (!dih->grid) continue;

#  if 0
    if (dih->hist[dih->grid]) {
      if (option('v')&1) prt("next dihedral hist[last]=%d",dih->hist[dih->grid]);
      dih->hist[dih->grid-1]+=dih->hist[dih->grid];
      dih->hist[dih->grid]=0; }
#  endif /*# 0 */

    trans=gauche=0;
    loop (i,0,dih->grid)
      if (i<2*dih->grid/3) gauche+=dih->hist[i];
      else trans+=dih->hist[i];
    sum=trans+gauche;

    if (sum || option('v')&2)
      prt_("dihedral (%d) %s-%s-%s-%s : ",
           idih+1,
           dih->type[0],dih->type[1],dih->type[2],dih->type[3]);

    if (sum) {
      if (!f) {
        _n
        backup("dia");
        prt_("                            ");
        f=fopen(Fn("dia"),"wt"); }

      fprintf(f,"# (%d) %s-%s-%s-%s : ",
              idih+1,
              dih->type[0],dih->type[1],dih->type[2],dih->type[3]);
      fprintf(f,"%d points, %.5f trans, %.5f %s\n",
              sum,(double)trans/sum,(double)gauche/sum,
              dih->gauche_trans<2?"cis":"gauche"); }

    if (sum || option('v')&2) prt_("%u points",sum);

    if (sum) prt(", %.5f trans, %.5f %s",
                 (double)trans/sum,(double)gauche/sum,
                 dih->gauche_trans<2?"cis":"gauche");
    else if (option('v')&2) _n

    if (sum) {
      loop (i,0,dih->grid) {
        d=dih->hist[i]/((double)sum/dih->grid*(2*PI));
        /*        if (i==0) fprintf(f,"%5.1f %7.5f\n",0.0,d); */
        fprintf(f,"%5.1f %7.5f\n", (i+0.5)*360.0/dih->grid,d); }
      /*      fprintf(f,"%5.1f %7.5f\n",360.0,d);*/
      fprintf(f,"\n"); }
    if (sum) idih++; }

  if (f) fclose(f);
}

void loaddih(int dihgrid) /***************************************** loaddih */
{
  dihhist_t *dih;
  struct {
    double dummy1;
    dihhist_t dih;
    double dummy2;} aux;

  if (!dihgrid) return;

  VarOpen(Fn("dih"),"r");
  for (dih=dihhead; dih; dih=dih->next) {
    VarRead((char*)aux.dih.type,VarFile.size); /* note: char aux.dih.type[][] */
    if (aux.dih.grid!=dih->grid
      || memcmp(aux.dih.type,dih->type,sizeof(aux.dih.type)))
      ERROR((
             "%s inconsistent (grid=%d type=%s, should be %d %s - try DIHOLDFMT?)",
             Fn("dih"),aux.dih.grid,aux.dih.type,dih->grid,dih->type))
        /* OLD version (grid in 180, safe bound): (dih->grid+1) */
        VarRead(dih->hist,dih->grid*sizeof(dih->hist[0])); }
  VarClose();
}

void savedih(int dihgrid) /***************************************** savedih */
{
  dihhist_t *dih;

  if (!dihgrid) return;

  backup("dih");
  VarOpen(Fn("dih"),"w");
  for (dih=dihhead; dih; dih=dih->next) {
    VarPut(dih->type,sizeof(dihhist_t)-(int)((char*)dih->type-(char*)dih));
    VarPut(dih->hist,dih->grid*sizeof(unsigned)); }
  VarClose();
}
#endif /*# DIHHIST */

static double chi2eps(double chi)
/* "b.c.-dependent susceptibility -> dielectric constant */
{
  return 1+1/(1/chi-1/(1+2*el.epsinf));
}

void prtdielconst(double V,int ewald) /************************ prtdielconst */
/*
  ewald=1: ewald-based
  ewald=0: meant for short-range potentials - restricted
*/
{
  double sat,factor,T;
  vector VarM;
  double sumM,sumMM,MM,dMM,xM;
#ifdef POLAR
  double chi_alpha=0;
#endif /*# POLAR */
  int sp,i,ichi,nchi,Nm;
  struct chi_s {
    double M; /* from var M (low-frequency) */
    double dM; /* error */
    double hf; /* high-frequency reponse (POLAR only) */
    double dhf; /* error (POLAR only) */
    double eps_hf; /* high-frequency diel. const. (POLAR only) */
    int Nm;     /* number of measurements */
  } chi[4]; /* 0=nonpolar, or with hf from polarizability (by Clausius-Mossotti)
               1=as 0 with error from 3 components
               2=as 0 with hf measured
               3=M,dM contain both low and high frequency (0+2 merged) */

  underline("dielectric constant");
  /* sum of dipole moments */

  sumM=sumMM=0;
  loop (sp,0,nspec) {
#ifdef POLAR
    sumM+=StaMean(string("mu_perm[%d]",sp))*spec[sp]->N;
    if (StaError) sumM+=StaMean(string("dipmom[%d]",sp))*spec[sp]->N; /* legacy name */
    sumMM+=StaMean(string("mu_tot[%d]",sp))*spec[sp]->N;
    if (StaError) sumMM+=StaMean(string("dipmomboth[%d]",sp))*spec[sp]->N; /* legacy name */
#else /*# POLAR */
    sumM+=StaMean(string("mu[%d]",sp))*spec[sp]->N;
    if (StaError) sumM+=StaMean(string("dipmom[%d]",sp))*spec[sp]->N; /* legacy name */
#endif /*#!POLAR */
  }
  prt("sum of permanent dipole moments = M_perm = %g p.u. = %g D = %g C m",
      xM=sumM,sumM*0.0117501081,sumM*3.91941417e-32);
#ifdef POLAR
  prt("sum of permanent+induced dipole moments = M_tot = %g p.u. = %g D = %g C m",
      xM=sumMM,sumMM*0.0117501081,sumMM*3.91941417e-32);
#endif /*# POLAR */
  Fn("cpa");
  prt("Hint: |M| or Mz is usually in column 8.\n\
After showcp has generated %s, this command will show saturation:\n\
  plot %s:0:H/%f\n",lastFn,lastFn,xM);

  T=StaMean("Tkin");
  if (StaError) return;
  factor=4*PI/(3*T*V);

  if (ewald) MM=StaMean("M^2"),dMM=StaStdErr("M^2"),Nm=StaN("M^2");
  else MM=StaMean("dipmomtot"),dMM=StaStdErr("dipmomtot"),Nm=StaN("dipmomtot");

  if (MM==0 || StaError) {
    prt("not dipolar system");
    return; }

  put2(MM,dMM)

  prt("these calculations are based on a fluctuation formula with el.epsinf=%g",el.epsinf);
  if (Eext.isE) prt("WARNING: EXTERNAL ELECTRIC FIELD");

  prt("saturation stdev(MM)/M_perm = %g",sqrt(MM)/sumM);
#ifdef POLAR
  prt("saturation stdev(MM)/M_tot = %g",sqrt(MM)/sumMM);
  if (sumMM) Eext.sumM=sumMM; else
#endif /*# POLAR */
    Eext.sumM=sumM;

  sat=sqrt(MM)/Eext.sumM;
  if (sat>0.1) prt("WARNING: The dipole moment of the cell is %.3f of the saturated one.\n\
*** The dielectric constant may be affected by nonlinear response.\n\
*** Consider decreasing el.epsinf or increasing the box size.",sat);

  sat=MM;
  if (Eext.isE) {
    loop (i,0,3) if (Eext.E[i]) {
      char *Mc=string("M%c",i+'x');
      double av=StaMean(Mc);

      if (StaError) av=StaMean(Mc=string("M[%d]",i)); /* legacy */
      prt("<M%c>^2=%g^2 subtracted from <M^2> because of the elst field",i+'x',av);
      MM-=av*av; }
    prt("corrected MM=%g (corrected MM)/(original MM)=%g",MM,MM/sat); }

  arrayzero(chi,4);

  chi[0].M=factor*MM;
  chi[0].dM=factor*dMM;
  chi[0].Nm=Nm;

  sumM=sumMM=0;
  if (ewald) loop (i,0,3) {
    char *Mc=string("M%c",i+'x');

    VarM[i]=StaVar(Mc);
    if (StaError) {
      /* legacy */
      Mc=string("M[%d]",i);
      VarM[i]=StaVar(Mc); }
    if (StaError) break;
    Nm=StaN(Mc);
    sumM+=VarM[i];
    sumMM+=Sqr(VarM[i]); }

  prt("Var M[i]= %g %g %g  sum = %g",VARG(VarM),sumM);

  chi[1].M=factor*sumM;
  chi[1].dM=factor*sqrt((sumMM/3-Sqr(sumM/3))/2);
  chi[1].Nm=Nm;

  nchi=2;

#ifdef POLAR
  chi_alpha=No.A*(4*PI/V);
  put(chi_alpha)
  if (option('p')/10%10==0) {
    chi_alpha=0;
    prt("Car-Parrinello-like: chi_alpha=0 (eps_hf=1) assumed (fluctuations of\n\
*** induced dipoles are described explicitly for thermalized Drude oscilators)"); }
  chi[0].eps_hf=chi[1].eps_hf=(3+2*chi_alpha)/(3-chi_alpha);
  chi[0].hf=chi[1].hf=(2*el.epsinf+1)*(chi[0].eps_hf-1)/(chi[0].eps_hf+2*el.epsinf);

  chi[2].hf=StaMean("high-frequency chi");
  nchi+=!StaError;
  if (nchi>2) {
    chi[2].dhf=StaStdErr("high-frequency chi");
    chi[2].M=chi[0].M;
    chi[2].dM=chi[0].dM;
    chi[2].Nm=StaN("high-frequency chi"); }

  chi[3].hf=StaMean("total chi");

  nchi+=!StaError;
  if (nchi>3) {
    double x;

    if (Eext.isE) {
      loop (i,0,3) if (Eext.E[i]) {
        if (scf.Estride>1) x=StaMean(string("xchi_%c",'x'+i));
        else x=sqrt(factor)*StaMean(string("M%c",'x'+i));
        put2(i,x)
        if (StaError) {
          static int pass=1;
          if (pass) WARNING(("internal: neither xchi_%c nor M%c found",'x'+i,'x'+i))
          pass=0;
          break; }
        chi[3].hf-=Sqr(x); } }
    chi[3].dhf=StaStdErr("total chi");
    chi[3].Nm=StaN("total chi");
    chi[3].M=0;
    chi[3].dM=0; }
#endif /*# POLAR */

  prt("NB: errors are expressed as:\n\
  MEAN STDERR  = values with linear estimates of standard errors\n\
  [ LOW HIGH ] = range of given confidence limit");

  loop (ichi,0,nchi) {
    double X=chi[ichi].M+chi[ichi].hf;
    double d=sqrt(Sqr(chi[ichi].dM)+Sqr(chi[ichi].dhf));
    double eps=chi2eps(X);
    double deps=(chi2eps(X+d*0.005)-chi2eps(X-d*0.005))*100;
    double eps_hf=chi2eps(chi[ichi].hf);
    double deps_hf=chi2eps(chi[ichi].hf+chi[ichi].dhf/2)-chi2eps(chi[ichi].hf-chi[ichi].dhf/2);

    switch (ichi) {
      case 0: prt("NONPOLAR or POLAR with Clausius-Mossotti-based high-frequency epsilon:"); break;
      case 1: prt("as above, error determined from three components (Var M_i):"); break;
#ifdef POLAR
      case 2:
        prt("POLAR with separately determined high-frequency epsilon\n\
* sampling elst field E = %g V/m in directions [x,-x,y,-y,z,-z]\n\
* sampling rate (scf.stride) = %d cycles = %g ps:",scf.E,scf.Estride,scf.Estride*No.dt);
        break;
      case 3:
        prt("POLAR: high-frequency response and Var M-based response measured at once:");
        if (Eext.isE) prt("WARNING: <M>^2 caused by external elst field subtracted");
        if (scf.Estride>1) prt("WARNING: low sampling rate (scf.Estride=%d)",scf.Estride);
        break;
#endif /*# POLAR */
      default: WARNING(("bad ichi - POLAR simulation continued by nonpolar cook*?")) }

    prt("  EPS%d  chi.M = %g %g  chi.hf = %g %g  chi.Nm = %d",
        ichi,chi[ichi].M,chi[ichi].dM,chi[ichi].hf,chi[ichi].dhf,chi[ichi].Nm);

    if (ichi<3) prt("  EPS%d  high-frequency epsilon = %8.6f %8.6f   (chi = %8.6f %8.6f)",
                    ichi, eps_hf,deps_hf, chi[ichi].hf,chi[ichi].dhf);
    prt("  EPS%d  total epsilon = %.4f %.4f",
        ichi,eps,deps);
    prt("  EPS%d  68%% confidence range = [ %.4f %.4f ]",
        ichi,chi2eps(X-d),chi2eps(X+d));
    prt("  EPS%d  95%% confidence range = [ %.4f %.4f ]",
        ichi,chi2eps(X-2*d),chi2eps(X+2*d));
  }

}

#ifdef SHEAR
void prtviscosity(void) /************************************** prtviscosity */
{
  if (shear) {
    double den=StaMean("shear:den");
    double dden=StaStdErr("shear:den");

    if (!StaError) {
      double num=StaMean("shear:num");
      double dnum=StaStdErr("shear:num");
      double eta=num/den;
      double relerr=sqrt(Sqr(dnum/num)+Sqr(dden/den));
      prt("\nviscosity = <shear:num>/<shear:den> = %g prog.units, rel.err=%g", eta,relerr);
      prt("viscosity/Pa.s = %g %g",eta*(Punit*timeunit),eta*(Punit*timeunit)*relerr);
    }
  }
}
#endif /*# SHEAR */

#if 0
/* fool-proof dipole moment (for debugging, POLAR+DRUDE) */
void dipmom(ToIntPtr A,char *info)
{
  molecule_t *mn;
  siteinfo_t *si;
  vector *r,*rpol;
  vector D,rpi;
  int n,i;

  VO(D,=0)

  loop (n,0,No.N) {
    mn=molec+n;
    si=spec[mn->sp]->si;
    r=rof(mn,A->rp);
    rpol=polarrof(mn,A->rp);

    loop (i,0,mn->ns) {
      VVV(rpi,=r[i],+rpol[i])
      VVV(D,+=si[i].charge*r[i],+si[i].chargepol*rpi) } }

  prt("%12s %13g %13g %13g",info,VARG(D));
}
#endif /*# 0 */

#ifdef RGYR
double measureplus(double *endend,int end[2],int I) /*********** measureplus */
#else /*# RGYR */
void measureplus(void) /**************************************** measureplus */
#endif /*#!RGYR */
/*
  calculates:
  - molecular dipole moments for all species separately
    for ions with respect to the center of charge^2
  - current autocorrelation function (if lag.cond)
  - velocity autocorrelation function (see lag.v, lag.dim)
  #ifdef RGYR also:
    gyration tensor = SUM m_i (r_i-r_CM)(r_i-r_CM)/SUM m_i
    radius of gyration = sqrt(SUM m_i (r_i-r_CM)^2/SUM m_i)
      (=return value; in statistics squared)
    |rhead-rtail|
      (=return value; in statistics squared)
*/
{
  molecule_t *mn;
  siteinfo_t *si;
  int ns,sp,lastsp=-1,moln=0,n,i,sumn;
  vector *r,*v,D,CQ,CM,current;
  double q,sumq,sumqq,sumM,sq;
  static struct sum_s {
    double D,DD,endend;
    double Rgyrq; /* trace of the gyration tensor = R_gyr^2 */
    vector RGm; /* eigenvalues (diagonalized) */
#ifdef POLAR
    double Dpol,DDpol,Dboth,DDboth,Dprj,DDprj;
#endif /*# POLAR */
    int nD,nRG,nend;
  } *sum;
  static int nspec0;
#ifdef RGYR
  double RG[3][3];
  double RGret=0,endendret=0,Rgyrqs;
  int sumnend;
  static int warning=10;
  vector *velocity;
  static double **RGa=NULL;
#endif /*# RGYR */

#ifdef POLAR
  vector Dpol;
  double qpol;
  vector *rpol,*vpol;
  vector sumD3,sumD3pol;
#  define sumD3both En.M
  VVO(sumD3pol,=sumD3both,=0)
#else /*# POLAR */
#  define sumD3 En.M
#endif /*#!POLAR */

  el.sumM=0;

  /* to avoid frequent alloc/dealloc, made static */
  if (!nspec0) {
    allocarrayzero(sum,nspec);
    nspec0=nspec; }
  else {
    /* bug fixed in 2.7f: not reset to zero */
    loop (i,0,nspec) memset(&sum[i],0,sizeof(sum[i]));
    if (nspec>nspec0) ERROR(("nspec changed (old=%d,new=%d)",nspec0,nspec)) }

#ifdef RGYR
  if (!RGa) alloc2Darray(RGa,3,3);
  Min(lag.nv,No.N)
  if (lag.dim>DIM) {
    WARNING(("lag.dim=%d out of range: changed to 3",lag.dim))
    lag.dim=3; }
  if (lag.v) allocarrayzero(velocity,lag.nv);
#endif /*# RGYR */

  VO(sumD3,=0)
  VO(current,=0)

  loop (n,0,No.N) {
    mn=molec+n;
    ns=mn->ns;
    sp=mn->sp;
    if (sp==lastsp) moln++;
    else moln=0;
    lastsp=sp;
    si=spec[sp]->si;
    r=rof(mn,cfg[0]->rp);
    v=rof(mn,cfg[1]->rp);
#ifdef POLAR
    rpol=polarrof(mn,cfg[0]->rp);
    vpol=polarrof(mn,cfg[1]->rp);
    VO(Dpol,=0)
#endif /*# POLAR */
    VVVO(D,=CQ,=CM,=0)
    sumq=sumqq=sumM=0;
#ifdef RGYR
    memset(RG,0,sizeof(RG));
#endif /*# RGYR */
    loop (i,0,ns) {
      /* WARNING: this test is generally incorrect for POLAR&32!!!*/
      if ( (q=si[i].charge) ) {
        /* dipole moments */
#ifdef POLAR
#  if POLAR&32
        if (si[i].qtype&32) {
          qpol=rpol[i][0]; /* q + delta q */
          if (el.centroid) VV(CQ,+=Sqr(qpol)*r[i]) /* ions: sqr charge centroid */
          sumqq+=Sqr(qpol);
          sumq+=qpol; /* NB: sum qpol = sum q */
          VV(current,+=qpol*v[i])
          VV(D,+=q*r[i])
          VV(Dpol,+=(qpol-q)*r[i])
        }
        else
        /* optimization possible - switch nonpol, skip uncharged... */
#  endif /*# POLAR&32 */
        { /* Drude */
          qpol=si[i].chargepol;

          /* current */
          VVV(current,+=q*v[i],+qpol*vpol[i])

          /* molecule without induced dipoles */
          q+=si[i].chargepol;
          if (el.centroid) VV(CQ,+=Sqr(q)*r[i]) /* ions: sqr charge centroid (w/o Drude) */
          sumqq+=Sqr(q);
          sumq+=q;
          VV(D,+=q*r[i])

          /* induced dipoles */
          VV(Dpol,+=qpol*rpol[i])
        }
#else /*# POLAR */
        /* nonpolar */
        if (el.centroid) VV(CQ,+=Sqr(q)*r[i]) /* ions: sqr charge centroid */
        sumqq+=Sqr(q);
        VV(current,+=q*v[i])
        sumq+=q;
        VV(D,+=q*r[i])
#endif /*#!POLAR */
      } /* if charged */

#ifdef RGYR
      /* summing up CM+RG things */
      q=si[i].mass;
      sumM+=q;
      VV(CM,+=q*r[i])
      if (lag.v) VV(velocity[n],+=q*v[i])
      { 
        int ii,jj;
        /* NB: symmetry used */
        loop (ii,0,3) loopto (jj,0,ii) RG[ii][jj]+=q*r[i][ii]*r[i][jj]; 
      }
#endif /*# RGYR */
    } /* i */

    if (sumqq) {
      sum[sp].nD++;

      /* dipole moment for ions to q^2 centroid (w/o Drude charge) */
      if (el.centroid) VV(D,-=sumq/sumqq*CQ)

      /* permanent dipole (only this one for nonpolarizable models) */
      q=SQR(D);
      sum[sp].D+=sq=sqrt(q);
      sum[sp].DD+=q;
      VV(sumD3,+=D)
#ifdef POLAR
      /* induced dipole */
      qpol=SQR(Dpol);
      sum[sp].Dpol+=sqrt(qpol);
      sum[sp].DDpol+=qpol;
      VV(sumD3pol,+=Dpol) /* bug fixed */

      /* projection of the induced dipole to the direction of the permanent one */
      qpol=SCAL(D,Dpol)/sq;
      sum[sp].Dprj+=qpol; /* bug fixed */
      sum[sp].DDprj+=Sqr(qpol);

      /* both permanent + induced */
      VV(D,+=Dpol)
      q=SQR(D);
      sum[sp].Dboth+=sqrt(q);
      sum[sp].DDboth+=q;
      VV(sumD3both,+=D)
#endif /*# POLAR */
    } /* sumqq */

#ifdef RGYR
    if (lag.v && n<lag.nv) {
      StaSet(0,lag.v,2,0);
      loop (i,0,lag.dim)
        StaAdd(string("v%c%d.%d",i+'x',sp,moln),velocity[n][i]/sumM); }

    if (ns>1) {
      double RGval,endendval,x;
      static int End[2],oldsp=-1;
      int ii,jj;

      /* filling the symmetric parts */
      RG[1][2]=RG[2][1];
      RG[0][1]=RG[1][0];
      RG[0][2]=RG[2][0];

      VO(CM,/=sumM)
      RGval=0;
      loop (ii,0,3) {
        loop (jj,0,3)
          RGa[ii][jj]=RG[ii][jj]/sumM - CM[ii]*CM[jj];
        RGval+=RGa[ii][ii]; }
      sum[sp].Rgyrq += RGval;

      Jacobi(-3,RGa,NULL,No.eps);

      /* order Rgxx<Rgyy<Rgzz */
      if (RGa[0][0]>RGa[1][1]) x=RGa[0][0],RGa[0][0]=RGa[1][1],RGa[1][1]=x;
      if (RGa[1][1]>RGa[2][2]) x=RGa[1][1],RGa[1][1]=RGa[2][2],RGa[2][2]=x;
      if (RGa[0][0]>RGa[1][1]) x=RGa[0][0],RGa[0][0]=RGa[1][1],RGa[1][1]=x;

      loop (ii,0,3) sum[sp].RGm[ii]=RGa[ii][ii];

      sum[sp].nRG++;

      if (sp!=oldsp) {
        if (sp<oldsp) warning=0;
        loop (i,0,2) {
          if (warning && (end[i]<-ns || end[i]>=ns)) {
            prt("WARNING: molecule %d, ns=%d: bad rg.end[%d]=%d",
                n,ns,i,end[i]);
            if (--warning==0) prt("... more warnings suppressed"); }

          if (end[i]<0) End[i]=ns+end[i]; else End[i]=end[i];
          if (End[i]<0) End[i]=0;
          if (End[i]>=ns) End[i]=ns-1; }

        if (End[0]==End[1]) End[0]=0,End[1]=ns-1;
        oldsp=sp; }

      VVV(CM,=r[End[0]],-r[End[1]])
      sum[sp].nend++;
      sum[sp].endend += (endendval=SQR(CM));

      if (I==n) RGret=RGval,endendret=endendval; }
    else
      sum[sp].endend=RGret=endendret=0;
#endif /*# RGYR */
  } /* n */

#ifdef RGYR
  warning=999;
#endif /*# RGYR */

//  VVV(En.current,=current,*3.91941417e-10/box.L)
  VV(En.J,=3.91941417e10/(box.L[0]*box.L[1]*box.L[2]*h)*current)

  if (lag.cond) {
    StaSet(0,lag.cond,2,0);
    StaAdd("Jx [A/m2]",En.J[0]);
    StaAdd("Jy [A/m2]",En.J[1]);
    StaAdd("Jz [A/m2]",En.J[2]); }

  sumn=0;
  StaSet(0,2,2,lag.n);
#ifdef POLAR
  loop (n,0,nspec) if (sum[n].nD) {
    sumn++;
    StaAdd(string("mu_perm[%d]",n),sum[n].D/sum[n].nD);
    /* sumM does not include saturation: el.sumM+=sum[n].D; */
    StaAdd(string("mu_perm[%d]^2",n),sum[n].DD/sum[n].nD);
    StaAdd(string("mu_ind[%d]",n),sum[n].Dpol/sum[n].nD);
    StaAdd(string("mu_ind[%d]^2",n),sum[n].DDpol/sum[n].nD);
    StaAdd(string("mu_tot[%d]",n),sum[n].Dboth/sum[n].nD); 
    /* sumM includes saturation: bug fixed in 2.7h */
    el.sumM+=sum[n].Dboth;
    StaAdd(string("mu_tot[%d]^2",n),sum[n].DDboth/sum[n].nD);
    StaAdd(string("mu_prj[%d]",n),sum[n].Dprj/sum[n].nD);
    StaAdd(string("mu_prj[%d]^2",n),sum[n].DDprj/sum[n].nD); }

  if (sumn) {
    StaAdd("(sum mu_perm)^2",SQR(sumD3));
    StaAdd("(sum mu_ind)^2",SQR(sumD3pol));
    StaAdd("(sum mu_tot)^2",SQR(sumD3both)); }
#else /*# POLAR */
  loop (n,0,nspec) if (sum[n].nD) {
    sumn++;

    StaAdd(string("mu[%d]",n),sum[n].D/sum[n].nD); el.sumM+=sum[n].D; 
    StaAdd(string("mu[%d]^2",n),sum[n].DD/sum[n].nD); }

  if (sumn) StaAdd("(sum mu)^2",SQR(sumD3));
#endif /*#!POLAR */

#ifdef RGYR
  sumn=sumnend=0;
  VO(En.RGm,=0)
  Rgyrqs=*endend=0;
  loop (n,0,nspec) {
    if (sum[n].nRG) {
      Rgyrqs+=sum[n].Rgyrq;
      VV(En.RGm,+=sum[n].RGm)
      sumn+=sum[n].nRG;
      StaAdd(string("Rgxx[%d]",n),sum[n].RGm[0]/sum[n].nRG); 
      StaAdd(string("Rgyy[%d]",n),sum[n].RGm[1]/sum[n].nRG); 
      StaAdd(string("Rgzz[%d]",n),sum[n].RGm[2]/sum[n].nRG); 
      StaAdd(string("Rgyr[%d]^2",n),sum[n].Rgyrq/sum[n].nRG); }
    if (sum[n].nend) {
      *endend+=sum[n].endend;
      sumnend+=sum[n].nend;
      StaAdd(string("endend[%d]^2",n),sum[n].endend/sum[n].nend); } }
  VO(En.RGm,/=sumn)
#endif /*# RGYR */

    //free(sum);

#ifdef RGYR
  if (lag.v) free(velocity);

  if (RGret>0) {
    *endend=sqrt(endendret);
    return sqrt(RGret); }
  else {
    if (sumnend==0) *endend=0; else *endend=sqrt(*endend/sumnend);
    if (sumn==0) return 0; else return sqrt(Rgyrqs/sumn); }
#endif /*# RGYR */
}

void ssdistance(void) /*************************************** distancecheck */
{
  if (ssd0) {

    struct ssd_s *ssd;
    vector *r=cfg[0]->rp,dr;

    StaSet(0,2,2,lag.n);

    for (ssd=ssd0; ssd; ssd=ssd->next) if (ssd->indx[0]>=0) {
      VVV(dr,=r[ssd->indx[0]],-r[ssd->indx[1]])
#ifndef FREEBC
      {
        int k;
        loop (k,0,3) {
          while (dr[k]>box.Lh[k]) dr[k]-=box.L[k];
          while (dr[k]<-box.Lh[k]) dr[k]+=box.L[k]; }
      }
#endif /*# FREEBC */
      ssd->u.dist=sqrt(SQR(dr));
      /* postponed: StaAdd(ssd->name,ssd->u.dist); */ } }
}

#ifdef HARMONICS
#  if HARMONICS==3
#    include "harmwat.c"
#  elif HARMONICS==2
#    include "harmlin.c"
#  elif HARMONICS==1
#    include "harmmeoh.c"
#  else /*#!HARMONICS==3!HARMONICS==2!HARMONICS==1 */
#    error "wrong value of HARMONICS: use 3=water, 2=linear molecule"
#  endif /*#!HARMONICS==3!HARMONICS==2!HARMONICS==1 */
#endif /*# HARMONICS */

/****************************************************************************/

#ifdef WIDOM
/*
  WIDOM insertion and scaled particle (grow molecule, or identity change)
  note that the LINKCELL method calculates the insertiona and grow
  energies on the configurateion BEFORE the current integraion step os
  performed while the standard version AFTER
*/

#  ifdef POLAR
#    warning "WIDOM+POLAR is not a good idea (unless uncharged molecule is inserted)"
#  endif /*# POLAR */
/* code doubled -- unify w. simils !!! */
/* USE HEADERS!!! */
double rndcos();
double rnd();
#  ifdef WALL
double mol_wall(vector f[],molecule_t *m,vector *rp);
#  endif /*# WALL */

extern vector **initsites;

#  ifdef SLAB
double Widom(int sp,int ncyc,double z0,double z1,double dz,int mode)
/*
   mode&1: incl. symmetrization
   mode&2: total chem.pot.
   mode&4: residual (without wall potential)
   mode&8: wall potential (angle-averaged)
*/
#  else /*# SLAB */
double Widom(int sp,int ncyc,double corr)
#  endif /*#!SLAB */
{
  En_t En0=En;
  rdf_t ***rdf0=rdf;
  int icyc,ns=spec[sp]->ns,i;
  static vector *rp,*rpf,*c;
  static molecule_t mol;
  double ps[3];
  double ret=0;
  vector dr;

#  ifdef SLAB
  double z,Ewall=0;
  double (*w)[3];
  int nz=(z1-z0)/dz+0.9999,iz,k;

  if (nz<=0 || ncyc<=0) return 0;
  if (mode&1 && fabs((z1-z0)/dz-nz)>1e-4)
    ERROR(("Widom: z1-z0 must be an integer multiple of dz if mode&1"))

  if (mode&1) allocarray(w,nz+1);
#  else /*# SLAB */
  /* cutoff correction - multiplication factor
     it is V-dependent, thus NPT is allowed though not recommended */
  double qcorr=exp(-corr/(L[0]*L[1]*L[2]*T));

  if (ncyc<=0) return 0;
#  endif /*#!SLAB */

  rdf=NULL; widomrdf(0);

  if (!rp) {
    allocarray(rp,No.s+ns); /* cfg+probe */
    allocarrayzero(rpf,No.s+ns); /* dummy forces */ }
  mol.sp=sp;
  mol.ns=ns;
  mol.nc=spec[sp]->nc;
  mol.ir=No.s*sizeof(vector);
  c=rp+No.s;
  mol.ig=0; /* not needed */

  StaSet(0,2,2,lag.n);

#  ifdef SLAB
  loopto (iz,0,nz) {
    z=z0+iz*dz;
#  endif /*# SLAB */

    ps[0]=ps[1]=ps[2]=0;

#  ifndef LINKCELL
    depend_r(cfg[0]); /* why this? */
    memcpy(rp,cfg[0]->rp,No.s*sizeof(vector));
#  endif /*# LINKCELL */

#  ifdef SLAB
    loop (icyc,0,ncyc) {
      dr[0]=rnd()*L[0];
      dr[1]=rnd()*L[1];
      dr[2]=z;
#  else /*# SLAB */
    loop (icyc,0,ncyc) {
      VV(dr,=rnd()*L)
#  endif /*#!SLAB */

      loop (i,0,ns) VV(c[i],=initsites[sp][i])
      rndorientation(c,ns);

      loop (i,0,ns) VV(c[i],+=dr)

#  ifdef LINKCELL
      if (ns>1) {
        /* the molecule position must be exactly normalized, i.e.,
           no coordinate may be negative; cf. normalize(int m) in norm.c
           (the direct (all-pairs) method is not sensitive
           to small negative cordinates)
           (it is assumed that 1 atom is always centered!!!) */
        vector rmin;
        int k;

        VV(rmin,=c[0])
        loop (i,1,ns) loop (k,0,DIM)
          if (c[i][k]<rmin[k]) rmin[k]=c[i][k];
        loop (k,0,DIM)
          if (rmin[k]<0) loop (i,0,ns) c[i][k]+=L[k]; }
#  endif /*# LINKCELL */

#  ifdef WALL
      En.el=0;
      Ewall=mol_wall(rof(&mol,rpf),&mol,rp);
      Ewall=-(Ewall+En.el)/T;
#  endif /*# WALL */

      /* insertion energy returned in En.pot */
      Einsert(&mol,rpf,rp);
      En.pot/=-T; /* WARNING: Widom requires thermostat! */

#  ifdef SLAB
      if (mode&2) if (En.pot>-700) ps[0]+=exp(En.pot);
      if (mode&4) if (Ewall+En.pot>-700) ps[1]+=exp(En.pot+Ewall);
      if (mode&8) if (Ewall>-700) ps[2]+=exp(Ewall); }

    if (mode&4) ret+=ps[1];
    else if (mode&2) ret+=ps[0];
    else ret+=ps[2];

    loop (k,0,3) {
      int m=2<<k;
      ps[k]/=ncyc;
      if (mode&m) StaAdd(string("Widom%d z=%g",m,z),ps[k]);
      if (mode&1) w[iz][k]=ps[k]; } } /* iz-loop */
    ret/=ncyc*nz;
#  else /*# SLAB */
    if (En.pot>-700) ps[0]+=exp(En.pot); }
  StaAdd("Widom raw",ps[0]/ncyc);
  StaAdd("Widom",ps[0]/ncyc*qcorr);
  ret=ps[0]/ncyc*qcorr;
#  endif /*#!SLAB */

#  ifdef SLAB
  if (mode&1) {
    loop (iz,0,nz/2) {
      z=z0+iz*dz;
      loop (k,0,3) {
        int m=2<<k;
        if (mode&m)
          StaAdd(string("Widom%d z=%g",m+1,z),(w[iz][k]+w[nz-iz][k])/2); } }
    free(w); }
#  endif /*# SLAB */

  En=En0;
  rdf=rdf0; widomrdf(1);

  return ret;
}


double XWidom(int sp,int spvirt) /********************************** XWidom */
{
  En_t En0=En;
  rdf_t ***rdf0=rdf;
  int n,nscale=0;
  static vector *rp,*rpf;
  double ret=0;
  double e,psum=0;

  rdf=NULL; widomrdf(0);

  if (!rp) {
    allocarray(rp,No.s); /* cfg */
    allocarrayzero(rpf,No.s); /* dummy */ }

  StaSet(0,2,2,lag.n);

#  ifndef LINKCELL
  depend_r(cfg[0]); /* why this? */
  memcpy(rp,cfg[0]->rp,No.s*sizeof(vector));
#  endif /*# LINKCELL */

  loop (n,0,No.N) if (cfg[n].sp==sp) {
    nscale++;
    Eidchange(spvirt,&cfg[n],rpf,rp);
    En.pot/=-T; /* WARNING: Widom requires thermostat! */
    e=0; if (En.pot>-700) e=exp(En.pot);
#  ifdef SLAB
    StaAdd(string("XWidom%d",n),e);
#  endif /*# SLAB */
    psum+=e; }

  ret=psum/nscale;
#  ifndef SLAB
  StaAdd("XWidom",ret);
#  endif /*# SLAB */

  En=En0;
  rdf=rdf0; widomrdf(1);

  return ret;
}
#endif /*# WIDOM */

#ifdef SLAB
/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                        density profiles in z-direction
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

static struct dpr_s {
  int4 size;          /* SDS size */
  int n;              /* # of bins; negative value means [0,Lz] scaled to [0,1] */
                      /* should be the same for all density profiles */
  int nmeas;          /* number of measurements */
  int SP;             /* species to autocenter the slab:
                         negative: all species 0..|sp| (incl.)
                         non-negative: species sp
                         out of range: no autocenter */
  char name[8];       /* atom name or mol0,mol1,.. */
  double grid;        /* per 1 A (if n>0) or per Lz (if n<0); all grids are the same */
  double Vbin;        /* averaged volume of one bin */
  double Lz;          /* averaged Lz */
  double mass_charge; /* molecular mass (or charge - not used) */
  union {
    unsigned ihist[1]; /* for site (number) density profiles */
    double   dhist[1]; /* for mass and charge density profiles */
  } hist; /* [n] */
} **dpr; /* indices:
  [0,nspec) : centers-of-mass of molecules, double dhist[]
  [nspec,2*nspec) : charges (by molecules), double dhist[]
  [2*nspec,2*nspec+nsites) : sites by types, unsigned ihist[] */


void initdpr(double grid,double zmax,int SP,char *fn) /************* initdpr */
{
  int n,sp;

  if (grid==0) return;

  underline("density profile");
  if (zmax==0) {
    prt("densprof.zmax = 0: the density profile is calculated with [0,Lz]\n\
(rescaled to [0,1]) divided into %d bins",grid);
    n=(int)grid;
    if (n!=grid) ERROR(("grid is not integer"))
    n=-n; }
  else
    n=zmax*grid+0.9999999999999;
  prt("initdpr: densprof.grid=%g densprof.zmax=%g n=%d",grid,zmax,n);
  if (abs(n)<2) WARNING(("only 1 histogram bin in the z-profile"))

  if (SP<nspec) {
    if (SP>=0) prt("density profile autocenter using CoM of species %d",SP);
    else prt("density profile autocenter using CoM of all species 0..%d (incl.)",-SP); }

  rallocarray(dpr,2*nspec+nsites);
  loop (sp,0,2*nspec+nsites) {
    int elemsize=
      sp<2*nspec ? sizeof(dpr[sp]->hist.dhist[0])
                 : sizeof(dpr[sp]->hist.ihist[0]);
    sdsralloczero(dpr[sp],sizeof(dpr[sp][0])+elemsize*(abs(n)-1));
    dpr[sp]->n=n; /* must be the same for all density profiles */
    dpr[sp]->grid=grid; /* must be the same for all density profiles */
    dpr[sp]->SP=SP; /* must be the same for all density profiles */
    if (sp<nspec) sprintf(dpr[sp]->name,"mol%d",sp);
    else if (sp<2*nspec) sprintf(dpr[sp]->name,"mol%d",sp-nspec);
    else strcpy(dpr[sp]->name,sitedef[sp-2*nspec].name); }

  loop (sp,0,nspec) {
    /* unnecessarily recalculated (with check) */
    int i;
    double m=0,q=0;

    loop (i,0,spec[sp]->ns) {
      siteinfo_t *si=spec[sp]->si+i;
      m += si->mass;
      q += si->charge;
#  ifdef POLAR
      q += si->chargepol; /* bug fixed by T.Trnka */
#  endif /*# POLAR */
    }
    dpr[sp]->mass_charge=m;
    dpr[sp+nspec]->mass_charge=q; /* not used */
    if (fabs(dpr[sp]->mass_charge-spec[sp]->mass)>1e-8) ERROR(("internal"))
    if (fabs(dpr[sp+nspec]->mass_charge-spec[sp]->charge)>1e-8) ERROR(("internal"))
  }
}

void loaddpr(double dprgrid) /************************************** loaddpr */
{
  double gr;
  int sp,nsp;

  if (dprgrid==0) return;
  if (!dpr) ERROR(("internal"))

  VarOpen(Fn("dpr"),"r");
  VarRead(&gr,sizeof(gr));
  if (gr!=dprgrid)
    ERROR(("%s: inconsistent grid (%g in file, %g in data)",lastFn,gr,dprgrid))
  VarRead(&nsp,sizeof(nsp));
  if (nsp!=nspec)
    ERROR(("%s: inconsistent # of species (%d in file, %d in data)",lastFn,nsp,nspec))
  loop (sp,0,2*nspec+nsites) VarReadSds(dpr[sp]);
  VarClose();
}

void savedpr(double dprgrid) /************************************** savedpr */
{
  int sp;

  if (dprgrid==0) return;
  if (!dpr) ERROR(("internal"))

  backup("dpr");
  VarOpen(Fn("dpr"),"w");
  VarPut(&dprgrid,sizeof(dprgrid));
  VarPut(&nspec,sizeof(nspec));
  loop (sp,0,2*nspec+nsites) VarPutSds(dpr[sp]);
  VarClose();
}

void measuredpr(void) /****************************************** measuredpr */
{
  int n,nbins;
  double z,zshift=0;

#  ifdef FREEBC
  ERROR(("implementation limitation: cannot measure z-profile in free b.c."))
#  endif /*# FREEBC */

  if (!dpr) return;
  nbins=abs(dpr[0]->n); /* all sizes assumed the same */

  /* autocenter density profile (for measurements only) */
  if (dpr[0]->SP<nspec) {
    double s=0,c=0;

    En.zshift=0;
    /* slab autocenter */
    loop (n,0,No.N) {
      double zcm=0; /* z-position of the center of mass */
      molecule_t *mn=molec+n;
      int sp=mn->sp;
      siteinfo_t *si=spec[sp]->si;
      int ns=mn->ns;
      vector *r=rof(mn,cfg[0]->rp);
      int i;

      if (dpr[sp]->SP<0) { if (sp>abs(dpr[sp]->SP)) continue; }
      else { if (sp!=dpr[sp]->SP) continue; }
      loop (i,0,ns) zcm+=si[i].mass*r[i][2];

      zcm*=2*PI/dpr[sp]->mass_charge/box.L[2];
      s+=sin(zcm);
      c+=cos(zcm); }

    En.zshift=atan2(c,s)/(2*PI)+0.25;
    if (En.zshift<0) En.zshift+=1;
    zshift=En.zshift*box.L[2]; 
    /* shifted by 0.5 for measurement/recording: */
    En.zshift-=0.5; if (En.zshift<0) En.zshift+=1; }

  loop (n,0,No.N) {
    double zcm=0; /* z-position of the center of mass */
    molecule_t *mn=molec+n;
    int sp=mn->sp;
    siteinfo_t *si=spec[sp]->si;
    int ns=mn->ns;
    vector *r=rof(mn,cfg[0]->rp);
    int i,nhsite,nhmol;
#  ifdef POLAR
    vector *rpol=(vector*)((char*)r+polar_off); /* relative to r, chargepol */
#  endif /*# POLAR */

    loop (i,0,ns) {
      z=r[i][2]+zshift;
      zcm+=si[i].mass*z;
      while (z>=box.L[2]) z-=box.L[2];
      /* NB: we assume all grids are the same ... */
      if (dpr[sp]->n>0) nhsite=z*dpr[0]->grid;
      else nhsite=(int)(z*(dpr[0]->grid/box.L[2]));
      if (nhsite<0)
        WARNING(("mol%d.%d negative z=%g in z-profile ignored",n,i,z))
      else {
        nhsite%=nbins;
        /* site density profile */
        dpr[2*nspec+si[i].st]->hist.ihist[nhsite]++;
        /* charge density profile */
#  ifdef POLAR
        if (si[i].chargepol==0)
          dpr[nspec+sp]->hist.dhist[nhsite]+=si[i].charge;
        else {
          /* charge density profile - Drude oscillator */
          int nminus,nplus;
          double qpol;

          if (dpr[sp]->n>0) {
            qpol=rpol[i][2]*si[i].chargepol*dpr[0]->grid;
            nminus=(z+rpol[i][2]/2)*dpr[0]->grid-0.5;
            nplus=nminus+1; }
          else {
            qpol=rpol[i][2]*si[i].chargepol*(dpr[0]->grid/box.L[2]);
            nminus=(int)((z+rpol[i][2]/2)/box.L[2]*dpr[0]->grid-0.5+nhsite)%nbins;
            nplus=(nminus+1)%nbins; }
          /* short dipole extended to grid-based dipole */
          if (nminus>=0)        dpr[nspec+sp]->hist.dhist[nminus]-=qpol;
          if (nplus<nbins) dpr[nspec+sp]->hist.dhist[nplus] +=qpol;

          /* partial charge part */
          dpr[nspec+sp]->hist.dhist[nhsite]+=si[i].charge+si[i].chargepol; }
#  else /*# POLAR */
        /* non-polar */
        dpr[nspec+sp]->hist.dhist[nhsite]+=si[i].charge;
#  endif /*#!POLAR */
      } }

    z=zcm/dpr[sp]->mass_charge;
    while (z>=box.L[2]) z-=box.L[2];
    if (dpr[sp]->n>0) nhmol=z*dpr[sp]->grid;
    else nhmol=(int)(z*(dpr[0]->grid/box.L[2]));
    if (nhmol<0)
      WARNING(("mol%d.%d negative z=%g in z-profile ignored",n,i,z))
    else {
      nhmol%=nbins;
      dpr[sp]->hist.dhist[nhmol]+=dpr[sp]->mass_charge; } }

/* statistics for [0] only */
  if (dpr[0]->n>0)
    dpr[0]->Vbin+=box.L[0]*box.L[1]/dpr[0]->grid;
  else
    dpr[0]->Vbin+=box.L[0]*box.L[1]*box.L[2]/dpr[0]->grid;
  dpr[0]->Lz+=box.L[2];
  dpr[0]->nmeas++;
}

void printdpr(int densprof_slab,double dV,char *PdVname) /********* printdpr */
/*
  print density profiles and report surface tension results
*/
{
  int sp,i,nbins;
  FILE *f;
  double s,satom;
  int number; /* 0: use [kg.m-3] or [e.A-3], 1: use [A-3] */
  char *fmt[]={" %10.4f", " %10.7f", " %11.8f"};
            /*  [kg.m-3]   [A-3]    [e.A-3] */
  double DZ;

  if (!dpr) return;
  nbins=abs(dpr[0]->n);

  if (dpr[0]->n>0) DZ=1./dpr[0]->grid;
  else DZ=dpr[0]->Lz/(dpr[0]->nmeas*dpr[0]->grid);

  loop (number,0,2) {

    /*** molecules: center-of-mass density profile ***/
    f=fopen(Fn(number?"cm.A-3.z":"cm.kgm-3.z"),"wt");
    fprintf(f,"# DENSITY PROFILES BY CENTER-OF-MASS OF MOLECULES\n\
#  calculated from %s\n\
#  z is in A, density profiles are in %s\n",
            Fn("dpr"),number?"A-3":"kg.m-3");
    if (dpr[0]->n<0) fprintf(f,"# accumulated in [0,Lz], using <Lz>=%.9g, dz=%.9g\n",dpr[0]->Lz/dpr[0]->nmeas,DZ);

    fprintf(f,"#    z   ");
    loop (sp,0,nspec) fprintf(f,"  %c=%-6s",sp+'B',dpr[sp]->name);
    fprintf(f,"    SUM\n");

    loop (i,0,nbins) {
      fprintf(f,"%6.3f ",(i+0.5)*DZ);

      s=0;
      loop (sp,0,nspec) {
        double x=dpr[sp]->hist.dhist[i]/dpr[0]->Vbin;

        if (number) x/=dpr[sp]->mass_charge;
        else x*=rhounit;
        fprintf(f,fmt[number],x);
        s+=x; }
      fprintf(f,fmt[number],s);
      fprintf(f,"\n"); }

    fclose(f);


    /*** sites-based density profile ***/
    f=fopen(Fn(number?"site.A-3.z":"site.kgm-3.z"),"wt");
    fprintf(f,"# DENSITY PROFILES BY SITES\n\
#  calculated from %s\n\
#  SUM(all)=sum over all sites, SUM(atoms)=only sites with nonzero mass\n\
#  z is in A, density profiles are in %s\n",
            Fn("dpr"),number?"A-3":"kg.m-3");
    if (dpr[0]->n<0) fprintf(f,"# accumulated in [0,Lz], using <Lz>=%.9g, dz=%.9g\n",dpr[0]->Lz/dpr[0]->nmeas,DZ);

    fprintf(f,"#    z   ");
    loop (sp,nspec*2,nspec*2+nsites)
      fprintf(f,"  %c=%-6s",sp-2*nspec+'B',dpr[sp]->name);
    fprintf(f,"  SUM(all) SUM(atoms)\n");

    loop (i,0,nbins) {
      fprintf(f,"%6.3f ",(i+0.5)*DZ);

      satom=s=0;
      loop (sp,2*nspec,2*nspec+nsites) {
        double x=dpr[sp]->hist.ihist[i]/dpr[0]->Vbin;
        double m=sitedef[sp-2*nspec].M/munit;

        if (!number) x*=m*rhounit;

        fprintf(f,fmt[number],x);
        s+=x;
        if (m) satom+=x; }
      fprintf(f,fmt[number],s);
      fprintf(f,fmt[number],satom);
      fprintf(f,"\n"); }

    fclose(f); }

  number=2; /* should be =2 after the above loop .. */

  /** charge density profile */
  f=fopen(Fn("q.eA-3.z"),"wt");
  fprintf(f,"# CHARGE DENSITY PROFILE BY MOLECULES\n\
#  calculated from %s\n\
#  z is in A, density profile is in e.A-3\n",Fn("dpr"));
  if (dpr[0]->n<0) fprintf(f,"# accumulated in [0,Lz], using <Lz>=%.9g, dz=%.9g\n",dpr[0]->Lz/dpr[0]->nmeas,DZ);

  fprintf(f,"#    z   ");
  loop (sp,nspec,2*nspec) fprintf(f,"  %c=%-6s",sp-nspec+'B',dpr[sp]->name);
  fprintf(f,"    SUM\n");

  loop (i,0,nbins) {
    fprintf(f,"%6.3f ",(i+0.5)*DZ);
    //    put_((i+0.5)/dpr[0]->grid)
    s=0;
    loop (sp,nspec,2*nspec) {
      double x=dpr[sp]->hist.dhist[i]/dpr[0]->Vbin;

      x/=electron;
      //      put_(x)
      fprintf(f,fmt[2],x);
      s+=x; }
    //    put2(number,s)
    fprintf(f,fmt[number],s);
    fprintf(f,"\n"); }

  fclose(f);


  /* surface tension */
  if (densprof_slab&1) {
    double Lzf=0.75e-10*dpr[0]->Lz/dpr[0]->nmeas;
    double c[3]={slabcorr(0),slabcorr(1),slabcorr(2)};
    double Esta=StaMean("Epot [\t"),Psta=StaMean("P [Pa]");
    double sc[3],Pc,dP=StaStdErr("P [Pa]")*Lzf;
    double V=box.L[0]*box.L[1]*box.L[2];

    underline("surface tension");
    prt_("\
====== slab-based cutoff corections to Epot, P, Pt=(Pxx+Pyy-2*Pzz)/3 ======\n\
raw slab corrections: c[0]=%g  c[1]=%g  c[2]=%g\n\
raw homogeneous fluid: corr=%g  corr/V=%g  corr/V^2=%g\n\
(homogeneous fluid corrections are already included in Epot and some reported\n\
P or Pt and have to be subtracted before adding the slab corrections)",
            c[0],c[1],c[2],  En.corr,En.corr/V,En.corr/(V*V));

    header("                               Epot [J/mol]        P [Pa]         Pt [Pa]  ");
    prt("homogeneous fluid correction %14.9g  %14.9g  %14.9g",
        En.corr/V*Eunit,En.corr/(V*V)*Punit,En.corr/(V*V)*Punit);
    prt("uncorrected values           %14.9g  %14.9g",
        Esta-En.corr/V*Eunit,Psta-En.corr/(V*V)*Punit);
    prt("slab-based cutoff correction %14.9g  %14.9g  %14.9g",
        c[0]*Eunit,c[1]*Punit/V,c[2]*Punit/V);
    prt("correction to be added       %14.9g  %14.9g  %14.9g",
        sc[0]=(c[0]-En.corr/V)*Eunit,
        sc[1]=(c[1]/V-En.corr/(V*V))*Punit,
        sc[2]=(c[2]/V-En.corr/(V*V))*Punit);
    prt("final corrected value        %14.9g  %14.9g  %14.9g",
        Esta+sc[0],Psta+sc[1],Pc=Psta+sc[2]);
    prt("corr_slab/corr_homog.fluid   %12.5f %15.5f %15.5f\n\
thick slab limits                    1               1              -",
         c[0]*V/En.corr,c[1]*V/En.corr,c[2]*V/En.corr);
    header("");

    prt("\n\
====== slab-based cutoff corections to surface tension ======\n\
NOTE: gamma/[N/m] = -Pt/[Pa]*%g = -P/[Pa]*%g (for NPT in x,y)",Lzf,Lzf/1.5);
    prt("correction (to be added to uncorrected value):     %11.6f N/m",
        -c[2]/V*Punit*Lzf);
    prt("wrong homogeneous correction:                      %11.6f N/m",
        En.corr/(V*V)*Punit*Lzf);
    prt("fix (to be added to homogeneously-corrected value):%11.6f N/m",
        -sc[2]*Lzf);

    header(" surface tension in N/m:       corrected     stdev   uncorrected");
    prt("from Pvir (assumes Psat=0)  %11.6f %11.6f %11.6f (a,b)",
        -Pc*Lzf, dP, -(Psta-En.corr/(V*V)*Punit)*Lzf);
    if (dV) {
      char shortname[8];

      memcpy(shortname,PdVname,7); shortname[7]=0;
      prt("%s%-7s) %11.6f %11.6f %11.6f%s",
          densprof_slab&2?"test-area (tangent ":
                          "virtual V change  (",
          shortname,
          -(StaMean(PdVname)+sc[2])*Lzf,
          StaStdErr(PdVname)*Lzf,
          -(StaMean(PdVname)-En.corr/(V*V)*Punit)*Lzf,
          densprof_slab&2?"":" (a)"); }
#  if PRESSURETENSOR&PT_VIR
    prt("from virial part of Pt      %11.6f %11.6f %11.6f (c)",
        -(StaMean("Pvir xy-zz [Pa]")+c[2]/V*Punit)*Lzf,
        StaStdErr("Pvir xy-zz [Pa]")*Lzf,
        -StaMean("Pvir xy-zz [Pa]")*Lzf);
#    if PRESSURETENSOR&PT_KIN
    prt("from full Pt (incl.kin.part)%11.6f %11.6f %11.6f",
        -(StaMean("Pt [Pa]")+c[2]/V*Punit)*Lzf,
        StaStdErr("Pt [Pa]")*Lzf,
       -StaMean("Pt [Pa]")*Lzf);
#    endif /*# PRESSURETENSOR&PT_KIN */
#  endif /*# PRESSURETENSOR&PT_VIR */
    header("");
    prt("(a) assumes <Pzz>=0: subtract %g*(saturated pressure in Pa)",Lzf);
    prt("(b) assumes elst.virial = -elst.energy: "
#  if COULOMB<0
        "see variable Pvirerr for accuracy"
#  else /*# COULOMB<0 */
        "imprecise for short-ranged electrostatics"
#  endif /*#!COULOMB<0 */
                                                   );
    prt("(c) valid only for monoatomic models or vibrating bonds (cook -u9999)\n\
    incorrect for constrained bonds or rigid molecules");
#  if (PRESSURETENSOR&PT_VIR) && !(PRESSURETENSOR&PT_KIN)
    prt("    (cook* compiled with PRESSURETENSOR&%d needed for this case)",PT_KIN);
#  endif /*# (PRESSURETENSOR&PT_VIR) && !(PRESSURETENSOR&PT_KIN) */
    _n }
}

static
double slabpaircorr(int is,int js,int isP) /******************* slabpaircorr */
/***
  Cutoff corections calculation in the slab geometry for one pair of atom types:
    isP=0: energy correction returned,
    isP=1: isotropic pressure correction returned (not needed),
    isP=2: pressure-equivalent correction for surface tension returned.
  Cf. setss() in setss.c;
    unlike setss(), slabpaircorr should be called after the density profile
    has been calculated.
  NOTE: gamma = -3/(4 Lx Ly)*(P-Pvapor)
    for isP=1,2: Delta P = -4/3* Lx Ly Delta gamma is calculated
  NEW 2.6c: extended to sum of periodic images in z upto 2*Lz
***/
{
#  ifndef COOK
#    error not COOK not supported
#  endif /*# COOK */

#  ifdef NIBC
#    error SLAB + NIBC not supported
#  endif /*# NIBC */

#  ifdef SHARPCUTOFF
  return 0;
#  else /*# SHARPCUTOFF */

  DECLARE_PARMS

  sitesite_t *ss=&sstab[is][js];
  int i,j,ig,n,nn,nbins,NBINS;
  double sumi,sumj,h,dz,xi,sum,x,rr,y,U,f,Urep,frep,z,xmax,hh,DZ;
  double *Iabsz;

  if (!dpr) return 0;
  nbins=abs(dpr[0]->n);
  NBINS=nbins*3; /* will integrate over z-periodic images */

  if (ss->A==0 || ss->C1q==0) return 0;

  allocarrayzero(Iabsz,NBINS); /* to Lz*2 */

  if (dpr[is]->n!=dpr[0]->n || dpr[js]->n!=dpr[0]->n)
    ERROR(("z-profiles of different sizes not supported"))

  /* r-integrals in advance (because time-consuming):
    isP=0: Iabsz(Dz) = INT_0^infty 2 PI r dr Du(sqrt(r^2+Dz^2))
    isP=1: Iabsz(Dz) = INT_0^infty 2 PI r dr r*f_ij(sqrt(r^2+Dz^2))/3
    isP=2: Iabsz(Dz) = INT_0^infty 2 PI r dr (r^2-2z^2)/r*f(sqrt(r^2+Dz^2))/3
    method: Gauss 2 point formula (error ~ h^4)
  */
#    define NGauss4 64
#    define Gauss4 0.2113248654051871

  hh=1/sqrt(ss->C1q)/NGauss4;

  if (dpr[0]->n>0) DZ=1./dpr[0]->grid;
  else DZ=dpr[0]->Lz/(dpr[0]->nmeas*dpr[0]->grid);

  sum=0;
  loop (i,0,NBINS) {
    //    dz=(double)i/dpr[0]->grid;
    dz=i*DZ;
    xmax=1/fmax(sqrt(ss->C1q),dz);
    nn=(int)(xmax/hh+1.99);
    h=xmax/nn;
    sum=0;

    loop (n,0,nn) loop (ig,0,2) {
      if (ig) xi=(n+1-Gauss4)*h;
      else xi=(n+Gauss4)*h;
      rr=1/Sqr(xi);
      /* cf. LJM in internp.c */
      SS_MEASURE_rep
      SS_MEASURE
      if (rr < ss->C2q) {
        x=rr-ss->C2q; U-=x*x*ss->A; f-=ss->A4*x; }
      switch (isP) {
        case 0: sum+=U*rr/xi; break;
        case 1: sum+=Sqr(rr)/xi*f/3; break;
        case 2: sum+=rr*(rr-3*Sqr(dz))/xi*f/3; } }
    Iabsz[i]=sum*PI*h; }

  /* optimize by calculating in advance ... */
  sumi=0; loop (i,0,nbins) sumi+=dpr[2*nspec+is]->hist.ihist[i];
  sumj=0; loop (j,0,nbins) sumj+=dpr[2*nspec+js]->hist.ihist[j];

  sum=0;
  loop (i,0,nbins)
    loop (j,0,nbins) {
      int k;
      double s=0;

      for (k=i-j-NBINS; k<NBINS; k+=nbins) if (abs(k)<NBINS) s+=Iabsz[abs(k)];
      sum+=(double)dpr[2*nspec+is]->hist.ihist[i]
          *(double)dpr[2*nspec+js]->hist.ihist[j]*s; }

  if (option('v')&4) {
    loop (i,0,NBINS) prt("%g %g Iabsz%d",(i+0.5)*DZ,Iabsz[i],isP);
    _n }

  free(Iabsz);

  if (sumi*sumj==0) {
    if (sum!=0) ERROR(("slabpaircorr: wrong sums"))
    return 0; }

  return sum/(sumi*sumj);
#  endif /*#!SHARPCUTOFF */
} /* slabpaircorr */

double slabcorr(int isP) /**************************************** slabcorr */
/***
  Counterpart of cutcorr() in siminit.c for slab geometry:
    isP=0: energy correction returned,
    isP=1: isotropic pressure correction returned (not needed),
    isP=2: pressure-equivalent correction for surface tension returned.
  However, this function should be called at the end of the simulation run!!!
  NOTE: gamma = -3/(4 Lx Ly)*(P-Pvapor)
    for isP=1,2: Delta P = -4/3* Lx Ly Delta gamma is calculated
***/
{
  int n,m,i,j,np;
  double c,corr=0;

#  ifdef SHARPCUTOFF
  WARNING(("SLAB and SHARPCUTOFF: no slab cutoff correction"))
  return 0;
#  endif /*# SHARPCUTOFF */

  loop (n,0,nspec)
    loopto (m,0,n) {
      if (m==n) np=spec[n]->N*(spec[n]->N-1)/2;
      else np=spec[n]->N*spec[m]->N;
      c=0;
      loop (i,0,spec[n]->ns) loop (j,0,spec[m]->ns)
        c += slabpaircorr(spec[n]->si[i].st,spec[m]->si[j].st,isP);
      corr+=c*np; }

  return corr/(box.L[0]*box.L[1]);
  /* pressure correction (isP=1,2) should be divided by V */

} /* slabcorr */

#endif /* SLAB */ /*# SLAB */
