#ifndef UNITS_H
#define UNITS_H

#ifdef REDUNITS
/* Reduced units - all factors=1. Not tested recently!!! */
#define massunit 1
#define lengthunit 1
#define timeunit 1
#define munit 1
#define Eunit 1
#define Punit 1
#define rhounit 1

#define kcal 1
#define Debye 1
#define electron 1

#else
/* real units */

/*
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %                       units of measurement                      %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

MACSIMUS may records the results either in the internal program units or
SI units 

In many cases the user output is in SI units, which is indicated in
[brackets].  If no units are given, the internal program units are used.

WARNING: The author of this program thinks that anybody using non-SI units
should be fined and if does not pay then hanged by legs in a chilly draft.
Unfortunately, direct using of SI units would cause over- and underflows in
calculations so that special units had to be used internally in the program.
In addition, some multiplications are spared if the unit of charge is chosen
so that the electrostatic energy is charge*charge/length

===============================================================================
quantity    | program unit |value in SI and other units
------------+--------------+---------------------------------------------------
temperature |K             |1 K
length      |A             |Angstroem = 1e-10 m
time        |ps            |1e-12 s
energy      |k*K           |1.3806504e-23 J = 8.314472 J/mol= 1.9872065 cal/mol
            |              |
velocity    |A/ps          |100 m/s
mass        |k*K/(A/ps)^2  |1.3806504e-27 kg = 0.8314472 g/mol
mass density|k*K*ps^2/A^5  |1380.6504 kg/m^3
pressure    |k*K/A^3       |13806504 Pa
charge      |(k*K*A)^1/2   |SI: (k*K*A*1e7*F/m)^(1/2)/{c} = 3.91941417e-22 C
            |              |where c={c} m/s, {c}=299792458, and F=Farad=C^2/J
            |              |CGS: 1.17501081e-12 g^1/2 cm^3/2 s^-1
dipole      |A*(k*K*A)^1/2 |SI: 3.91941417e-32 C m
moment      |              |CGS: 1.17501081e-20 g^1/2 cm^5/2 s^-1
            |              |     = 0.0117501081 Debye
polariz-    |A^3           |SI: 10e-23 {c}^-2 C m^2/V = 1.112650056e-40 C m^2/V
ability     |              |CGS: 10e-24 cm^3
===============================================================================

NOTES:
  loosely, energy is measured in Kelvins
  k is the Boltzmann constant
  1 cal = thermochemical calorie = 4.184 J
  program electrostatic energy is q1*q2/r
  charge of electron is 408.779582 charge units, e^2=167100.747 charge units^2
  1 mol = 6.02214179e23 (30) (=Avogadro number)
  SI polarizability volume = SI polarizability/(4 PI epsilon_0)
  numerical factors are based on CODATA 2006,
    see http://physics.nist.gov/cuu/Constants/index.html
  Note that sometimes the values are inconsistent within error bars

useful constants:
  1/(4 PI epsilon_0) = {c}^2 1e-7 = 8987551787 [Jm/C^2=m/F]
  1e = 1.602176487e-19 (40) C = 4.80320427e-10 g^1/2 cm^3/2 s^-1
     = 96485.3399 (24) C/mol (Faraday constant)
  1eV = 96485.3399 J/mol = 23.060550 kcal/mol
  e^2/A = 1.389354558 MJ/mol = 332.063709 kcal/mol = 2.307077128 e-18 J
  1eA = 4.80320427 Debye
  1Cm = 2.99792458e+29 Debye
  me = 9.10938215e-31 kg
  h=6.62606896e-34 Js
  a=0.52917720859 A (Bohr atomic unit of length = (h/2 pi)^2 4 pi eps0/e^2/me)
  1 Hartree = 2625499.618 J/mol = 627.509469 kcal/mol = 4.35974394e-18 J
  mu= 1.660538782 e-27 kg
  c=299792458 m/s
  G=6.67428e-11 m^3 kg^-1 s^-1
  1/alpha=137.03599968

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*** basic SI factors ***/
#define massunit 1.3806504e-27 /* kg */
#define lengthunit 1e-10       /* m */
#define timeunit   1e-12       /* s */
#define chargeunit 3.91941417e-22 /* C (some quantities may contain 4pi) */
#define energyunit 1.3806504e-23 /* J */
#define Avogadro (1e-3*munit/massunit) /* 1/mol */
/* Examples:
   1) viscosity has dimensionality  mass length^-1 time^-1 = pressure*time
      the program unit of viscosity is thus
        viscunit=massunit/lengthunit/timeunit
                =Punit*timeunit
                =1.3806504e-05 Pa.s
        viscosity[SI] = viscunit * programviscosity  
   2) electric current density has dimensionality charge length^-2 time^-1 
      the program unit of current density is thus
        Junit=chargeunit/lengthunit^2/timeunit
        3.91941417e+10 A/m2
      J[SI] = Junit * programJ 
*/

/*** some derived program units in SI ***/
#define munit 0.8314472    /* g/mol */
#define Eunit 8.314472     /* J/mol */
#define Punit 13806504.0   /* Pa */
#define rhounit 1380.6504  /* kg/m^3 */

#define Debye 0.0117501081 /* this is 1 Debye in program units */
#define kcal 4184.0        /* J (some criminals still do use calories) */

#define electron 408.779582 /* charge of electron in program units */


#endif
#endif /* UNITS_H */
