/* \make plb2diff

07/2011: env.variable PLB2DIFF==3: fit MSD(t) to a+bt+c/t^1/2
06/2005: unsuccessful attempt to extend to viscosity via Einstein relation
         (it is likely impossible)
10/2000: interface to new option -g of cook

This utility automates diffusion (D) and conductivity (kappa) calculations.
See also `sim/new', section
  "6/00 conductivity and autodiffusion measurements by Einstein relation"

Requirements:
~~~~~~~~~~~~~
  Periodic b.c.
  Stored configurations:
    either in a plb-file (SIMNAME.plb, generated by cook* -y-1)
    or SIMNAME.1,SIMNAME.2,... (generated by cook* -r)
  Utility plot to view graphs

Synopsis - call plb2diff without args!
~~~~~~~~

More info:

  COMMAND should be a script accepting one arg=SIMNAME, or a call to
  cook* with all necessary arguments but SIMNAME; in this case it
  should be enclosed in "".  It must call the corresponding version of
  cook with option -f (if SIMNAME.plb is to be read)
  Options -f2, -f-2, -f3, ... are NOT supported - plb2diff will crash!!!
  [Option -f-1 (SIMNAME.1,...have been saved) has not been tested]

  NO is the total # of frames (configurations) or less, see
  SIMNAME.prt or use
    plbinfo SIMNAME.plb
  [[or see the last SIMNAME.###]]

  START is just to skip unequilibrated portion of the trajectory

  One call to cook* -f generates one curve of msd, mscd, or cond (see sim/new
  for their definiton).  msd and mscd are only averaged over atoms, but cond
  is not averaged in principle (better said, the only averaging is over 3
  Cartesian coordinates).  To obtain ENSEMBLE AVERAGES, we divide the
  trajectory into NBLOCKS subtrajectories (blocks) of length BLOCK, analyze
  them (by separate calls to cook* -f), and make averages and error analysis.
  It is recommended that the blocks overlap; see the `coverage' info
  at the beginning of the output file SIMNAME.cdf!  The recommended coverage
  is about 3 (see below).

  OMIT is the number of points of MSD(t) omitted in regression analysis
  from the beginning of each block. OMIT=0 means that MSD(0)=0 is included.

  MFACTOR and QFACTOR should be calculated in advance from the data in
  SIMNAME.prt.  If they are omitted, the final results of D and kappa
  are not available but must be calculated from other results.

Results:

  SIMNAME.m0.dat, SIMNAME.m1.dat ... : results of linear regression of
    individual atom msd of all blocks.  The points (column 2 on the files)
    are the slopes with OMIT points from the block omitted, i.e. after
    dividing by dt.plb [[or dt.cfg]] the diffusion coefficients D.
    Note that the larger coverage, the more correlated data.
  SIMNAME.mdif.dat : these data should be very small because of momentum
    conservation!!!
  These files are plotted in the first window named "diffusion SIMNAME".
  Three curves of the same color correspond to linear regression error bars.

  SIMNAME.q0.dat, SIMNAME.q1.dat, as SIMNAME.m0.dat,... with charge weights,
    see mscd
  SIMNAME.cond.dat : total conductivities estimated from blocks multiplied by
    dt.plb*T*V.
  These files are plotted in the 2nd window named "conductivity SIMNAME"

  SIMNAME.avm.dat : estimated diffusion coefficients multiplied by dt.plb.
    Column1 = species 0, last column = difm (should be around zero)
  This file (all columns) is plotted in the 3rd window named "<msd(t)
    SIMNAME>"

  SIMNAME.avq.dat : estimated partial conductivities and conductivity (last
    column=cond) multiplied by dt.plb*T*V.
  This file (all columns) is plotted in the 4th window named "<mscd(t)
    SIMNAME>"

  If the last two plots are bent close to origin, OMIT should be increased to
  start from a linear portion.

  SIMNAME.cdf :
    1/ coverage = see above
    2/ block info
    3/ averages and error analysis over linear regression results of
       individual blocks (=data in SIMNAME.m0.dat, ..., SIMNAME.visc.dat).
       The errors should be estimated from the statistical analysis!!!
       NEW: Approximate error estimates (stdef*sqrt(coverage)) are given.
    4/ linear regression of averaged data (=files SIMNAME.av[mq].dat)
       NOTE: the `l.r.err' should NOT be used as the estimate of error
       because the data are correlated!!!

Problems:
  If difm is not around zero, a warning is printed.  It may be caused by:
  - too long dt.plb so that particles travel by more than L/2 between
    frames and the algorithm cannot follow the trajectory over
    periodic images.  A separate warning may be printed if jumps close
    to L/2 are found, these warnings can be safely ignored if difm
    oscillates uniformly around zero and is tiny
  - changes of parameters during simulation (like noint, dt.plb, tau.P,...)
  - a restart from crash (more frames in a plb-file)
  - wrong SIMNAME.cfg (only with tau.P)
  - wrong newt used instead of SIMNAME.cfg (only with tau.P)
  - wrong (shifted) convergence profile, .cp x .cpz clash (only with tau.P)

  Error `XXX truncated' may be caused by -f# with #>1; use -f1 or -f-1 only.

Optimization:
  Using 256 molecules of tip4pew water, 80ps blocks and 10ps omitted,
  3ns total length,  the following dependence of the relative error on
  the "true" coverage was obtained:
# coverage  error/stdev
  10        0.71471
   5        0.71683
   3.33333  0.72299
   2.5      0.73523
   2        0.74933
   1.66667  0.74498
   1.42857  0.79346
   1.25     0.82888
   1.11111  0.88120
   1        0.98133
   0.83333  1.13038
   0.625    1.25066
   0.5      1.39198
   0.4      1.58114
   1        1 (independent blocks)
It means that coverage around 3 gives about 0.73 times smaller error
  than if the same data were generated as independent blocks (i.e.,
  almost 0.73^2~0.5 CPU time is needed).
Efficiency inreases by marginal 1-2% if the coverage is increased to 5
  or more.

Bugs:
  MFACTOR and QFACTOR are not calculated automatically
  Viscosity is under development, see sfdx.c
*/

#include "ground.h"
#include "alloc.h"
#include "linregr.h"
#include "sdfit.h"
#include "cpmark.h"
#include "statics.h"
#include "maxjump.h"
#include <unistd.h>

FILE *sh;

char *vname(int icp)
{
  return
      icp<9   ? string("avv%c%c",icp%3+'x',icp/3+'x')
    : icp==9  ? "visc"
    : icp==10 ? "vmid"
    : icp==11 ? "vis0"
    : icp==12 ? "visl"
    : icp==13 ? "vcon"
    : "ERR ";
}

void Setenv(char *name,char *value)
{
  if (setenv(name,value,1)) ERROR(("setenv(%s,%s,1)",name,value))
  fprintf(sh,"export %s=\"%s\"\n",name,getenv(name));
}

int plbno(char *fn)
{
  FILE *f=fopen(fn,"r");
  float h[2];
  int len;

  if (!f) ERROR(("cannot open %s",fn))
  if (fread(h,4,2,f)!=2) ERROR(("%s too short",fn))
  fseek(f,0,SEEK_END);

  len=ftell(f)/(long)(((int)h[0]+(h[1]==-3.))*12);
  fclose(f);

  fprintf(stderr,"%d frames detected in %s\n",len,fn);

  return len;
}

int main(int narg,char **arg)
{
  int NO,START=0,BLOCK,OMIT,NBLOCKS=0;
  double MFACTOR=0,QFACTOR=0,VFACTOR=0,DBLOCK;
  int PLOT=15;
  int init,no,iblock,ncps[128],ncp,icp,pass;
  int ir;
  FILE *get=NULL,**mf=NULL,**qf=NULL,**vf=NULL;
  float *r=NULL;
  double **mr=NULL,**qr=NULL,**vr=NULL,**xr;
  struct sum_s {
    double s,q;
    int n; } *summ,*sumq,*sumv=NULL,*sum;
  double ERR;
  char *COMMAND,*SIMNAME,*fn,*mode,*ch;
  static char plotm[1024]="plot ",plotq[1024]="plot ",plotv[1024]="plot ";
  double av=0,err,cov,COVERAGE=0;
  double msdb=0,mscdb=0;

  void (*MYAdd)(char *name,double weight,double X,double Y);
  double (*MYRes)(char *name,char *key);
  void (*MYFree)(void);

#if CHECKHEAP==2
  AllocRange=128;
#endif

  memset(ncps,0,sizeof(ncps));

  if (narg<=5) {
    fprintf(stderr,"\
Playback file analysis to calculate diffusivity and conductivity. Call by:\n\
  plb2diff COMMAND SIMNAME NO[-START] BLOCK[-OMIT] {NBLOCKS|-COVERAGE}\n\
           [MFACTOR-[QFACTOR-[VFACTOR]]] [PLOT]]\n\
where\n\
  COMMAND : to generate SIMNAME.[mq].cp, e.g., \"cook -f -g SYSNAME\"\n\
            (more cook* options may be needed)\n\
  SIMNAME : simulation name for cook*\n\
  NO      : # of frames from SIMNAME.plb (use plbinfo); NO=0: autodetect\n\
  START   : # of frames skipped from CP start (0=none skipped)\n\
            -> use START>0 if the beginning is unequilibrated\n\
  BLOCK   : block size for 1 analysis in frames; BLOCK<1: in full length NO\n\
  OMIT    : # of frames not used for linear regression at block beginning\n\
            (default=BLOCK/4; check the <..(t)> graphs for linear parts)\n\
  NBLOCKS : # of blocks; they may overlap (coverage>1) or be separated by\n\
            unused frames (coverage<1)  -> recommended coverage ~ 3\n\
  COVERAGE: min. coverage (NBLOCKS will be calculated)\n\
  == the following factors are calculated in program units (ps,A^3,K) ==\n\
  MFACTOR : (for diffusivity): 1/dt.plb (NVT), <V>^(2/3)/dt.plb (NPT)\n\
  QFACTOR : (for cond.): 1/(V*T*dt.plb) (NVT), 1/(<V>^(1/3)*T*dt.plb) (NPT)\n\
  VFACTOR : (for visc, WRONG), 1/(T*dt.plb) (NVT), (NPT not implemented)\n\
  PLOT    : launch plots 1=dif/blk 2=cond/blk 4=MSD(t) 8=MSCD(t)... (df.=15)\n\
            plots of MSD,MSCD have a linear term substracted, with b exported\n\
            for MSD of the 1st species, for MSCD of the sum (giving the bulk\n\
            conductivity): use b B ^B / * to change b\n\
            The last column for MSD should be noisy zero (with b=0)\n\
Output files:\n\
  SIMNAME.G1 : rewritten by plb2diff, used by cook* -f -g as inputs\n\
  SIMNAME.S1 : outputs of cook* -f -g\n\
  SIMNAME.cdf : plb2diff protocol\n\
  SIMNAME.av[mq].dat : <MSD(t)> and <MSCD(t)> averaged over all blocks\n\
  SIMNAME.[mq].cp : last block MSD, MSCD\n\
  SIMNAME.[mq][0-].dat : linear regression results of blocks, species 0,..\n\
  SIMNAME.difm.dat,SIMNAME.difm.dat : as above, total MSD and MSCD\n\
Environment:\n\
  PLB2DIFF=2 will use fit to a+b*t (default)\n\
  PLB2DIFF=3 will use fit to a+b*t+c/sqrt(t) (hydrodynamic tail)\n\
             NOT IMPLEMENTED: error of the fit not calculated\n\
Bugs:\n\
  viscosity support is wrong (Einstein formula probably does not exist)\n\
  version cook -f-1 (stored SIMNAME.1,.. instead of SIMNAME.plb) not tested\n\
Example:\n\
  plb2diff \"cookpol -f -g ions\" salt 2000 200-30 20 13171-0.0007807\n\
See also:\n\
  plbmsd plbbox plbinfo plot showcp\n");
   exit(0); }

  initscroll(0);

  MYFree=LRFree;
  MYAdd=LRAdd;
  MYRes=LRRes;

  if (getenv("PLB2DIFF")) {
    int i=atoi(getenv("PLB2DIFF"));
    if (i==2) /**/ ;
    else if (i==3) {
      fprintf(stderr,"WARNING: using fit to  a+b*t+c/sqrt(t), error of fit not available\n");
      MYFree=SDFree;
      MYAdd=SDAdd;
      MYRes=SDRes; }
    else ERROR(("PLB2DIFF must be 2 or 3")) }

  COMMAND=arg[1];
  if (!(strstr(COMMAND,"-f") && strstr(COMMAND,"-g")) && !strstr(COMMAND,".sh"))
    ERROR(("no options -f -g nor .sh in \"%s\"\n\
*** if you are sure, ignore this error",COMMAND))
  SIMNAME=arg[2];

  NO=atoi(arg[3]);
  if (NO<0) ERROR(("invalid NO=%d",NO))
  if (NO==0) NO=plbno(string("%s.plb",SIMNAME));

  ch=strchr(arg[3],'-');
  if (ch) START=atoi(ch+1);

  DBLOCK=atof(arg[4]);
  if (DBLOCK<0) { ERROR(("invalid BLOCK=%g",DBLOCK)) DBLOCK=0; }
  if (DBLOCK==0) DBLOCK=0.25;
  if (DBLOCK<=1) {
    BLOCK=(int)(DBLOCK*NO);
    prt("BLOCK=%d set",BLOCK); }
  else
    BLOCK=DBLOCK;
  ch=strchr(arg[4],'-');
  if (ch) OMIT=atoi(ch+1);
  else {
    OMIT=BLOCK/4;
    prt("OMIT=%d set"); }

  if (arg[5][0]=='-') COVERAGE=-atof(arg[5]);
  else NBLOCKS=atoi(arg[5]);

  if (narg>6) {
    MFACTOR=atof(arg[6]);
    ch=strchr(arg[6],'-');
    if (ch) {
      QFACTOR=atof(++ch);
      ch=strchr(ch,'-');
      if (ch) VFACTOR=atof(++ch); } }

  if (narg>7) PLOT=atof(arg[7]);

  fprintf(stderr,"COMMAND=\"%s\"  SIMNAME=\"%s\"\n",COMMAND,SIMNAME);

  if (NO<BLOCK) ERROR(("NO=%d < BLOCK=%d",NO,BLOCK))
  if (OMIT>=BLOCK) ERROR(("OMIT=%d >= BLOCK=%d",OMIT,BLOCK))
  if (START>=NO) ERROR(("START=%d >= NO=%d",START,NO))

  if (COVERAGE==0 && NBLOCKS==0) ERROR(("zero COVERAGE and NBLOCKS"))

  if (COVERAGE!=0)
    for (NBLOCKS=1; COVERAGE>(double)BLOCK*(NBLOCKS-0.999999999999999)/(NO-BLOCK-START); NBLOCKS++) ;

  fprintf(stderr,"NO=%d START=%d BLOCK=%d OMIT=%d NBLOCKS=%d\n",
          NO,START,BLOCK,OMIT,NBLOCKS);

  out=fopen(string("%s.cdf",SIMNAME),"wt");

  if (OMIT==0) {
    if (MYRes==LRRes) WARNING(("OMIT=0 means theat MSD(0)=0 is included in regression"))
    else ERROR(("OMIT=0, i.e., MSD(0)=0 not defined for c/sqrt(t)")) }

  prt_("Called by:\n ");
  loop (ir,0,narg)
    if (strpbrk(arg[ir]," \t")) prt_(" \"%s\"",arg[ir]);
    else prt_(" %s",arg[ir]);

  fn=string("%s.aux",SIMNAME);
  prt("\n\nAuxiliary file %s %sremoved\n",fn,remove(fn)?"not ":"");

  prt("no=%d frames",NO);
  prt("Coverage = (frames in all blocks)/(all frames) = %g",
      cov=(double)BLOCK/(NO-BLOCK-START)*(NBLOCKS-1));
  if (cov<0.9) WARNING(("Too small coverage - there are unused data, recommended 3"))
  if (cov>10) WARNING(("Too large coverage - data are correlated, recommended 3"))
  if (cov<2 || cov>5) fprintf(stderr,"Coverage=%g is outside the optimum range [2,5]\n",cov);

  StaSet(1,2,2,4);

  loop (iblock,0,NBLOCKS) {
    if (NBLOCKS==1) init=1;
    else init=1+START+(long long int)iblock*(NO-BLOCK-START)/(NBLOCKS-1);
    no=init+BLOCK-1;
    prt("block %d: init=%d no=%d",iblock+1,init,no);
    fprintf(stderr,"## %d/%d ## frames %d-%d ## ",iblock+1,NBLOCKS,init,no);
    get=fopen(fn=string("%s.G1",SIMNAME),"wt");
    if (!get) ERROR(("cannot write to %s",fn))

    fprintf(get,"! ");
    loop (ir,0,narg)
      if (strpbrk(arg[ir]," \t")) fprintf(get," \"%s\"",arg[ir]);
      else fprintf(get," %s",arg[ir]);
    fprintf(get,"\nno=%d init=%d el.kappa=0; quit=1;\n",no,init);
    if (fclose(get)) ERROR(("%s write error",fn))
    fn=string("%s %s",COMMAND,SIMNAME);
    if (system(fn)) ERROR(("`%s' execution error",fn))

    for (mode="mqv"; *mode; mode++) {
      FILE *cp=fopen(fn=string("%s.%c.cp",SIMNAME,*mode),"rb");
      float h=0;

      if (cp) {
        fread(&h,4,1,cp);
        if (h!=CPmark) ERROR(("%s bad header",fn))
        fread(&ncp,4,1,cp);
#define NVCP 14 /* normally 10 */
        if (*mode=='v') {
          if (ncp!=NVCP) ERROR(("%s: ncp=%d (%d expected)",fn,ncp,NVCP)) }
        else {
          if (ncp<2 || ncp>1000) {
            /* cf. HOOK in old version of plb2diff */
            ERROR(("%s : NCP=%d read",fn,ncp))
          ncp=2; } }
        ncps[(int)*mode]=ncp;
        loop (icp,2,ncp) fread(&h,4,1,cp);

        if (ncps['q'] && ncps['m']!=ncps['q'])
          ERROR(("NCP=%d in %s.m.cp != NCP=%d in %s.q.cp",
                 ncps['m'],SIMNAME,ncps['q'],SIMNAME))

        /* allocations in 1st iblock only */
        if (iblock==0) switch (*mode) {
          case 'm':
            alloc(r,max(NVCP,ncp)*4);
            allocarray(mr,BLOCK);
            loop (ir,0,BLOCK) allocarrayzero(mr[ir],ncp);
            allocarrayzero(mf,ncp);
            allocarrayzero(summ,ncp);
            break;
          case 'q':
	    allocarray(qr,BLOCK);
	    loop (ir,0,BLOCK) allocarrayzero(qr[ir],ncp);
	    allocarrayzero(qf,ncp);
	    allocarrayzero(sumq,ncp);
	    break;
          case 'v':
            if (ncp!=NVCP) Error("internal");
            allocarray(vr,BLOCK);
            loop (ir,0,BLOCK) allocarrayzero(vr[ir],ncp);
            allocarrayzero(vf,ncp); /* vf[0..8] not used */
            allocarrayzero(sumv,ncp);
            break;
          default: ERROR((mode)); }

	loop (ir,0,BLOCK) {
	  if (ncp!=fread(r,4,ncp,cp)) ERROR(("%s truncated",fn))
          if (ir>=OMIT) {
            if (*mode=='v') loop (icp,9,ncp) {
              MYAdd(vname(icp),
                    1./(ir+(ir==0)), /* weight, NB: ir=0 is MSD=0 by def. */
                    ir, r[icp]); }
            else loop (icp,0,ncp)
              MYAdd(icp<ncp-1 ? string("%c%d",*mode,icp)
                              : *mode=='m' ? "difm" : "cond",
                    1./(ir+(ir==0)), /* weight */
                    ir, r[icp]); }
          //          prt("%g %g %g DBG",ir+1.,r[0],1/(ir+1.));
          loop (icp,0,ncp)
            (*mode=='m' ? mr : *mode=='q' ? qr : vr) [ir][icp] += r[icp]; }

        loop (icp,0,ncp) {
	  FILE **f = *mode=='m' ? mf : *mode=='q' ? qf : vf;
	  char *name=icp<ncp-1
	    ? string("%c%d",*mode,icp)
	    : *mode=='m' ? "difm" : "cond";

	  if (*mode=='v') {
	    if (icp<9) continue;
	    name=vname(icp); }

          sum=*mode=='m' ? summ : *mode=='q' ? sumq : sumv;
          //prt("%s a=%g b=%g c=%g",name,MYRes(name,"A"),MYRes(name,"B"),MYRes(name,"C"));
	  av=MYRes(name,"B");
	  sum[icp].s+=av;
	  sum[icp].q+=av*av;
	  sum[icp].n++;

	  err=MYRes(name,"DB");
	  if (!f[icp]) {
	    char *fn=string("%s.%s.dat",SIMNAME,name);

	    f[icp]=fopen(fn,"wt");
	    sprintf(strend(*mode=='m'?plotm:*mode=='q'?plotq:plotv),
		    " %s:A:B:-:C",fn);

	    fprintf(f[icp],"# ");

	    loop (ir,0,narg)
	      if (strpbrk(arg[ir]," \t")) fprintf(f[icp]," \"%s\"",arg[ir]);
	      else fprintf(f[icp]," %s",arg[ir]);

	    fprintf(f[icp],"\n# init lin.regr.slope lin.regr.err\n"); }
	  fprintf(f[icp],"%d %g %g\n",init,av,err);
	  StaAdd(name,av); }
	StaAdd(string("sqrt(|%c|)",*mode),sqrt(fabs(av)));
	fclose(cp); }
      else {
	static int pass;

	if (!pass++)
          WARNING(("missing %s (MORE WARNINGS SUPPRESSED)",fn)) }
      } /* mode */

    MYFree();
   } /* iblock */

  loop (icp,0,ncps['m'])
    if (!mf[icp]) ERROR(("close m %d",icp)) else fclose(mf[icp]);
  loop (icp,0,ncps['q'])
    if (!qf[icp]) ERROR(("close q %d",icp)) else fclose(qf[icp]);
  if (vf) loop (icp,9,ncps['v'])
    if (!vf[icp]) ERROR(("close v %d",icp)) else fclose(vf[icp]);

  StaPrintAll(NULL);

  /* MSD(t) and MSCD(t): print files and calculate fits */
  for (mode="mq"; *mode; mode++) {
    FILE *msd;

    xr=*mode=='m' ? mr : qr;
    msd=fopen(string("%s.%s.dat",SIMNAME,*mode=='m'?"msd":"mscd"),"wt");
    fprintf(msd,"# %s %s %s\n",arg[3],arg[4],arg[5]);
    fprintf(msd,"#t/dt  <MSD_0(t)> ... <sum MSD(t)>=0\n");
    fprintf(msd,"#    indented points not used in the fit\n");

    pass=1;
    loop (ir,0,BLOCK) {
      if (ir>=OMIT) {
	loop (icp,0,ncps[(int)*mode]) {
          char *name=icp<ncps[(int)*mode]-1 ? string("%c%d",*mode,icp)
                                            : *mode=='m'?"difm":"cond";
          if (pass) fprintf(msd," %s",name);
          MYAdd(name, 1./(ir+(ir==0)), ir, xr[ir][icp]/NBLOCKS); }
        fprintf(msd,"%2d ",ir); }
      else
        fprintf(msd,"   %2d ",ir); /* line shifted: data not used in fit */
      loop (icp,0,ncps[(int)*mode])
	fprintf(msd,"%g%c", xr[ir][icp]/NBLOCKS,"\n "[icp<ncps[(int)*mode]-1]);
      if (pass) fprintf(msd,"\n");
      pass=0; }

    fclose(msd); }

  /* MSD(t) and MSCD(t): print fits */
  for (mode="mq"; *mode; mode++) {
    FILE *fit;
    double a,b,c,x,y;

    xr=*mode=='m' ? mr : qr;
    fit=fopen(string("%s.%s.fit",SIMNAME,*mode=='m'?"msd":"mscd"),"wt");
    fprintf(fit,"# %s %s %s\n",arg[3],arg[4],arg[5]);
    fprintf(fit,"#t/dt  <MSCD_0(t)> ... <sum MSCD(t)>=0\n");
    fprintf(fit,"#    indented points not used in the fit\n");
    loop (icp,0,ncps[(int)*mode]) {
      char *name=icp<ncps[(int)*mode]-1 ? string("%c%d",*mode,icp)
                                        : *mode=='m'?"difm":"cond";
      fprintf(fit,"# species %s: a=%.9g b=%.9g %s\n",
              icp==ncps[(int)*mode]-1?"sum":string("%d",icp),
              MYRes(name,"A"),MYRes(name,"B"),
              MYRes==SDRes?string("c=%.9g",MYRes(name,"C")):""); }

    pass=1;
    for (x=0.5; x<BLOCK-0.3; x+=0.25) {
      if (fabs(OMIT+0.5-x)<0.03) fprintf(fit,"# end of extrapolated line\n#");
      fprintf(fit,"%.2f ",x);
      loop (icp,0,ncps[(int)*mode]) {
        char *name=icp<ncps[(int)*mode]-1 ? string("%c%d",*mode,icp)
                                          : *mode=='m'?"difm":"cond";
        if (pass) fprintf(fit," %s",name);
        a=MYRes(name,"A");
        b=MYRes(name,"B");
        if (*mode=='m' && !icp) msdb=b; /* the 1st species */
        if (*mode=='q') mscdb=b; /* the sum */
        y=a+x*b;
        if (MYRes==SDRes) {
          c=MYRes(name,"C");
          y=a+x*b+c/sqrt(x); }

	fprintf(fit,"%g%c", y,"\n "[icp<ncps[(int)*mode]-1]); }

      if (pass) fprintf(fit,"\n");
      pass=0; }

    fclose(fit); }

  if (vf) { /* unsuccessful: do not use */
    FILE *msptd=fopen(string("%s.msptd.dat",SIMNAME),"wt");
    fprintf(msptd,"#");
    pass=1;
    loop (ir,0,BLOCK) {
      if (ir>=OMIT)
	loop (icp,0,ncps['v']) {
	char *name=vname(icp);

	if (pass) fprintf(msptd," %s",name);
	MYAdd(name, 1./(ir+(ir==0)), ir, vr[ir][icp]/NBLOCKS); }
      loop (icp,0,ncps['v'])
	fprintf(msptd,"%g%c", vr[ir][icp]/NBLOCKS,"\n "[icp<NVCP-1]);
      if (pass) fprintf(msptd,"\n");
      pass=0; }

    fclose(msptd); }

  prt("\n\n\
In the following statistics, ERR was obtained from blocks (see data in\n\
files SIMNAME.m0.dat, SIMNAME.q0.dat, SIMNAME.m1.dat, ..., SIMNAME.visc.dat)\n\
treated as independent data and for coverage>1 multiplied by sqrt(coverage)\n\
Relative error=ERR/fit slope\n");

  if (MYAdd==LRAdd) prt("FITTING FUNCTION = a+b*t (no hydrodynamic tail)");
  else prt("FITTING FUNCTION = a+b*t+c/sqrt(t) (linear with with hydrodynamic tail)");

  for (mode="mqv"; *mode; mode++) {
    int NCP=ncps[(int)*mode];

    switch (*mode) {
      case 'm':
	prt("\nDIFFUSION");
	header(" name   n   fit slope  fit err  stderr     D/MFACTOR     D[m^2/s]   stderr ");
	break;
      case 'q':
	prt("\nCONDUCTIVITY");
	header(" name   n   fit slope  fit err  stderr    kappa/QFACTOR  kappa[S/m] stderr ");
	break;
      case 'v':
	if (!vf) break;
	prt("\nVISCOSITY (THIS IS WRONG!)");
	header(" name   n   fit slope  fit err  stderr    eta/VFACTOR    eta[kg/m/s] stderr");
      }

    loop (icp,0,NCP) {
      char *name=icp<NCP-1
	? string("%c%d",*mode,icp)
	: *mode=='m' ? "difm" : "cond";
      sum=*mode=='m' ? summ : *mode=='q' ? sumq : sumv;

      if (!sum) continue;

      if (vf && *mode=='v') name=vname(icp);

      av=MYRes(name,"B");
      err=MYRes(name,"DB");

      if (fabs(av-sum[icp].s/sum[icp].n)/av>3e-6)
	WARNING(("%s: average inconsistent: %g %g",name,av,sum[icp].s/sum[icp].n))
      if (sum[icp].n>1)
	ERR=sqrt(fmax(cov,1.0)*(sum[icp].q/sum[icp].n-Sqr(sum[icp].s/sum[icp].n))/(sum[icp].n-1));
      else
	ERR=0;

      prt_("%5s %4.0f %12.6g %8.2g %9.3g ",name,MYRes(name,"N"),av,err,ERR);
      switch (*mode) {
        case 'm':
	  prt(" %12.6g %12.6g %8.3g",
              av*1e-8,av*1e-8*MFACTOR, sqrt(err*err+ERR*ERR)*1e-8*MFACTOR);
	  break;
        case 'q':
	  prt(" %12.6g %12.6g %8.3g",
              av*111.26502, av*111.26502*QFACTOR, sqrt(err*err+ERR*ERR)*111.26502*QFACTOR);
	  break;
        case 'v':
	  if (!vf) break;
	  if (icp>=9)
            prt(" %12.6g %12.6g %8.3g",
                av*1.3806503e-5,av*1.3806503e-5*VFACTOR,sqrt(err*err+ERR*ERR)*1.3806503e-5*VFACTOR);
	  else
	    prt(" (s=%g)",av/err);
	  break; } }
    header("");

    switch (*mode) {
      case 'm':
        prt("diffusion coefficients [in m^2/s] are obtained by multiplying column\n\
D/MFACTOR by MFACTOR, which is:\n\
  1/dt.plb          for NVT simulation (tau.P=0)\n\
  <V>^(2/3)/dt.plb  for NPT simulation (tau.P>0) [??? for not-cube]\n\
<V> = averaged volume in A^3 (see SIMNAME.prt)\n\
dt.plb is in ps [to be replaced by dt.cfg if SIMNAME.1,... are used and -f-1]\n\
NOTE: the generated .dat files are in program units and the additional factor\n\
(to get units m^2/s from the slope) is 1e-8");
	if (MFACTOR) prt("you specified MFACTOR=%g",MFACTOR);
	break;
      case 'q':
	prt("conductivities [in S/m] are obtained by multiplying column kappa/QFACTOR by\n\
QFACTOR, which is:\n\
  1/(V*T*dt.plb)          for NVT simulation (tau.P=0)\n\
  1/(<V>^(1/3)*T*dt.plb)  for NPT simulation (tau.P>0) [??? for not-cube]\n\
(T = temperature in K, V = volume in A^3, dt.plb is in ps)\n\
NOTE: the generated .dat files are in program units and the additional factor\n\
(to get units S/m from the slope) is 111.26502");
	if (QFACTOR) prt("you specified QFACTOR=%g",QFACTOR);
	break;
      case 'v':
	if (!vf) break;
	prt("viscosity [in kg/m/s] is obtained by multiplying eta/VFACTOR by\n\
VFACTOR, which is:\n\
  V/(T*dt.plb)            for NVT simulation (tau.P=0)\n\
  ???                     for NPT simulation (tau.P>0 - not implemented)\n\
(T = temperature in K, V = volume in A^3, dt.plb is in ps)\n\
NOTE: avvXY for X!=Y should be zero within err\n\
NOTE: the generated .dat files are in program units and the additional factor\n\
(to get units kg/m/s from the slope) is 1.3806503e-5");
	if (VFACTOR) prt("you specified VFACTOR=%g",VFACTOR);
	break; }
     }
  {
    static struct maxjump_s maxjump,max1jump;
    struct maxjump_s mj,m1j;
    int k;
    FILE *aux;

    prt("\nPERIODIC BOX STATISTICS");

    aux=fopen(fn=string("%s.aux",SIMNAME),"rb");
    if (!aux) ERROR(("open %s",fn))
    loop (iblock,0,NBLOCKS) {

      if (!fread(&mj,sizeof(mj),1,aux)) ERROR(("%s truncated",fn))
      if (!fread(&m1j,sizeof(m1j),1,aux)) ERROR(("%s truncated",fn))

      loop (k,0,3) {
	if (fabs(mj.xi[k])>fabs(maxjump.xi[k])) {
	  maxjump.xi[k]=mj.xi[k];
	  maxjump.i[k]=mj.i[k];
	  maxjump.n[k]=mj.n[k];
	  maxjump.frame[k]=mj.frame[k];
	  maxjump.init=mj.init;
	  maxjump.no  =mj.no;
	  maxjump.f   =mj.f; }
	if (fabs(m1j.xi[k])>fabs(max1jump.xi[k])) {
	  max1jump.xi[k]=m1j.xi[k];
	  max1jump.i[k]=m1j.i[k];
	  max1jump.n[k]=m1j.n[k];
	  max1jump.frame[k]=m1j.frame[k];
	  max1jump.init=m1j.init;
	  max1jump.no  =m1j.no;
	  max1jump.f   =m1j.f; } } }

    if (fread(&mj,sizeof(mj),1,aux)) ERROR(("unexpected extra data in %s",fn))

    loop (k,0,3) {
      prt("\n%c-axis\n\
max jump over periodic boxes in blocks:\n\
  %f*L (init=%d frame=%d mol.site=%d.%d)",
         k+'x',
         maxjump.xi[k],maxjump.init,maxjump.frame[k],maxjump.n[k],maxjump.i[k]);
     prt("max jump between consecutive frames:\n\
  %f*L (init=%d frames=%d->%d mol.site=%d.%d)",
     max1jump.xi[k],max1jump.init,max1jump.frame[k]-1,max1jump.frame[k],max1jump.n[k],max1jump.i[k]);
     if (fabs(max1jump.xi[k])>.4)
       WARNING(("max jump between consecutive frames is very close to L/2\n\
*** diffusions may be wrong - check difm (should be tiny)")) }

    if (fabs(MYRes("difm","B"))>3e-7)
      WARNING(("difm=%g is not tiny (up to 1e-5 acceptable for large systems)\n\
*** possible causes:\n\
*** - nonzero momentum (check variable  drift )\n\
*** - nearest image problems because of too long  dt.plb\n\
***   (see max jump between consecutive frames)",MYRes("difm","B")))

    fclose(aux);
    if (remove(fn)) ERROR(("cannot remove %s",fn))
  }

  sh=fopen(fn=string("%s.plot.sh",SIMNAME),"wt");
  if (!sh) ERROR(("open %s",fn))

  fprintf(sh,"#!/bin/sh\n");

  Setenv("PLOTGEOMETRY","600x400+15+15");
  Setenv("PLOTNAME",string("diffusion %s",arg[2]));
  strcat(plotm," &");
  fprintf(sh,"%s\n\n",plotm);
  if (PLOT&1) { system(plotm); sleep(1); }

  Setenv("PLOTGEOMETRY","600x400+30+30");
  Setenv("PLOTNAME",string("conductivity %s",arg[2]));
  strcat(plotq," &");
  fprintf(sh,"%s\n\n",plotq);
  if (PLOT&2) { system(plotq); sleep(1); }

  if (vf) {
    Setenv("PLOTGEOMETRY","600x400+80+80");
    Setenv("PLOTNAME",string("viscosity %s",arg[2]));
    strcat(plotv," &");
    fprintf(sh,"%s\n\n",plotv);
    if (PLOT&16) { system(plotv); sleep(1); } }

  sprintf(plotm,"plot %s.msd.dat",SIMNAME);
  fn=strend(plotm);
  loop (icp,0,ncp) sprintf(strend(plotm),":1:\"%c-b*A:%do\" ",'B'+icp,icp+1);
  Setenv("PLOTGEOMETRY","600x400+45+45");
  Setenv("PLOTNAME",string("<MSD(t) %s>",arg[2]));
  Setenv("b",string("%.9g",msdb));
  sprintf(strend(plotm)," %s.msd.fit",SIMNAME);
  loop (icp,0,ncp) sprintf(strend(plotm),":1:\"%c-b*A:%d-\" ",'B'+icp,icp+1);
  strcat(plotm," &");
  fprintf(sh,"%s\n\n",plotm);
  if (PLOT&4) { system(plotm); sleep(1); }

  sprintf(plotm,"plot %s.mscd.dat",SIMNAME);
  fn=strend(plotm);
  loop (icp,0,ncp) sprintf(strend(plotm),":1:\"%c-b*A:%do\" ",'B'+icp,icp+1);
  Setenv("PLOTGEOMETRY","600x400+60+60");
  Setenv("PLOTNAME",string("<MSCD(t) %s>",arg[2]));
  Setenv("b",string("%.9g",mscdb));
  sprintf(strend(plotm)," %s.mscd.fit",SIMNAME);
  loop (icp,0,ncp) sprintf(strend(plotm),":1:\"%c-b*A:%d-\" ",'B'+icp,icp+1);
  strcat(plotm," &");
  fprintf(sh,"%s\n\n",plotm);
  if (PLOT&8) system(plotm);

  if (vf) {
    Setenv("PLOTGEOMETRY","600x400+95+95");
    Setenv("PLOTNAME",string("<MSVISC(t) %s>",arg[2]));
    fn[-5]='v'; /* dirty trick */
    strcpy(fn,":0:\"J+n*a\"");
    fprintf(sh,"%s\n",plotm);
    if (PLOT&32) system(plotm); }

  fclose(sh);
  system(string("chmod +x %s.plot.sh",SIMNAME));

  return 0;
}
